/*
 * ACPI configurator and AML interpreter
 */
#include "kernel.h"
#include "acpi.h"
#include "aml.h"

enum
{
	Maxtblsz = 64 * 1024,
};

typedef struct Parse Parse;

struct Parse
{
	char*	sig;
	int	(*f)(uchar*, int);	/* return nil to keep vmap */
};

Biobuf* bout;
Biobuf boutbuf;

static Parse ptables[] =
{
	"SRAT",	acpisrat,
	"MSCT",	acpimsct,
	"DSDT", acpidsdt,
	"SSDT", acpidsdt,
};

static void
usage(void)
{
	fprint(2, "usage: %s: [-AESbcdinosx] [-f file] [op...]\n", argv0);
	fprint(2, "where op is\n");
	fprint(2, "\t/a/path\tto dump the ns rooted there\n");
	fprint(2, "\t%%/a/path\tto configure devices in that tree\n");
	fprint(2, "\t!/method/path\tto call a method\n");
	fprint(2, "\t@/method/path\tto dump source for a method\n");
	fprint(2, "\t+/method/path\tto set tracing for a method\n");
	fprint(2, "\t-/method/path\tto clear tracing for a method\n");
	exits(nil);
}

static int
readtable(Biobuf *bin)
{
	Parse *p;
	char *ln;
	char *toks[50];
	int ntoks;
	int n;
	long len;
	int i;
	uchar *amlblock;
	int ishexdump;

	while((ln = Brdstr(bin, '\n', 1)) != nil){
		for(i = 0; i < nelem(ptables); i++){
			p = &ptables[i];
			if(strncmp(ln, p->sig, 4) == 0)
				goto Found;
		}
		free(ln);
	}
	return -1;
Found:
	len = Maxtblsz;	/* BUG: should allocate only what we need */

	dprint("reading %s...\n", ln);
	free(ln);
	amlblock = malloc(len);
	if(amlblock == nil)
		sysfatal("no memory");
	n = ishexdump = 0;
	while(n < len && (ln = Brdstr(bin, '\n', 1)) != nil){
		if(ishexdump && strlen(ln) > 56)
			ln[56] = 0;
		ntoks = tokenize(ln, toks, nelem(toks));
		if(ntoks <= 0)
			break;
		if(ntoks > 17){
			ishexdump = 1;
			ntoks = 17;
		}
		for(i = 1; n < len && i < ntoks; i++){
			if(strlen(toks[i]) > 2)
				break;
			amlblock[n++] = strtol(toks[i], nil, 16);
			//Bprint(bout, "%s ", toks[i]);
		}
		//Bprint(bout, "\n");
		free(ln);
	}
	if(n <= Sdthdrsz){
		free(amlblock);
		return -1;
	}
	if(p->f(amlblock, n) <= 0)
		free(amlblock);
	return 0;
}

static int
dumpmethod(Aobj *o, int flags)
{
	xprint("method %s/%d:\n", o->name->path, o->method.nargs);
	if(o->method.f != nil)
		xprint("\tbuiltin\n");
	else
		amleval(o->method.ps, o->method.pe, nil, o->name, nil, flags);
	xprint("\n");
	return 1;
}

/*
 * We leak args here.
 */
static int
fetchargs(int nargs, Aobj **args, int argc, char **argv)
{
	int n;

	for(n = 0; n < nargs && n < argc; n++)
		if(argv[n][0] == '0')
			args[n] = amlnewi(strtoul(argv[n], 0, 0));
		else
			args[n] = amlnews(argv[n]);
	if(n < nargs)
		return -1;
	return 0;
}

static void
loadtables(char *fname, int amlflags)
{
	Biobuf *bin;

	bin = Bopen(fname, OREAD);
	if(bin == nil)
		sysfatal("%s: %s: %r", argv0, fname);
	amlinit();
	acpiflags = amlflags;

	dprint("%s: reading %s...\n", argv0, fname);
	while(readtable(bin) == 0)
		Bflush(bout);
	Bterm(bin);
}

static void
acpi(char *fname, int amlflags, int docfg)
{
	acpiflags = amlflags;
	if(acpiinit() < 0)
		return;
	loadtables(fname, amlflags);
	if(docfg == 0)
		return;
	dxprint("9am acpi gpes...\n");
	acpigpes();
	dxprint("9am acpi pwrs...\n");
	acpipwrs();
	dxprint("9am acpi apics...\n");
	acpiapics();
	dxprint("9am acpi conf...\n");
	acpiconf(acpins);
	dxprint("9am acpi done\n");
}

static void
cmds(int argc, char *argv[], int amlflags)
{
	Aobj*args[Nmargs], *o;
	int x, i, na, flags;
	Aname *n;

	for(i = 0; i < argc; i++){
		x = argv[i][0];
		if(x == '!' || x == '@' || x == '%' || x == '+' || x == '-')
			argv[i]++;
		n = amlwalk(argv[i], Walkns, acpins);
		if(n == nil){
			xprint("%s: name not found\n", argv[i]);
			continue;
		}
		if(x == '!' || x == '@' || x == '+' || x == '-')
			if(n->o->type != Omethod){
				xprint("%s: not a method\n", argv[i]);
				break;
			}
		switch(x){
		case '!':
			flags = amlflags;
			memset(args, 0, sizeof(args));
			na = n->o->method.nargs;
			if(fetchargs(na, args, argc-i-1, argv+i+1)<0){
				xprint("%s: not enough args\n", argv[i]);
				break;
			}
			o = amlcall(n->o, args, nil, flags);
			if(o == nil)
				xprint("%s: %s\n", argv[i], amlerrstk.msg);
			else
				xprint("%s: %O\n", argv[i], o);
			amlclose(o);
			i += na;
			break;
		case '@':
			flags = amlflags|Emdump|Eidump|Enoeval;
			dumpmethod(n->o, flags);
			break;
		case '+':
			n->o->method.trace = 1;		/* force on */
			break;
		case '-':
			n->o->method.trace = -1;	/* force off */
			break;
		case '%':
			acpiflags = amlflags;
			acpiconf(n);
			break;
		default:
			dumpns(n, 0, 1);
		}
		FLUSH;
	}

}

static int
bind9am(void)
{
	bind("#ec", "/env", MREPL);
	bind("#e", "/env", MBEFORE|MCREATE);
	if(access("/dev/acpictl", AWRITE) == 0)
		return 0;
	return bind("#9", "/dev", MAFTER);
}

static void
checkno9am(char *argv[])
{
	char *p;

	p = getenv("*9am");
	if(p == nil || p[0] != '1' || bind9am() < 0){
		fprint(2, "no 9am\n");
		exec("/boot/boot", argv);
		exits("noboot");
	}
	print("9am...\n");
	free(p);
}

static void
hwinit(void)
{
	int fd;

	dxprint("9am: hwinit...\n");
	fd = open("#9/acpictl", OWRITE);
	if(fd < 0)
		sysfatal("can't open ctl: %r");
	if(write(fd, "init", 4) != 4)
		sysfatal("can't hwinit");
	close(fd);
	dxprint("9am: hwinit done\n");
}

void
main(int argc, char *argv[])
{
	char **bootargv;
	char *p9args[10];
	char *p, *fname;
	int sflag, amlflags, cfgflags;

	acpistackbottom = (char*)&argc;	/* debug */

	/*
	 * If we are called as /boot/boot, argv is for
	 * the real /boot/boot, and our args come from 9amargs.
	 */
	argv0 = "9am";
	bootargv = nil;
	if(strcmp(argv[0], "/boot/boot") == 0){
		bootargv = argv;
		checkno9am(bootargv);
		argv = p9args;
		p9args[0] = "9am";
		argc = 1;

		p = getenv("*debug9am");
		if(p != nil)
			argc += tokenize(p, p9args+1, nelem(p9args)-2);
		p9args[argc] = nil;
	}

	amlflags = sflag = 0;
	fname = nil;

	ARGBEGIN{
	case 'f':			/* use file to read tables */
		fname = EARGF(usage());
		break;

	/* only useful for tracing and debugging */
	case 's':			/* dump stats */
		sflag++;
		break;
	case 'n':			/* do not perform I/O to /dev */
		amlflags |= Edryrun;
		break;
	case 'd':			/* activate "dprint" messages */
		amlflags |= Evdump;
		break;
	case 'i':			/* dump instructions for tables */
		amlflags |= Eidump;
		break;
	case 'x':			/* report actions */
		amlflags |= Exdump;
		break;
	case 'b':			/* report ns changes */
		amlflags |= Ebdump;
		break;
	case 'o':
		amlflags |= Eodump;	/* report object i/o */
		break;

	/* more debugging */
	case 'A':			/* abort on errors */
		amlflags |= Eabort;
		break;
	case 'E':			/* dump environments after ops */
		amlflags |= Eedump;
		break;
	case 'S':			/* check out the stack (slow!!) */
		amlflags |= Escheck;
		break;
	default:
		usage();
	}ARGEND;

	if(fname == nil){
		if(bind9am() < 0)
			sysfatal("no #9");
		fname = "/dev/acpitbl";
	}
	Binit(&boutbuf, 1, OWRITE);
	bout = &boutbuf;

	cfgflags = amlflags;
	if(argc == 0)
		cfgflags &= Edryrun|Eabort|Escheck;

	acpi(fname, cfgflags, argc == 0);
	cmds(argc, argv, amlflags);

	if(sflag){
		dumpstats();
		dumpopstats();
	}
	Bterm(bout);

	if(bootargv != nil){
		hwinit();
		dprint("boot...\n");
		exec("/boot/boot", bootargv);
		exits("noboot");
	}
	exits(nil);
}
