/*
 * ACPI 4.0 Hardware configuration.
 *
 * Parse tables needed for booting and provide
 * data for p9am(3) and devices. 
 */

#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"../port/error.h"
#include	"mp.h"
#include	"acpi.h"

typedef struct Apicst Apicst;
typedef struct Atable Atable;
typedef struct Facs Facs;
typedef struct Fadt Fadt;
typedef struct Gas Gas;
typedef struct Gpe Gpe;
typedef struct Madt Madt;
typedef struct Parse Parse;
typedef struct Pwr Pwr;
typedef struct Rsdp Rsdp;
typedef struct Sdthdr Sdthdr;

#pragma	varargck	type	"G"	Gas*


enum
{
	Npwr		= 6,	/* number of power states */

	Sdthdrsz	= 36,	/* size of SDT header */

	/* ACPI PM1 control */
	Pm1SciEn	= 0x1,		/* Generate SCI and not SMI */

	/* FADT bootarch flags */
	BAlegacy	= 1,
	BAkbd		= 2,
	BAnovga		= 4,

	/* Apic structure types */
	ASlapic = 0,	/* processor local apic */
	ASioapic,	/* I/O apic */
	ASintovr,	/* Interrupt source override */
	ASnmi,		/* NMI source */
	ASlnmi,		/* local apic nmi */
	ASladdr,	/* local apic address override */
	ASiosapic,	/* I/O sapic */
	ASlsapic,	/* local sapic */
	ASintsrc,	/* platform interrupt sources */
	ASlx2apic,	/* local x2 apic */
	ASlx2nmi,	/* local x2 apic NMI */
};

struct Parse
{
	char*	sig;
	void	(*f)(uchar*, int);
};

struct Gpe
{
	uintptr	stsio;		/* port used for status */
	int	stsbit;		/* bit number */
	uintptr	enio;		/* port used for enable */
	int	enbit;		/* bit number */
	int	nb;		/* event number */
	char*	obj;		/* handler object  */
	int	id;		/* id as supplied by user */
};

struct Pwr
{
	int pm1a;	/* value for pm1a.slp_typ to enter state */
	int pm1b;	/* value for pm1b.slp_typ to enter state */
};

/*
 * Generic address structure. 
 */
struct Gas
{
	u8int	spc;	/* address space id */
	u8int	len;	/* register size in bits */
	u8int	off;	/* bit offset */
	u8int	accsz;	/* 1: byte; 2: word; 3: dword; 4: qword */
	u64int	addr;	/* address (or acpi encoded tbdf + reg) */
};

/* Multiple APIC description table (sw)
 * ACPI system interrupt numbers are virtual, with each
 * ioapic knowing the base number for its first irq.
 */
struct Madt
{
	u64int	laddr;	/* local APIC addr */
	int	pcat;	/* also has PC/AT 8259s */
	Apicst*	st;
};

struct Apicst
{
	int	type;
	Apicst*	next;
	union{
		struct{
			int	pid;	/* processor id */
			int	id;	/* apic id (<= 255) */
		} lapic;
		struct{
			int	id;	/* io apic id */
			u32int	ibase;	/* interrupt base addr. */
			u64int	addr;	/* base address */
		} ioapic, iosapic;
		struct{
			int	irq;	/* bus intr. source */
			int	intr;	/* system interrupt */
			int	flags;	/* INTI flags */
		} intovr;
		struct{
			int	intr;	/* system interrupt */
			int	flags;	/* apic flags */
		} nmi;
		struct{
			int	pid;	/* processor id */
			int	flags;	/* LINT flags */
			int	lint;	/* apic LINTn for nmi */
		} lnmi;
		struct{
			int	pid;	/* processor id */
			int	id;	/* apic id */
			int	eid;	/* apic eid */
			int	puid;	/* processor uid */
			char*	puids;	/* same thing */
		} lsapic;
		struct{
			int	pid;	/* processor id */
			int	peid;	/* processor eid */
			int	iosv;	/* io sapic vector */
			int	intr;	/* global sys intr. */
			int	type;	/* intr type */
			int	flags;	/* apic flags */
			int	any;	/* err sts at any proc */
		} intsrc;
		struct{
			int	id;	/* x2 apic id (lapic >255) */
			int	puid;	/* processor uid */
		} lx2apic;
		struct{
			int	puid;
			int	flags;
			int	intr;
		} lx2nmi;
	};
};

/* Fixed ACPI description table (sw version)
 * Describes implementation and hardware registers.
 * PM* blocks are low level functions.
 * GPE* blocks refer to general purpose events.
 * P_* blocks are for processor features.
 * Contains the address for the DSDT.
 */
struct Fadt
{
	u32int	facs;
	u32int	dsdt;
	/* 1 reserved */
	u8int	pmprofile;
	u16int	sciint;
	u32int	smicmd;
	u8int	acpienable;
	u8int	acpidisable;
	u8int	s4biosreq;
	u8int	pstatecnt;
	u32int	pm1aevtblk;
	u32int	pm1bevtblk;
	u32int	pm1acntblk;
	u32int	pm1bcntblk;
	u32int	pm2cntblk;
	u32int	pmtmrblk;
	u32int	gpe0blk;
	u32int	gpe1blk;
	u8int	pm1evtlen;
	u8int	pm1cntlen;
	u8int	pm2cntlen;
	u8int	pmtmrlen;
	u8int	gpe0blklen;
	u8int	gpe1blklen;
	u8int	gp1base;
	u8int	cstcnt;
	u16int	plvl2lat;
	u16int	plvl3lat;
	u16int	flushsz;
	u16int	flushstride;
	u8int	dutyoff;
	u8int	dutywidth;
	u8int	dayalrm;
	u8int	monalrm;
	u8int	century;
	u16int	iapcbootarch;
	/* 1 reserved */
	u32int	flags;
	Gas	resetreg;
	u8int	resetval;
	/* 3 reserved */
	u64int	xfacs;
	u64int	xdsdt;
	Gas	xpm1aevtblk;
	Gas	xpm1bevtblk;
	Gas	xpm1acntblk;
	Gas	xpm1bcntblk;
	Gas	xpm2cntblk;
	Gas	xpmtmrblk;
	Gas	xgpe0blk;
	Gas	xgpe1blk;
};

/*
 * ACPI table (sw)
 */
struct Atable
{
	Atable*	next;		/* next table in list */
	int	is64;		/* uses 64bits */
	char	sig[5];		/* signature */
	char	oemid[7];	/* oem id str. */
	char	oemtblid[9];	/* oem tbl. id str. */
	uchar*	p;		/* table data, including hdr */
	long	len;		/* size of data in table, including hdr */
};

/*
 * hardware structures
 */

/* Root system description table pointer.
 * Used to locate the root system description table RSDT
 * (or the extended system description table from version 2) XSDT.
 * The XDST contains (after the DST header) a list of pointers to tables:
 *	- FADT	fixed acpi description table.
 *		It points to the DSDT, AML code for the acpi namespace.
 *	- SSDTs	tables with AML code to add to the acpi namespace.
 *	- pointers to other tables (Apics, etc.)
 */
struct Rsdp
{
	u8int	signature[8];		/* "RSD PTR " */
	u8int	rchecksum;
	u8int	oemid[6];
	u8int	revision;
	u8int	raddr[4];		/* RSDT */
	u8int	length[4];
	u8int	xaddr[8];		/* XSDT */
	u8int	xchecksum;		/* XSDT */
	u8int	_33_[3];		/* reserved */
};

/* Header for ACPI description tables
 */
struct Sdthdr
{
	u8int	sig[4];			/* "FACP" or whatever */
	u8int	length[4];
	u8int	rev;
	u8int	csum;
	u8int	oemid[6];
	u8int	oemtblid[8];
	u8int	oemrev[4];
	u8int	creatorid[4];
	u8int	creatorrev[4];
};

/* Firmware control structure
 */
struct Facs
{
	u32int	hwsig;
	u32int	wakingv;
	u32int	glock;
	u32int	flags;
	u64int	xwakingv;
	u8int	vers;
	u32int	ospmflags;
};

static void acpifadt(uchar*, int);
static void acpimadt(uchar*, int);
static void acpitable(uchar*, int);

#define l16get(p)	(((p)[1]<<8)|(p)[0])
#define l32get(p)	(((u32int)l16get(p+2)<<16)|l16get(p))

#pragma varargck	type	"I"	uint

#define dprint		if(debug)print
#define BIOSSEG(a)	KADDR(((uint)(a))<<4)

/* tables listed here are loaded from the XSDT */
static Parse ptables[] =
{
	"FACP", acpifadt,
	"APIC",	acpimadt,
	"SSDT", acpitable,
	"SRAT",	acpitable,
	"MSCT",	acpitable,
};

static char *arnames[] = {
	"irq",	"dma",	"io",	"mem",	"memas",
	"ioas",	"busas",	"pir",	"bbn"
};

static Facs	*facs;		/* Firmware ACPI control structure */
static Fadt	fadt;		/* Fixed ACPI description. ACPI registers */
static Madt	*apics;		/* Fixed MADT entries */
static Atable	*tfirst;	/* loaded DSDT/SSDT/... tables */
static Atable	*tlast;		/* pointer to last table loaded */
static Gpe	*gpes;		/* General purpose events */
static int	ngpes;
static Pwr	slptyp[Npwr];	/* write to pm1.slptyp to enter Sn */
static int	debug = 1;

int legacydevices;
int keyboardpresent;
int novgapresent;

static int
Gfmt(Fmt* f)
{
	Gas *g;
	ulong x;

	g = va_arg(f->args, Gas*);
	switch(g->spc){
	case Rsysmem:
	case Rsysio:
	case Rembed:
	case Rsmbus:
	case Rcmos:
	case Rpcibar:
	case Ripmi:
		fmtprint(f, "[%s ", acpiregstr(g->spc));
		break;
	case Rpcicfg:
		x = g->addr >> 16;
		fmtprint(f, "[pci ");
		fmtprint(f, "dev %#ulx ", ACPIpcidev(x));
		fmtprint(f, "fn %#ulx ", ACPIpcifn(x));
		x = g->addr & 0xFFFF;
		fmtprint(f, "adr %#ulx ", x);
		break;
	case Rfixedhw:
		fmtprint(f, "[hw ");
		break;
	default:
		fmtprint(f, "[spc=%#ux ", g->spc);
	}
	return fmtprint(f, "off %d len %d addr %#ullx sz %d]",
		g->off, g->len, g->addr, g->accsz);
}

/*
 * interrupt flags
 */
static int
Ifmt(Fmt *f)
{
	int r;
	char flg[] = "--";

	r = va_arg(f->args, int);
	if(r == -1)
		return fmtstrcpy(f, "none");
	if((r&PcmpELMASK) == PcmpEDGE)
		flg[0] = 'e';
	if((r&PcmpELMASK) == PcmpLEVEL)
		flg[0] = 'l';
	if((r&PcmpPOMASK) == PcmpHIGH)
		flg[1] = 'h';
	if((r&PcmpPOMASK) == PcmpLOW)
		flg[1] = 'l';
	return fmtstrcpy(f, flg);
}

static int
Rfmt(Fmt *f)
{
	ACPIres *r;
	int i;

	r = va_arg(f->args, ACPIres*);
	switch(r->type){
	case Rirq:
		fmtprint(f, "irq flg %#ux %I", r->irq.flags, r->irq.flags);
		for(i = 0; i < nelem(r->irq.nb); i++)
			if(r->irq.nb[i] == ~0)
				break;
			else
				fmtprint(f, " %d", r->irq.nb[i]);
		break;
	case Rdma:
		fmtprint(f, "dma chans %#ux flags %#ux",
			r->dma.chans, r->dma.flags);
		break;
	case Rio:
	case Rmem:
		fmtprint(f, "%s %s min %#ullx max %ullx align %ud",
			arnames[r->type], r->mem.isrw ? "rw" : "ro",
			r->mem.min, r->mem.max, r->mem.align);
		break;
	case Rmemas:
	case Rioas:
	case Rbusas:
		fmtprint(f, "%s flags %#ux [%#ullx-%ullx]"
			" %#ullx %+#ld attr %#ullx",
			arnames[r->type], r->as.flags,
			r->as.min, r->as.max, r->as.len,
			r->as.off, r->as.attr);
		break;
	case Rpir:
		fmtprint(f, "pir dev %d pin %d link %d aint %d",
			r->pir.dno, r->pir.pin, r->pir.link,
			r->pir.airq);
		break;
	case Rbbn:
		fmtprint(f, "bbn %d", r->bbn);
		break;
	default:
		fmtprint(f, "unknown res %#ux", r->type);
	}
	return 0;
}

static char*
seprinttable(char *s, char *e, Atable *t)
{
	uchar *p;
	int i;

	p = t->p;
	s = seprint(s, e, "%s @ %#p\n", t->sig, p);
	for(i = 0; i < t->len; i++){
		if((i % 16) == 0)
			s = seprint(s, e, "%x: ", i);
		s = seprint(s, e, " %2.2ux", p[i]);
		if((i % 16) == 15)
			s = seprint(s, e, "\n");
	}
	return seprint(s, e, "\n\n");
}

char*
smprinttables(void)
{
	char *ttext;
	char *ns, *s, *e, *ntext;
	Atable *t;
	int tlen;

	tlen = 16*1024;
	ttext = malloc(tlen);
	if(ttext == nil){
		print("acpi: no memory\n");
		return 0;
	}
	s = ttext;
	e = ttext + tlen;
	strcpy(s, "no tables\n");
	for(t = tfirst; t != nil; t = t->next){
		ns = seprinttable(s, e, t);
		while(ns >= e - 1){
			dprint("tables: grow %#p %d\n", ttext, tlen*2);
			ntext = realloc(ttext, tlen*2);
			if(ntext == nil)
				panic("acpi: no memory\n");
			s = ntext + (s - ttext);
			ttext = ntext;
			tlen *= 2;
			e = ttext + tlen;
			ns = seprinttable(s, e, t);
		}
		s = ns;
	}
	return ttext;
}

static u64int
l64get(u8int* p)
{
	/*
	 * Doing this as a define
	 * #define l64get(p)	(((u64int)l32get(p+4)<<32)|l32get(p))
	 * causes 8c to abort with "out of fixed registers".
	 */
	return (((u64int)l32get(p+4)<<32)|l32get(p));
}

static Atable*
newtable(uchar *p)
{
	Atable *t;
	Sdthdr *h;

	t = malloc(sizeof(Atable));
	if(t == nil)
		panic("no memory for aml tables");
	t->p = nil;
	h = (Sdthdr*)p;
	t->is64 = h->rev >= 2;
	t->len = l32get(h->length);
	memmove(t->sig, h->sig, sizeof(h->sig));
	t->sig[sizeof(t->sig)-1] = 0;
	memmove(t->oemid, h->oemid, sizeof(h->oemid));
	t->oemtblid[sizeof(t->oemtblid)-1] = 0;
	memmove(t->oemtblid, h->oemtblid, sizeof(h->oemtblid));
	t->oemtblid[sizeof(t->oemtblid)-1] = 0;
	t->next = nil;
	if(tfirst == nil)
		tfirst = tlast = t;
	else{
		tlast->next = t;
		tlast = t;
	}
	return t;
}

static void*
sdtchecksum(void* addr, int len)
{
	u8int *p, sum;

	sum = 0;
	for(p = addr; len-- > 0; p++)
		sum += *p;
	if(sum == 0)
		return addr;

	return nil;
}

static void*
sdtmap(uintptr pa, int *n, int cksum)
{
	Sdthdr* sdt;

	sdt = vmap(pa, sizeof(Sdthdr));
	if(sdt == nil)
		return nil;
	*n = l32get(sdt->length);
	vunmap(sdt, sizeof(Sdthdr));
	if((sdt = vmap(pa, *n)) == nil)
		return nil;
	if(cksum != 0 && sdtchecksum(sdt, *n) == nil){
		dprint("acpi: SDT: bad checksum\n");
		vunmap(sdt, sizeof(Sdthdr));
		return nil;
	}
	return sdt;
}

static int
loadfacs(uintptr pa)
{
	int n;

	facs = sdtmap(pa, &n, 0);
	if(facs == nil)
		return -1;
	if(memcmp(facs, "FACS", 4) != 0){
		vunmap(facs, n);
		facs = nil;
		return -1;
	}
	/* don't vunmap */

	if(debug){
		print("acpi: facs\n");
		print("\thwsig: %#ux\n", facs->hwsig);
		print("\twakingv: %#ux\n", facs->wakingv);
		print("\tflags: %#ux\n", facs->flags);
		print("\tglock: %#ux\n", facs->glock);
		print("\txwakingv: %#llux\n", facs->xwakingv);
		print("\tvers: %#ux\n", facs->vers);
		print("\tospmflags: %#ux\n", facs->ospmflags);
		print("\n");
	}
	return 0;
}

static void
loaddsdt(uintptr pa)
{
	int n;
	uchar *dsdtp;

	dsdtp = sdtmap(pa, &n, 1);
	if(dsdtp == nil)
		return;
	acpitable(dsdtp, n);
	vunmap(dsdtp, n);
}

static void
gasget(Gas *gas, uchar *p)
{
	gas->spc = p[0];
	gas->len = p[1];
	gas->off = p[2];
	gas->accsz = p[3];
	gas->addr = l64get(p+4);
}

static void
dumpfadt(Fadt *fp)
{
	print("acpi fadt:\n");
	print("\tfacs: %#ux\n", fp->facs);
	print("\tdsdt: %#ux\n", fp->dsdt);
	print("\tpmprofile: %#ux\n", fp->pmprofile);
	print("\tsciint: %#ux\n", fp->sciint);
	print("\tsmicmd: %#ux\n", fp->smicmd);
	print("\tacpienable: %#ux\n", fp->acpienable);
	print("\tacpidisable: %#ux\n", fp->acpidisable);
	print("\ts4biosreq: %#ux\n", fp->s4biosreq);
	print("\tpstatecnt: %#ux\n", fp->pstatecnt);
	print("\tpm1aevtblk: %#ux\n", fp->pm1aevtblk);
	print("\tpm1bevtblk: %#ux\n", fp->pm1bevtblk);
	print("\tpm1acntblk: %#ux\n", fp->pm1acntblk);
	print("\tpm1bcntblk: %#ux\n", fp->pm1bcntblk);
	print("\tpm2cntblk: %#ux\n", fp->pm2cntblk);
	print("\tpmtmrblk: %#ux\n", fp->pmtmrblk);
	print("\tgpe0blk: %#ux\n", fp->gpe0blk);
	print("\tgpe1blk: %#ux\n", fp->gpe1blk);
	print("\tpm1evtlen: %#ux\n", fp->pm1evtlen);
	print("\tpm1cntlen: %#ux\n", fp->pm1cntlen);
	print("\tpm2cntlen: %#ux\n", fp->pm2cntlen);
	print("\tpmtmrlen: %#ux\n", fp->pmtmrlen);
	print("\tgpe0blklen: %#ux\n", fp->gpe0blklen);
	print("\tgpe1blklen: %#ux\n", fp->gpe1blklen);
	print("\tgp1base: %#ux\n", fp->gp1base);
	print("\tcstcnt: %#ux\n", fp->cstcnt);
	print("\tplvl2lat: %#ux\n", fp->plvl2lat);
	print("\tplvl3lat: %#ux\n", fp->plvl3lat);
	print("\tflushsz: %#ux\n", fp->flushsz);
	print("\tflushstride: %#ux\n", fp->flushstride);
	print("\tdutyoff: %#ux\n", fp->dutyoff);
	print("\tdutywidth: %#ux\n", fp->dutywidth);
	print("\tdayalrm: %#ux\n", fp->dayalrm);
	print("\tmonalrm: %#ux\n", fp->monalrm);
	print("\tcentury: %#ux\n", fp->century);
	print("\tiapcbootarch: %#ux\n", fp->iapcbootarch);
	print("\tflags: %#ux\n", fp->flags);
	print("\tresetreg: %G\n", &fp->resetreg);
	print("\tresetval: %#ux\n", fp->resetval);
	print("\txfacs: %#llux\n", fp->xfacs);
	print("\txdsdt: %#llux\n", fp->xdsdt);
	print("\txpm1aevtblk: %G\n", &fp->xpm1aevtblk);
	print("\txpm1bevtblk: %G\n", &fp->xpm1bevtblk);
	print("\txpm1acntblk: %G\n", &fp->xpm1acntblk);
	print("\txpm1bcntblk: %G\n", &fp->xpm1bcntblk);
	print("\txpm2cntblk: %G\n", &fp->xpm2cntblk);
	print("\txpmtmrblk: %G\n", &fp->xpmtmrblk);
	print("\txgpe0blk: %G\n", &fp->xgpe0blk);
	print("\txgpe1blk: %G\n", &fp->xgpe1blk);
	print("\n");
}

static void
acpifadt(uchar *p, int)
{
	Fadt *fp;

	fp = &fadt;
	fp->facs = l32get(p + 36);
	fp->dsdt = l32get(p + 40);
	fp->pmprofile = p[45];
	fp->sciint = l16get(p+46);
	fp->smicmd = l32get(p+48);
	fp->acpienable = p[52];
	fp->acpidisable = p[53];
	fp->s4biosreq = p[54];
	fp->pstatecnt = p[55];
	fp->pm1aevtblk = l32get(p+56);
	fp->pm1bevtblk = l32get(p+60);
	fp->pm1acntblk = l32get(p+64);
	fp->pm1bcntblk = l32get(p+68);
	fp->pm2cntblk = l32get(p+72);
	fp->pmtmrblk = l32get(p+76);
	fp->gpe0blk = l32get(p+80);
	fp->gpe1blk = l32get(p+84);
	fp->pm1evtlen = p[88];
	fp->pm1cntlen = p[89];
	fp->pm2cntlen = p[90];
	fp->pmtmrlen = p[91];
	fp->gpe0blklen = p[92];
	fp->gpe1blklen = p[93];
	fp->gp1base = p[94];
	fp->cstcnt = p[95];
	fp->plvl2lat = l16get(p+96);
	fp->plvl3lat = l16get(p+98);
	fp->flushsz = l16get(p+100);
	fp->flushstride = l16get(p+102);
	fp->dutyoff = p[104];
	fp->dutywidth = p[105];
	fp->dayalrm = p[106];
	fp->monalrm = p[107];
	fp->century = p[108];
	fp->iapcbootarch = l16get(p+109);
	fp->flags = l32get(p+112);
	gasget(&fp->resetreg, p+116);
	fp->resetval = p[128];
	fp->xfacs = l64get(p+132);
	fp->xdsdt = l64get(p+140);
	gasget(&fp->xpm1aevtblk, p+148);
	gasget(&fp->xpm1bevtblk, p+160);
	gasget(&fp->xpm1acntblk, p+172);
	gasget(&fp->xpm1bcntblk, p+184);
	gasget(&fp->xpm2cntblk, p+196);
	gasget(&fp->xpmtmrblk, p+208);
	gasget(&fp->xgpe0blk, p+220);
	gasget(&fp->xgpe1blk, p+232);

	legacydevices = fp->iapcbootarch & BAlegacy;
	keyboardpresent = fp->iapcbootarch & BAkbd;
	novgapresent = fp->iapcbootarch & BAnovga;

	if(debug)
		dumpfadt(fp);
	if(fp->xfacs != 0)
		loadfacs(fp->xfacs);
	else
		loadfacs(fp->facs);
	if(fp->xdsdt != 0)
		loaddsdt(fp->xdsdt);
	else
		loaddsdt(fp->dsdt);
}

static void
dumpmadt(Madt *apics)
{
	Apicst *st;

	print("acpi: madt lapic addr %llux pcat %d:\n",
		apics->laddr, apics->pcat);
	for(st = apics->st; st != nil; st = st->next)
		switch(st->type){
		case ASlapic:
			print("\tlapic pid %d id %d\n",
				st->lapic.pid, st->lapic.id);
			break;
		case ASioapic:
		case ASiosapic:
			print("\tioapic id %d addr %#llux ibase %d\n",
				st->ioapic.id, st->ioapic.addr,
				st->ioapic.ibase);
			break;
		case ASintovr:
			print("\tintovr irq %d intr %d %##ux\n",
				st->intovr.irq, st->intovr.intr,
				st->intovr.flags);
			break;
		case ASnmi:
			print("\tnmi intr %d %##ux\n", st->nmi.intr,
				st->nmi.flags);
			break;
		case ASlnmi:
			print("\tlnmi pid %d lint %d %#ux\n",
				st->lnmi.pid, st->lnmi.lint,
				st->lnmi.flags);
			break;
		case ASlsapic:
			print("\tlsapic pid %d id %d eid %d"
				" puid %d puids %s\n",
				st->lsapic.pid, st->lsapic.id,
				st->lsapic.eid, st->lsapic.puid,
				st->lsapic.puids);
			break;
		case ASintsrc:
			print("\tintr type %d pid %d peid %d"
				" iosv %d intr %d %#ux\n",
				st->type, st->intsrc.pid,
				st->intsrc.peid, st->intsrc.iosv,
				st->intsrc.intr, st->intsrc.flags);
			break;
		case ASlx2apic:
			print("\tlx2apic puid %d id %d\n",
				st->lx2apic.puid, st->lx2apic.id);
			break;
		case ASlx2nmi:
			print("\tlx2nmi puid %d intr %d %##ux\n",
				st->lx2nmi.puid, st->lx2nmi.intr,
				st->lx2nmi.flags);
			break;
		default:
			print("\t<unknown type %#ux>\n", st->type);
		}
	print("\n");
}

enum{ FakeISA = 16 };

static void
mknmi(int apicno, int intin, int flags)
{
	PCMPintr intr;

	if(flags < 0)
		return;
	memset(&intr, 0, sizeof(intr));
	intr.busno = FakeISA;		/* our fake isa bus */
	intr.intr = PcmpNMI;
	intr.flags = flags;
	intr.irq = 2;
	intr.apicno = apicno;
	intr.intin = intin;
	dprint("acpi lnmi apic %d in %d flags %#ux\n",
		apicno, intin, flags);
	mpnewlintr(&intr);
}

static void
acpilapic(Apicst *a)
{
	Apicst *l;
	Apic *apic;
	uint lint[2];
	int flags;
	static int mpmachno;
	static void *va;

	lint[0] = lint[1] = -1;

	flags = PcmpEN;
	if(mpmachno == 0){
		flags |= PcmpBP;
		va = vmap(apics->laddr, 1024);
	}
	if(va == nil){
		print("can't map lapic\n");
		return;
	}
	apic = mpnewapic(PcmpPROCESSOR, a->lapic.id, flags);
	if(apic == nil){
		print("lapic already defined\n");
		return;
	}
	apic->machno = mpmachno++;
	apic->paddr = apics->laddr;
	apic->addr = va;
	dprint("lapic %d at %#p l0\n", apic->apicno, apic->paddr);
	if(mpmachno == 1){
		mpnewbus(BusPCI, 0);		/* fake pci bus */
		mpnewbus(BusISA, FakeISA);	/* fake isa bus */
	}
	for(l = apics->st; l != nil; l = l->next)
		if(l->type == ASlnmi && l->lnmi.pid == a->lapic.pid)
		if(l->lnmi.lint != 0 && l->lnmi.lint != 1)
			print("acpilapic: bad lint %d\n", l->lnmi.lint);
		else
			mknmi(apic->apicno, l->lnmi.lint, l->lnmi.flags);
}

/*
 * Must start the ioapic to learn its mre.
 * Needed to locate ioapics by acpi irq numbers.
 */
static void
acpiioapic(Apicst *a)
{
	Apic *apic;
	void *va;

	apic = mpnewapic(PcmpIOAPIC, a->ioapic.id, PcmpEN);
	if(apic == nil){
		print("ioapic already defined");
		return;
	}
	apic->paddr = a->ioapic.addr;
	apic->ibase = a->ioapic.ibase;
	if((va = vmap(apic->paddr, 1024)) == nil)
		apic->flags &= ~PcmpEN;
	else{
		apic->addr = va;
		ioapicinit(apic, apic->apicno);
	}
	dprint("ioapic %d at %#p ibase %d\n",
		apic->apicno, apic->paddr, apic->ibase);
}

static Apic*
getioapic(int airq)
{
	int ibase;
	Apic *apic;

	for(apic = mpioapics; apic != nil; apic = apic->next){
		ibase = apic->ibase;
		if(airq >= ibase && airq <= ibase+apic->mre)
			return apic;
	}
	return nil;
}

/*
 * add an isa interrupt entry.
 */
static void
acpiint(int irq, int airq, int flags, int nmi)
{
	PCMPintr intr;
	Apic *a;

	a = getioapic(airq);
	if(a == nil){
		print("acpi: no ioapic for airq %d\n", airq);
		return;
	}
	if((flags&PcmpPOMASK) == 0)
		flags |= PcmpHIGH;
	if((flags&PcmpELMASK) == 0)
		flags |= PcmpEDGE;
	memset(&intr, 0, sizeof(intr));
	intr.busno = FakeISA;		/* our fake isa bus number */
	intr.flags = flags;
	intr.intr = PcmpINT;
	if(nmi != 0)
		intr.intr = PcmpNMI;
	intr.irq = irq;
	intr.apicno = a->apicno;
	intr.intin = airq - a->ibase;
	dprint("acpi int: isa irq %d apic %d in %d flags %I\n",
		intr.irq, intr.apicno, intr.intin, intr.flags);
	mpnewintr(&intr);
}

static void
acpiapics(void)
{
	Apicst *l;
	int i, irq, isairqs;

	if(apics == nil)
		return;

	/*
	 * Configure lapics. The first one is always the boot processor.
	 */
	for(l = apics->st; l != nil; l = l->next)
		switch(l->type){
		case ASlapic:
			acpilapic(l);
			break;
		case ASioapic:
			acpiioapic(l);
			break;
		}
	/*
	 * Add interrupt info now that we have configured all apics.
	 * This means nmis and ISA irqs which, unless overriden,
	 * are assumed to be identity mapped by ACPI.
	 */
	isairqs = 0;
	for(l = apics->st; l != nil; l = l->next)
		switch(l->type){
		case ASintovr:
			irq = l->intovr.irq;
			if(irq >= 0 && irq < 16)
				isairqs |= 1<<irq;
			acpiint(l->intovr.irq, l->intovr.intr,
				l->intovr.flags, 0);
			break;
		case ASnmi:
			acpiint(2, l->intovr.intr,
				l->intovr.flags, 1);
			break;
		}
	for(i = 0; i < 16; i++)
		if((isairqs & 1<<i) == 0)
			acpiint(i, i, PcmpHIGH|PcmpEDGE, 0);

}

static void
acpimadt(uchar *p, int len)
{
	uchar *pe;
	Apicst *st, *l, **stl;
	int stlen, id;

	if(len < 44)
		return;
	apics = mallocz(sizeof(Madt), 1);
	if(apics == nil)
		panic("acpimadt: no memory");
	apics->laddr = l32get(p+36);
	apics->pcat = l32get(p+40);
	apics->st = nil;
	stl = &apics->st;
	pe = p + len;
	for(p += 44; p < pe; p += stlen){
		st = mallocz(sizeof(Apicst), 1);
		st->type = p[0];
		st->next = nil;
		stlen = p[1];
		switch(st->type){
		case ASlapic:
			st->lapic.pid = p[2];
			st->lapic.id = p[3];
			if(l32get(p+4) == 0){
				free(st);
				st = nil;
			}
			break;
		case ASioapic:
			st->ioapic.id = p[2];
			st->ioapic.addr = l32get(p+4);
			st->ioapic.ibase = l32get(p+8);
			break;
		case ASintovr:
			st->intovr.irq = p[3];
			st->intovr.intr = l32get(p+4);
			st->intovr.flags = l16get(p+8);
			break;
		case ASnmi:
			st->nmi.flags = l16get(p+2);
			st->nmi.intr = l32get(p+4);
			break;
		case ASlnmi:
			st->lnmi.pid = p[2];
			st->lnmi.flags = l16get(p+3);
			st->lnmi.lint = p[5];
			break;
		case ASladdr:
			apics->laddr = l64get(p+8);
			break;
		case ASiosapic:
			id = st->iosapic.id = p[2];
			st->iosapic.ibase = l32get(p+4);
			st->iosapic.addr = l64get(p+8);
			for(l = apics->st; l != nil; l = l->next)
				if(l->type == ASioapic && l->ioapic.id == id){
					l->ioapic = st->iosapic;
					free(st);
					st = nil;
					break;
				}
			break;
		case ASlsapic:
			st->lsapic.pid = p[2];
			st->lsapic.id = p[3];
			st->lsapic.eid = p[4];
			st->lsapic.puid = l32get(p+12);
			if(l32get(p+8) == 0){
				free(st);
				st = nil;
			}else
				kstrdup(&st->lsapic.puids, (char*)p+16);
			break;
		case ASintsrc:
			st->intsrc.flags = l16get(p+2);
			st->type = p[4];
			st->intsrc.pid = p[5];
			st->intsrc.peid = p[6];
			st->intsrc.iosv = p[7];
			st->intsrc.intr = l32get(p+8);
			st->intsrc.any = l32get(p+12);
			break;
		case ASlx2apic:
			st->lx2apic.id = l32get(p+4);
			st->lx2apic.puid = l32get(p+12);
			if(l32get(p+8) == 0){
				free(st);
				st = nil;
			}
			break;
		case ASlx2nmi:
			st->lx2nmi.flags = l16get(p+2);
			st->lx2nmi.puid = l32get(p+4);
			st->lx2nmi.intr = p[8];
			break;
		default:
			print("acpi: unknown madt entry\n");
			free(st);
			st = nil;
		}
		if(st != nil){
			*stl = st;
			stl = &st->next;
		}
	}
	dumpmadt(apics);
	acpiapics();
}

/*
 * Keep a copy of a table, for p9am(3).
 */
static void
acpitable(uchar *p, int len)
{
	Atable *t;

	if(len < Sdthdrsz){
		print("acpi: short table: %d bytes\n", len);
		return;
	}
	t = newtable(p);
	t->p = malloc(len);
	if(t->p == nil)
		panic("acpitable: no memory");
	memmove(t->p, p, len);
}

/*
 * Process the xsdt table and load tables found there.
 */
static int
acpixsdtload(u8int* p, int len, int asize)
{
	int i, l, t, found;
	uintptr dhpa;
	uchar *sdt;
	char tsig[5];

	found = 0;
	for(i = 0; i < len; i += asize){
		if(asize == 8)
			dhpa = l64get(p+i);
		else
			dhpa = l32get(p+i);
		if((sdt = sdtmap(dhpa, &l, 1)) == nil)
			continue;
		memmove(tsig, sdt, 4);
		tsig[4] = 0;
		dprint("acpi: %s addr %#p paddr %#p\n", tsig, sdt, dhpa);
		for(t = 0; t < nelem(ptables); t++)
			if(strcmp(tsig, ptables[t].sig) == 0){
				ptables[t].f(sdt, l);
				break;
			}
		vunmap(sdt, l);
	}
	return found;
}

static void*
rsdscan(u8int* addr, int len, char* signature)
{
	int sl;
	u8int *e, *p;

	e = addr+len;
	sl = strlen(signature);
	for(p = addr; p+sl < e; p += 16){
		if(memcmp(p, signature, sl))
			continue;
		return p;
	}

	return nil;
}

static void*
rsdsearch(char* signature)
{
	uintptr p;
	u8int *bda;
	void *rsd;

	/*
	 * Search for the data structure signature:
	 * 1) in the first KB of the EBDA;
	 * 2) in the BIOS ROM between 0xE0000 and 0xFFFFF.
	 */
	if(strncmp((char*)KADDR(0xFFFD9), "EISA", 4) == 0){
		bda = BIOSSEG(0x40);
		if((p = (bda[0x0F]<<8)|bda[0x0E])){
			if(rsd = rsdscan(KADDR(p), 1024, signature))
				return rsd;
		}
	}
	return rsdscan(BIOSSEG(0xE000), 0x20000, signature);
}

static uint
bankedin(uintptr ra, uintptr rb, int sz)
{
	uint r;

	r = 0;
	switch(sz){
	case 1:
		if(ra != 0)
			r |= inb(ra);
		if(rb != 0)
			r |= inb(rb);
		break;
	case 2:
		if(ra != 0)
			r |= ins(ra);
		if(rb != 0)
			r |= ins(rb);
		break;
	case 4:
		if(ra != 0)
			r |= inl(ra);
		if(rb != 0)
			r |= inl(rb);
		break;
	default:
		print("bankedin: wrong size\n");
	}
	return r;
}

static uint
bankedout(uintptr ra, uintptr rb, int sz, int v)
{
	uint r;

	r = -1;
	switch(sz){
	case 1:
		if(ra != 0)
			outb(ra, v);
		if(rb != 0)
			outb(rb, v);
		break;
	case 2:
		if(ra != 0)
			outs(ra, v);
		if(rb != 0)
			outs(rb, v);
		break;
	case 4:
		if(ra != 0)
			outl(ra, v);
		if(rb != 0)
			outl(rb, v);
		break;
	default:
		print("bankedout: wrong size\n");
	}
	return r;
}

static uint
getpm1ctl(void)
{
	return bankedin(fadt.pm1acntblk, fadt.pm1bcntblk, fadt.pm1cntlen);
}

static void
setpm1sts(uint v)
{
	bankedout(fadt.pm1aevtblk, fadt.pm1bevtblk, fadt.pm1evtlen/2, v);
}

static uint
getpm1sts(void)
{
	return bankedin(fadt.pm1aevtblk, fadt.pm1bevtblk, fadt.pm1evtlen/2);
}

static uint
getpm1en(void)
{
	int sz;

	sz = fadt.pm1evtlen/2;
	return bankedin(fadt.pm1aevtblk+sz, fadt.pm1bevtblk+sz, sz);
}

static int
getgpeen(int n)
{
	return inb(gpes[n].enio) & 1<<gpes[n].enbit;
}

static void
setgpeen(int n, uint v)
{
	int old;

	old = inb(gpes[n].enio);
	if(v)
		outb(gpes[n].enio, old | 1<<gpes[n].enbit);
	else
		outb(gpes[n].enio, old & ~(1<<gpes[n].enbit));
}

static void
clrgpests(int n)
{
	outb(gpes[n].stsio, 1<<gpes[n].stsbit);
}

static uint
getgpests(int n)
{
	return inb(gpes[n].stsio) & 1<<gpes[n].stsbit;
}

void
acpigpe(int i, char *obj)
{
	if(i >= ngpes)
		error("gpe out of range");
	kstrdup(&gpes[i].obj, obj);
	dprint("acpi: gpe %d %s\n", i, gpes[i].obj);
	setgpeen(i, 1);
}

void
acpipwr(int i, int pm1a, int pm1b)
{
	if(i >= Npwr)
		error("bad power state number");
	slptyp[i].pm1a = pm1a;
	slptyp[i].pm1b = pm1b;
	dprint("acpi: pwr %d pm1a %#ux pm1b %#ux\n",
		i, slptyp[i].pm1a, slptyp[i].pm1b);
}

/*
 * ACPI interrupt routine. Something happen to the fixed
 * hardware or we must notify an event to 9am.
 */
static void
acpiintr(Ureg*, void*)
{
	int i;
	uint sts, en;

//	print("acpi: intr\n");

	for(i = 0; i < ngpes; i++)
		if(getgpests(i)){
			print("gpe %d on\n", i);
 			en = getgpeen(i);
			setgpeen(i, 0);
			clrgpests(i);
			if(en != 0)
				print("acpiitr: calling gpe %d\n", i);
		//	queue gpe for calling gpe->ho in the
		//	aml process.
		//	enable it again when it returns.
		}
	sts = getpm1sts();
	en = getpm1en();
//	print("acpiitr: pm1sts %#ux pm1en %#ux\n", sts, en);
//	if(sts&en)
//		print("have enabled events\n");
//	if(sts&1)
//		print("power button\n");
	// XXX serve other interrupts here.
	setpm1sts(sts);	
}

static void
initgpes(void)
{
	int i, n0, n1;

	n0 = fadt.gpe0blklen/2;
	n1 = fadt.gpe1blklen/2;
	ngpes = n0 + n1;
	gpes = mallocz(sizeof(Gpe) * ngpes, 1);
	for(i = 0; i < n0; i++){
		gpes[i].nb = i;
		gpes[i].stsbit = i&7;
		gpes[i].stsio = fadt.gpe0blk + i>>3;
		gpes[i].enbit = (n0 + i)&7;
		gpes[i].enio = fadt.gpe0blk + ((n0 + i)>>3);
	}
	for(i = 0; i + n0 < ngpes; i++){
		gpes[i + n0].nb = fadt.gp1base + i;
		gpes[i + n0].stsbit = i&7;
		gpes[i + n0].stsio = fadt.gpe1blk + i>>3;
		gpes[i + n0].enbit = (n1 + i)&7;
		gpes[i + n0].enio = fadt.gpe1blk + ((n1 + i)>>3);
	}
	for(i = 0; i < ngpes; i++){
		setgpeen(i, 0);
		clrgpests(i);
	}
}

static void
acpiioalloc(uint addr, int len)
{
	if(addr != 0)
		ioalloc(addr, len, 0, "acpi");
}

static void
acpirsdptr(void)
{
	Rsdp *rsd;
	int asize;
	uintptr sdtpa;
	u8int *p;
	int len;

	if((rsd = rsdsearch("RSD PTR ")) == nil){
		dprint("no rsd ptr\n");
		return;
	}
	assert(sizeof(Sdthdr) == Sdthdrsz);

	dprint("acpi: RSD PTR @ %#p, physaddr %#ux length %ud %#llux rev %d\n",
		rsd, l32get(rsd->raddr), l32get(rsd->length),
		l64get(rsd->xaddr), rsd->revision);

	if(rsd->revision >= 2){
		if(sdtchecksum(rsd, 36) == nil){
			dprint("acpi: RSD: bad checksum\n");
			return;
		}
		sdtpa = l64get(rsd->xaddr);
		asize = 8;
	}
	else{
		if(sdtchecksum(rsd, 20) == nil){
			dprint("acpi: RSD: bad checksum\n");
			return;
		}
		sdtpa = l32get(rsd->raddr);
		asize = 4;
	}

	/*
	 * process the RSDT or XSDT table.
	 */
	if((p = sdtmap(sdtpa, &len, 1)) == nil)
		return;
	if((p[0] == 'R' || p[0] == 'X') && memcmp(p+1, "SDT", 3) == 0){
		dprint("acpi: %cSDT %#p\n", p[0], p);
		acpixsdtload(p + Sdthdrsz, len - Sdthdrsz, asize);
	}
	vunmap(p, len);
}

void
acpistart(void)
{
	if(0)
	if(fadt.sciint != 0)
		intrenable(fadt.sciint, acpiintr, 0, BUSUNKNOWN, "acpi");
}

void
acpiinit(void)
{
	int i;
	char *p;
	static int initialized;

	if((p = getconf("*noacpi")) != nil && strtol(p, 0, 0) != 0)
		return;
	if(initialized++ > 0)
		return;

	fmtinstall('G', Gfmt);
	fmtinstall('R', Rfmt);
	fmtinstall('I', Ifmt);

	acpirsdptr();
	if(fadt.smicmd == 0)
		return;

	/* stop if we want just acpi tables; but no aml */
	if((p = getconf("*9am")) == nil || strtol(p, 0, 0) == 0)
		return;

	/*
	 * should use fadt->xpm* and fadt->xgpe* registers for 64 bits.
	 */
	dprint("acpi io alloc\n");
	acpiioalloc(fadt.smicmd, 1);
	acpiioalloc(fadt.pm1aevtblk, fadt.pm1evtlen);
	acpiioalloc(fadt.pm1bevtblk, fadt.pm1evtlen );
	acpiioalloc(fadt.pm1acntblk, fadt.pm1cntlen);
	acpiioalloc(fadt.pm1bcntblk, fadt.pm1cntlen);
	acpiioalloc(fadt.pm2cntblk, fadt.pm2cntlen);
	acpiioalloc(fadt.pmtmrblk, fadt.pmtmrlen);
	acpiioalloc(fadt.gpe0blk, fadt.gpe0blklen);
	acpiioalloc(fadt.gpe1blk, fadt.gpe1blklen);

	dprint("acpi init gpes\n");
	initgpes();

	/*
	 * This starts ACPI, which may require we handle
	 * events ourselves. Use with care.
	 */
	dprint("enabling acpi...");
	if((getpm1ctl() & Pm1SciEn) != 0){
		dprint("always enabled\n");
		return;
	}
	outb(fadt.smicmd, fadt.acpienable);
	for(i = 0; i < 10; i++)
		if(getpm1ctl() & Pm1SciEn)
			break;
	if(i == 10){
		dprint("failed to enable\n");
		outb(fadt.smicmd, fadt.acpidisable);
		fadt.smicmd = 0;
	}else
		dprint("\n");
}

static int
gasio(Gas *g, void *p, int len)
{
	Reg r;

	if(g->len == 0)
		return -1;
	r.name = "gas";
	r.spc = g->spc;
	if(r.spc != Rsysmem && r.spc != Rsysio){
		print("bug: gasio:space %#ux not implemented\n", r.spc);
		/* should at least support pci for reset */
		return -1;
	}
	r.len = g->len/8;
	r.base = g->addr;
	r.accsz = g->accsz/8;
	if(r.accsz == 0)
		r.accsz = 1;
	return acpiregio(&r, p, len, g->off, 1);
}

void
acpireset(void)
{
	gasio(&fadt.resetreg, &fadt.resetval, fadt.resetreg.len);
}

