.FP lucidasans
.HTML "Plan 9 AML Interpreter and ACPI support
.TL
The Plan 9 AML Interpreter and ACPI support
.AU
Francisco J. Ballesteros
nemo@lsub.org
.AB
This is a technical note regarding the implementation of the Plan 9
AML interpreter and support for ACPI.
It is intended as an aid for those who must change or
understand its implementation.
.AE
.SH
Introduction.
.PP
ACPI is an standard for configuration and power management on the PC.
In ACPI, the platform (motherboard and devices) provides tables that convey
configuration information to the OS.
.PP
Besides the static information found in tables, ACPI configuration includes
bytecode for a machine language known as AML. The OS is expected to execute
AML code both to initialize ACPI and to trigger actions on devices (e.g., to change
the power state). Also, some hardware interrupts may require that we should call AML
to handle the implied events.
.PP
AML is the machine language for an abstract language known as ASL. It turns out
that AML is more a representation of ASL than it is a language for a machine. In
understanding what the interpreter must do it helps to look at ASL definitions
for the operations considered. AML is close to them.
.PP
The language operates on objects that live in a tree-shaped namespace, global to
AML.
The AML ``machine'' does not have a general purpose stack. It operates
on the name space and has ``builtin objects'' that represent arguments and local
variables. Thus,
thanks to the AML specification, a name may represent a method object, or may not;
when it does, what
follows in the code may be arguments for it. There is no operation for calling
a routine. In general, it is not feasible to say if a bytecode
is an operation or raw data. 
.PP
The name space is initialized as a side-effect of loading
.CW DSDT
and
.CW SSDT
tables (and evaluating AML code on them).
.PP
Names are four characters, but
searching rules for names are not what could be expected. For example,
a relative name may be the resolved as the aunt of dot. As another example,
names
.CW _OS_
and
.CW _OS
are considered to be the same. The global
.CW acpins
keeps the root of the name space.
.PP
A name refers to an object (perhaps none), and besides that, may or may not
have children. So, names are both like files and directories, depending on the
point of view. This is an example (portion) of a name space:
.P1
	   /_SB_/PBTN= dev <nil> pnp 'PNP0C0C' addr 0x0 {
	      /_SB_/PBTN/_HID=  0xc0cd041
	      /_SB_/PBTN/_PRW= pkg[2] { 0x8,  0x4}
	      /_SB_/PBTN/_PSW= method _PSW/1
	   }
	   /_SB_/_INI= method _INI/0
.P2
.LP
Here,
.CW PBTN
is both a device and the root for a (sub)tree of ACPI objects referring to
.CW PBTN .
Syntax for names in ACPI is actually
.CW \e_SB_.PBTN
and not
.CW /_SB_/PBTN .
This implementation uses the more familiar UNIX syntax (with the required ACPI
semantics).
.PP
There are different object types including strings, buffers,
integers, references to other objects, buffer slices, register (bit) fields, memory
(or I/O) regions, etc.
An important object type is called a
.I method
in AML. It's not a ``method'', but a subroutine that may  be called.
In general, methods create objects in the name space and use them
very much like local variables and, therefore, are not reentrant.
.PP
Execution of AML operations is performed by using (and modifying) objects
in the name space.
There are control operations, arithmetic operations, and operations on objects
(roughly speaking). 
There is no operation to call a method. AML is not a language
for an abstract machine, but a compact representation of a source language.
Thus, evaluating AML is similar to interpreting source (ASL). For example, to
call a method, the name is placed into the bytecode and then the arguments.
Depending on the context, a name found in the bytecode stream may imply
a method call.
.PP
Most operations copy objects and may lead to either explicit or
implicit type casts, which must follow rules indicated on the spec. 
Beware that casts are in many cases not what could be expected. For
example, obtaining an integer from a string differs
from what
.I strtoul
would do. Buffers may be handled as integers or buffers depending on their
size. Converting a buffer to a string and writing it into a register does not
produce the same effect than writing the buffer. Etcetera.
.PP
To look for further information on ACPI, ASL, and AML it helps to consider:
.IP \(bu
.I
Advanced Configuration and Power Interface Specification. 4.0. June 2009.
.br
.R
This is the ACPI specification. Chapter 5 is a description of the various ACPI tables,
the namespace, and important objects. Chapter 18 is a description of ASL. Chapter
19 is a description of AML (or, rather, a description of how ASL is encoded into AML).
Section 18.2.5 lists the different objects and type conversion rules.
.IP \(bu
.I
OpenBSD ACPI implementation.
.br
.R
See
.CW /sys/dev/acpi
in the system kernel source. Many operating systems use an Intel provide implementation
for the interpreter. However, this one is easier to understand and can be used as
additional documentation for ACPI and AML.
.LP
The AML interpreter is currently work in progress, and can be used as a user program
to parse and execute AML code as dumped by a kernel during boot, using a temporary
.I acpi
device for such purpose.
.PP
When ready, the interpreter must execute at
.I links
time during boot, or anytime before configuring devices. 
By that time, there are no processes in the kernel.
.SH
Implementation
.LP
The important files are:
.IP \f(CWaml.h\fP
.br
It defines the data structures used by the interpreter.
.IP \f(CWaml.c\fP
.br
A driver for the interpreter. Most of it would go when in the kernel.
It contains the main program and code to translate the dumps made
by the kernel to binary data that can be interpreted as AML bytecodes.
.IP \f(CWamlrt.c\fP
.br
The run time for the interpreter. Includes helpers and several important
entry points to evaluate AML code.
.IP \f(CWamlop.c\fP
.br
Implementation for AML operations.
.IP \f(CWamlconv.c\fP
.br
Implementation for type casts and object copying. This part of the
standard is a tangled web and its code is kept appart.
.IP \f(CWkernel.c\fP
.br
Compatibility tools to keep the interpreter code as ready for the kernel as
feasible.
.IP \f(CWacpi.c\fP
.br
Contains tools that should be provided by the kernel ACPI driver. They are
not part of the interpreter but may be necessary for this program. There is
a version for the kernel in
.CW 9pc/acpi.c ,
intended to replace this file when the interpreter gets into the kernel.
.SH
Names and objects
.LP
The global
.CW acpins
keeps the root of the name space. This is the data type for a name:
.P1
struct Aname
{
	char	s[5];		/* NNNN\0 */
	char*	path;
	Aobj*	o;
	Aname*	parent;
	Aname*	childhd;	/* first child */
	Aname*	childtl;	/* last child */
	Aname*	sibling;	/* pointer to next sibling */
	Atable*	table;		/* [ds]sdt table id; for unload */
	Aname*	mnext;		/* in list of per-method binds */
};
.P2
.LP
The list of children starts at
.CW childhd .
The actual name is
.CW s ,
but the entire
.CW path
is included as a convenience. The object bound to this name, if any, is kept
at
.CW o .
.PP
The data type for an ACPI, or AML, object is this:
.P1
/*
 * ACPI Aobject.
 */
struct Aobj
{
	Ref;
	int	type;				/* for this object */
	Aname*	name;				/* ACPI name if bound or nil */
	Aobj*	next;				/* in list */
.P2
.P1
	union{
		u64int		ival;		/* doubles as handle */
		char*		sval;		/* strings and names */
		Nbuf		buf;		/* byte buffer */
		Nslice		slice;		/* buffer slice (bits) */
		Nreg		reg;		/* memory, i/o, ... region */
		Nfield		field;		/* slice (bits) of object */
		Nref		oref;		/* local, ref, arg */
		Nlist		list;		/* packages */
		Nscope		scope;		/* for parsing */
		Nmethod		method;
		Ndev		dev;
		Npwr		pwr;
		Nproc		proc;
		Nthermal	thermal;
		Nmutex		mutex;
		Nevent		event;
	};
};
.P2
.LP
Objects may be kept in a list of free objects (when free) or in a list of
objects contained in a
.I package
object (an ACPI array). They are reference counted (including as
references those from names, from other objects, from the
interpreter stack, etc.).
.CW Type
may be:
.P1
Onone	=  0,
Ofree,			/* debug */
Oslice,			/* buffer field or slice */
Oscope,			/* scope */
Omethod,		/* method */
Odev,			/* device */
Oproc,			/* processor */
Opwr,			/* pwrrsrc */
Othermal,		/* thermal zone */
Oref,			/* ref to object */
Olocal,			/* ref to local object */
Oarg,			/* ref to arg object */
Ohandle,		/* definition block handle */
Obuf	= 'B',		/* buf */
Oevent	= 'E',		/* sleep/wakeup */
Ofield	= 'F',,		/* register field */
Oint	= 'I',		/* integer */
Oname	= 'N',		/* a name */
Omutex	= 'M',		/* mutex */
Opkg	= 'P',		/* package object (array) */
Oreg	= 'R',		/* data region */
Ostr	= 'S',		/* string */
Odecstr	= 'D',		/* like string, but bytes from buffer */
			/* values are printed in decimal */
Oany	= 'O',		/* any object after resolving names */
.P2
.LP
.CW Onone
is uninitialized.
.CW Ofree
is used while the object is on the free list (for debug checks).
.CW Oslice
is called a
.I field
in ACPI, but there are three or four types of fields it is easier to understand
this object as a slice of a buffer. We use
.CW Ofield
for fields that identify bit-slices of other fields or system memory and I/O spaces.
.CW Opkg
is actually an array.
.PP
A slice value (buffer field in ACPI) is defined as
.P1
/*
 * buffer field. slice of bits in buffer.
 */
struct Nslice
{
	Aobj*	src;		/* source Obuf object */
	uintptr	off;		/* starting bit index */
	uintptr	len;		/* number of bits */
};
.P2
.LP
We keep offset and length in bits because
.CW Nslice
is used to represent several ACPI field types with bit/byte/... granularity.
A general purpose
.CW bitcpy
function in the interpreter is used to move bytes when feasible
and bits otherwise.
.PP
An address space region (memory, I/O, PCI, etc.) is represented by
.CW Nreg :
.P1
struct Nreg
{
	int	spc;		/* io space type */
	u64int	base;		/* address, physical */
	uchar*	p;		/* address, kmapped */
	uintptr	len;
	uint	tbdf;		/* pci only */
};
.P2
.LP
All other field types in ACPI are represented by
.CW Nfield .
This represents a slice (bits) of a region or another field and is the
main object used to perform I/O.
.P1
/*
 * bit field for a region, or banked-register bit field, or
 * (index,data) register bit field.
 */
struct Nfield
{
	char*	fname;		/* field name */
	int	accsz;		/* access sz in bits for source*/
	int	locking;	/* needs locking */
	int	update;		/* how to update */
	uintptr	off;		/* in bits */
	int	len;		/* in bits */
	Aobj*	reg;		/* region the field is for */
	Aobj*	bank;		/* bank name (for banked field) or nil */
	Aobj*	bankval;	/* bank value (for banked field) */
	Aobj*	idx;		/* index reg. (for index field) or nil */
	Aobj*	idxval;		/* value for index */
	Aobj*	data;		/* data (for index field) or nil */
};
.P2
.LP
To make things uniform, it has bit granularity so we may
forget in general if it is a bit or byte aligned field.
.CW Accsz
dictates the number of bits to I/O at a time.
.CW Update
dictate what to do with remaining bits in the source object upon writes to the field
(preverve them by reading them before, write as one or write as zero). There are
two kinds of fields:
.IP "\fIIndex fields\fP
.br
A slice of a
.CW data
register available after setting an index register,
.CW idx ,
to a value.
.IP "\fIRegion fields\fP
.br
A slice of a region, perhaps requiring a write to a
.CW bank
register before using.
.PP
Scope objects are not ACPI objects. They are used to represent a scope
(also introduced by an AML
.I scope
operation) as an aid in parsing.
.P1
struct Nscope
{
	char*	name;
	uchar*	pe;		/* end of AML for it */
};
.P2
.LP
A method including information about the number of arguments and the
portion of AML implementing the method's body.
.P1
struct Nmethod
{
	char*	name;
	int	nargs;		/* arity */
	int	isexcl;		/* is serialized? */
	int	synclvl;
	int	trace;		/* 1: on; -1: off; 0: as you were */
	uchar*	ps;		/* aml for this method */
	uchar*	pe;		/* end of aml */
	void	(*f)(Aobj*);	/* C implementation for builtins */
};
.P2
.LP
There are other objects that should be either self-explanatory or uninteresting
by now.
.SH
Devices
.LP
Devices include information about resources for them. They are kept in al list
of
.CW Res
items.
.P1
/*
 * ACPI device as seen by drivers.
 */
struct ACPIdev
{
	char*	name;
	char*	pnpid;		/* pnp id */
	u64int	addr;		/* address */
	Res*	res;		/* current resource settings */
};

struct Ndev
{
	ACPIdev;
	uchar*	pe;
	void	(*handler)(Aobj*, int);	/* for events */
	void*	aux;
};
.P2
.LP
A resource is described by this data type:
.P1
struct Res
{
	Res*	next;		/* in resource list */
	int	type;		/* resource type */
	union{
		struct{
			u32int	irqs[16];
			uint	flags;
			int	idx;	/* index in src */
			char*	src;	/* where resources come from */
		}irq;
		struct{
			uint	chans;
			uint	flags;
		}dma;
.P2
.P1
		struct{
			u64int	min;	/* address of the range */
			u64int	max;	/* address of the range */
			uint	align;
			int	isrw;	/* memory only, ro or rw? */
		}io, mem;
.P2
.P1
		struct{
			u64int	mask;	/* which bits are decoded */
			u64int	min;	/* address on the other side */
			u64int	max;
			u64int	off;	/* translation adds this */
			int	idx;	/* index in src */
			char*	src;	/* where resources come from */
			int	flags;
			int	rev;	/* acpi revision id or 0 */
			u64int	attr;
		}as;
		Gas field;
.P2
.P1
		/* non ACPI resources; invented by us to keep them here */
		struct{
			u64int	addr;
			int	pin;
			char*	src;
			int	idx;
		}prt;			/* pci intr. routing, from _PRT */
	};
};
.P2
.LP
This is a ``compact'' representation of the ACPI data structures for resource
settings, parsed by the implementation. A cleaner structure should be used
instead, once we know which pieces of information are really necessary.
.PP
There are other objects for processors, thermal zones, mutexes, etc. Just keep
in mind that they are not devices in ACPI.
.SH
The interpreter
.LP
The AML interpreter is a table driven implementation.
It is a single thread that
loops fetching operation codes and executing operations
according to them. Operation codes may be one or two bytes. Arguments
depend not only on the operation code, but also on the state of the interpreter.
.PP
A table,
.CW amlops ,
provides the implementation for each 1-byte operation code. Another table,
.CW amlxops ,
provides the implementation for, so called, extended operation codes. This
is an excerpt:
.P1
[OpZero]	{opnum,		"zero",		Dry, ""},
[OpOne]		{opnum,		"one",		Dry, ""},
[OpAlias]	{opalias,	"alias",	0, "On"},
[OpName]	{opname,	"name",		0, "nO"},
[OpMid]		{opmid,		"mid",		0, "OIIt"},
.P2
.LP
Each entry is defined by this type:
.P1
typedef void (*Opx)(void);
/*
 * AML operation
 */
struct Op
{
	Opx	x;		/* implementation */
	char*	name;		/* debug */
	int	flags;		/* call on dry runs? Print { } on dumps? */
	char*	args;		/* arguments, for AML parsing */
	int	ncalls;		/* how many times used */
};
.P2
.LP
The implementation assumes that each operation is responsible for
building an object (in general) and may require arguments to perform its
task. A string in each entry encodes argument processing so that,
after fetching an operation, the
.CW pargs
function may parse as much AML as needed to build argument objects
as dictated by the argument string. For example,
.CW opalias
requires an evaluated object and a name, as encoded by the string
.CW "On" ''. ``
.PP
In the argument string, a lowercase refers to an object of a particular type with
no evaluation performed on it. An uppercase refers to an object that, after evaluation,
has a particular type. For example,
.CW S '' ``
means to parse an object, evaluate it, and obtain a string object (perhaps by
doing a type cast to string). On the other hand,
.CW n '' ``
means to parse an object that should be a name.
.PP
These characters may be used to specify arguments:
.P1
Obuf	= 'B',		/* buf */
Oevent	= 'E',		/* sleep/wakeup */
Ofield	= 'F',,		/* register field */
Oint	= 'I',		/* integer */
Oname	= 'N',		/* a name */
Omutex	= 'M',		/* mutex */
Opkg	= 'P',		/* package object (array) */
Oreg	= 'R',		/* data region */
Ostr	= 'S',		/* string */
Odecstr	= 'D',		/* like string, but bytes from buffer */
			/* values are printed in decimal */
Oany	= 'O',		/* any object after resolving names */
.P2
.LP
These other characters, quoted from the implementation, may be
used but do not correspond to any object type as known by
the interpreter (they build objects, but do not represent different
object types):
.P1
/* 
 *	- 'l' means packagelen 
 * 	- 'b' means byte, 'w' word, 'd' d-word,
 * 	- 'f' fieldlist
 *	- 's' string
 *	- 'n' name or arg or local
 *	- 't' target (name, perhaps created, dst object, nil)
 *	- 'o' any object, maybe a name.
 * Upper case args are evaluated by pargs to obtain the typed arg.
 * Other args are not evaluated before calling the operation.
 */
.P2
.LP
But for the pointer to the implementation for each operation and the argument
string, the only interesting information in the operation table is a flag to see
if the operation must be called on dry runs or not and a flag to help pretty-print (sic)
AML for debugging.
.PP
The main entry point for the interpreter is
.P1
char*
amleval(uchar *ps, uchar *pe, Aobj **po, Name *n, Atable* table, int flags)
.P2
.LP
which evaluates AML code between pointers
.CW ps
and
.CW ps
using
.CW n
as the namespace dot for the evaluation. It returns an error string and the resulting
object in
.CW *po .
.PP
Another function evaluates a single call to a method object (just a subroutine):
.P1
Aobj*
amlcall(Aobj *mo, Aobj **args, Atable *table, int flags)
.P2
.LP
It accepts arguments for the method call and returns a result object.
.PP
Both functions are implemented by calling
.CW pamlblock ,
which parses (and evaluates) an AML block.
.CW Pamlblock
is mostly a loop that calls
.CW pamlobj ,
which parses (and evaluates) as much AML as needed to build a single
object. 
.PP
Depending on the flags given to the evaluator, it may just dissasemble AML, execute
it, print actions as they are performed, and perform various debug checks:
.IP \f(CWEdryrun\fP
.br
asks for a dry run. Usually combined with other flags that dump information.
.IP \f(CWEidump\fP
.br
asks for instruction dumps. This is close to a disassembler.
.IP \f(CWExdump\fP
.br
asks for messages about actions performed.
.IP \f(CWEscheck\fP
.br
asks for full stack checks to ensure that reference counting is correct and
to see if objects seem to be correct. This slows down execution a lot, but is useful
after making changes to the interpreter, because it is easy to make mistakes regarding
object references.
.LP
When the interpreter finds a method definition is skips its implementation and only
records where the code starts and its length. Method evaluation is performed by
using
.CW pamlblock
on that part of the program.
.SH
Environments
.LP
.LP
.CW Env
represents an evaluation environment at a particular scope. Multiple
.CW Env s
are nested as a result of entering and leaving scopes in the AML program.
This happens both to implement actual AML scopes and also to open
temporary scopes to save the current environment (e.g., during argument
evaluation). All
.CW Env
structures live in the C stack.
.PP
A global variable,
.CW amlenv ,
points to the current (i.e., last, or inner-most) environment. This
.CW Env
can be considered the state of the abstract machine and
is implicit to all AML operations.
.P1
struct Env
{
	char*	name;		/* debug */
	Env*	prev;		/* in env stack */
	int	ic;		/* instruction code */
	uchar*	ps;		/* start of what's left of program (PC) */
	uchar*	pe;		/* end of program */
	int	lvl;		/* env nesting level (debug) */

	Op*	op;		/* current op */
	uchar*	opp0;		/* operation address 0 */
	Nlist	olist;		/* list of obj in env */
	Aobj*	args[Nargs];	/* arguments for op */
	int	nargs;		/* max number of args used */
	uchar*	argpe;		/* pe resulting from arg 'l' */
	Aobj*	res;		/* result from op */
	
	Mframe*	mp;		/* Method frame pointer */
	Aname*	dot;		/* current acpi name */
	Atable*	table;		/* [ds]sdt table id */
	int	flags;
};
.P2
.LP
As said, the implementation looks up an operation code at a time and processes
its arguments as indicated by an argument string. Some arguments are simple
and are parsed directly. Others are not and may change the environment while
they are evaluated. In this case, a nested environment is pushed to protect the
environment for the operation (so that its implementation may assume that the
environment corresponds to the operation and has not been altered much during
argument processing).
.PP
When an operation is called:
.IP \f(CWamlenv->op\fP
.br
points to the operation being executed (its table entry).
.IP \f(CWamlenv->ic\fP
.br
is the instruction code for the operation (the second byte for
two-byte operations).
.IP \f(CWamlenv->ps\fP
.br
is the program counter (the start of what's left of the program). When
the operation includes an argument string, it points past the AML code
for the arguments. Upon return, it must point past the last byte of AML
code for the operation (including arguments and body, if any).
N.B.: the implementation for
.CW opwhile
violates this assumption for a good reason.
.IP
Operations with implementations that are also used to parse arguments should look at
.CW amlenv->ps[0]
and not to
.CW amlenv->ic
to see their IC.
That is because the IC is assumed to correspond to the operation itself
and not to its arguments.
.IP \f(CWamlenv->pe\fP
.br
is the end of the program. Operations may include a length argument that
constraints the program size (e.g., to define a portion of code as the
.I then -arm
of an
.I if
instruction).
.IP  \f(CWamlenv->opp0\fP
.br
is the program counter for the current operation (the zero address for its AML code). This
is important to
.CW opwhile .
.LP
When operations with argument strings are called:
.IP \f(CWamlenv->ps\fP
.br
is past the arguments.
.IP \f(CWamlenv->args\fP
.br
includes pointers to argument objects for the operation.
.IP \f(CWamlenv->argpe\fP
.br
is the end of the program according to the length argument. The operation should
not touch AML code past it.
.LP
Upon return from an operation:
.IP \f(CWamlenv->res\fP
.br
points to the object returned from the operation (perhaps
.CW nil ).
.LP
Arguments and locals are closed when the operation returns. When the
environment is terminated (pop) or the next operation executes,
.CW amlenv->res
is closed. Thus, operations that don't have to return a value may leave an
object linked there, so it is collected upon errors.
.PP
Environments used to evaluate methods include a pointer to an
.CW Mframe .
.P1
/*
 * Method activation frame.
 */
struct Mframe
{
	Aobj*	args[Nmargs];	/* method arguments */
	int	nargs;
	Aobj*	lcls[Nmlcls];	/* method locals */
	int	nlcls;
	Aobj**	targs;		/* tmp. method args while parsing them */
	Aname*	binds;		/* done by this method call */
};
.P2
This is just a place to store method arguments and locals, as well as a
place to keep the list of names bound by a method call (which must be
unbound upon returns). 
A method call pushes a new
.CW Env
(in the C stack) that points to a new
.CW Mframe
(in the C stack)
via
.CW mp .
Further scopes entered inherit
the
.CW mp
pointer from the parent
.CW Env .
.SH
Error handling
.LP
Remember that there are no processes by the time the interpreter runs.
Errors are reported using an error stack, similar to that used by processes.
This is an example:
.P1
static void
oppackage(void)
{
	Env env;
	...
	amlpush(&env, amlenv->op->name, nil, amlenv->argpe);
	if(wasamlerror()){
		amlpop();
		amlerror(nil);	/* pass a string to raise it */
	}
	pamlblock(Keep);
	...
	popamlerror();
	amlpop();
}
.P2
Operations
.CW break ,
.CW continue ,
and
.CW return
raise their names and rely on this exception mechanism for their implementation.
