/*
 * AML conversion, copy, and I/O routines.
 * ACPI has all kinds of requirements on casts and
 * assignments. It is easy to make mistakes there and
 * this file keeps that code apart.
 *
 * Regarding I/O, we must support different regions with
 * different alignment constraints and preserving rules.
 * Only some of the required regions are here. Others are missing.
 */
#include "kernel.h"
#include "acpi.h"
#include "aml.h"

static void fieldwr(Aobj*,Aobj*);
static void fieldio(Aobj *o, void *a, int iswr);
static Aobj*tcast(Aobj*, int);
static Aobj*tobuf(Aobj*);
static Aobj*dupobj(Aobj*, Aobj*);


/*
 * bit numbers: [7-0] [15-8] ...
 */

static void
bitset(void *d, int dbit, int val)
{
	char *dp;

	dp = d;
	dp += dbit >> 3;
	dbit &= 7;
	if(val != 0)
		*dp |= 1<<dbit;
	else
		*dp &= ~(1<<dbit);
}

static int
bitget(void *d, int dbit)
{
	char *dp;

	dp = d;
	dp += dbit >> 3;
	dbit &= 7;
	return (*dp & 1<<dbit) != 0;
}

/*
 * copy bits from s to d starting at sbit in s, dbit in d.
 * copy at most len bits.
 */
static void
bitcpy(void *d, int dbit, void *s, int sbit, long len)
{
	char *dp, *sp;
	int n, i;

	dp = d;
	sp = s;
	/* see if we can copy some whole bytes */
	if(((dbit|sbit) & 7) == 0 && len > 7){
		n = len >> 3;
		memmove(dp + (dbit>>3), sp + (sbit>>3), n);
		dbit += n<<3;
		sbit += n<<3;
		len -= n<<3;
	}
	/* copy remaining bits */
	for(i = 0; i < len; i++)
		bitset(dp, dbit++, bitget(sp, sbit++));

}

static void
idxfieldio(Aobj *o, void *a, int iswr)
{
	Aobj *data;
	uchar *p;
	int nbytes;

	amltcheck(o->field.idx, Ofield, 0);
	data = o->field.data;
	amltcheck(data, Ofield, 0);
	fieldwr(o->field.idx, o->field.idxval);
	nbytes = (data->field.len+7)/8;
	p = amlmalloc(nbytes);
	if(wasamlerror()){
		free(p);
		amlerror(nil);
	}
	if(iswr && (o->field.off != 0 || o->field.len != nbytes*8))
		switch(o->field.update){
		case AFwrones:
			memset(p, ~0, nbytes);
			break;
		case AFpreserve:
			fieldio(data, p, 0);
			break;
		}
	if(iswr){
		bitcpy(p, o->field.off, a, 0, o->field.len);
		fieldio(data, p, 1);
	}else{
		fieldio(data, p, 0);
		bitcpy(a, 0, p, o->field.off, o->field.len);
	}
	popamlerror();
	free(p);
	return;
}

/*
 * Read all field bits into a or write all of them from a,
 * preverve old bits in source if needed.
 */
static void
fieldio(Aobj *o, void *a, int iswr)
{
	uchar *p;
	int nbytes, mask;
	uintptr saddr, eaddr;

	doprint("field%s [%O] %#p\n", iswr ? "out" : "in", o, a);
	if(o->field.idx != nil){
		idxfieldio(o, a, iswr);
		return;
	}
	if(o->field.bank != nil)
		fieldwr(o->field.bank, o->field.bankval);
	/* compute addresses relative to the region */
	mask = o->field.accsz/8 - 1;
	saddr = o->field.off/8;				/* bytes */
	saddr &= ~mask;
	eaddr = (o->field.off + o->field.len + 7) / 8;	/* bytes */
	eaddr = (eaddr + mask) & ~mask;
	nbytes = eaddr - saddr;
	p = amlmalloc(nbytes);
	if(wasamlerror()){
		free(p);
		amlerror(nil);
	}
	if(iswr && (o->field.off != saddr*8 || o->field.len != nbytes*8))
		switch(o->field.update){
		case AFwrones:
			memset(p, ~0, nbytes);
			break;
		case AFpreserve:
			regio(o->field.reg, p, saddr, nbytes, o->field.accsz/8, 0);
			break;
		}
	if(iswr){
		bitcpy(p, o->field.off - saddr*8, a, 0, o->field.len);
		regio(o->field.reg, p, saddr, nbytes, o->field.accsz/8, 1);
	}else{
		regio(o->field.reg, p, saddr, nbytes, o->field.accsz/8, 0);
		bitcpy(a, 0, p, o->field.off - saddr*8, o->field.len);
	}
	popamlerror();
	free(p);
}

/*
 * read from a field into a buffer object.
 */
static void
fieldrd(Aobj *fo, Aobj *o)
{
	uchar *p;

	dxprint("fieldrd %O %O\n", fo, o);
	amltcheck(fo, Ofield, 0);
	amltcheck(o, Obuf, 0);
	if(o->buf.len * 8 < fo->field.len){
		p = amlmalloc((fo->field.len+7)/8);
		if(wasamlerror()){
			free(p);
			amlerror(nil);
		}
		fieldio(fo, p, 0);
		memmove(o->buf.p, p, o->buf.len);
		popamlerror();
		free(p);
	}else{
		memset(o->buf.p, 0, o->buf.len);
		fieldio(fo, o->buf.p, 0);
	}
}

/*
 * write o into a field
 */
static void
fieldwr(Aobj *fo, Aobj *o)
{
	Aobj *no;
	char *s;
	void *p;
	int n, fbytes, tot, left;

	dxprint("fieldwr %O %O\n", fo, o);
	amltcheck(fo, Ofield, 0);
	p = nil;
	no = nil;
	if(wasamlerror()){
		amlclose(no);
		free(p);
		amlerror(nil);
	}
	fbytes = (fo->field.len+7)/8;
	switch(o->type){
	case Ostr:
	case Odecstr:
		/* write each character as an integer value */
		no = amlnewi(0);
		for(s = o->sval; *s != 0; s++){
			no->ival = *s;
			fieldwr(fo, no);
		}
		break;
	case Oint:
		/* write bits from it, truncating or zero extending */
		if(fbytes > acpisz)
			p = amlmalloc(fbytes);
		else
			p = amlmalloc(acpisz);
		memmove(p, &o->ival, acpisz);
		fieldio(fo, p, 1);
		break;
	case Obuf:
		/* write from o->buf.p fo->len bits at a time */
		p = amlmalloc(fbytes);
		tot = 0;
		for(left = o->buf.len * 8; left > 0; left -= n){
			n = fo->field.len;
			if(left < n)
				n = left;
			bitcpy(p, 0, o->buf.p, tot, n);
			fieldio(fo, p, 1);
			memset(p, 0, fbytes);
			tot += n;
		}
		break;
	default:
		no = tcast(o, Obuf);
		fieldwr(fo, no);
		break;
	}
	popamlerror();
	amlclose(no);
	free(p);
}

/*
 * copy nbits bits from s (offset sbit bits) to a slice object.
 */
static void
slicecpy(Aobj *o, void *s, int sbit, int nbits)
{

	dxprint("slicecpy: s %#p off %#x:%d sz %#x -> [%O]\n",
		s, sbit/8, sbit&7, nbits, o);
	amltcheck(o, Oslice, 0);
	assert(o->slice.src != nil && o->slice.src->type == Obuf);

	bitcpy(o->slice.src->buf.p, o->slice.off, s, sbit, nbits);
}

/*
 * Type cast tools.
 */

static char*
inttostr(u64int ival, int type)
{
	char buf[15];
	char *fmt[2] = { "%#08ux", "%#016ullx" };
	char *s;

	if(type == Odecstr){
		fmt[0] = "%08ud";
		fmt[1] = "%016ud";
	}
	if(amlenv->table->is64 == 0)
		seprint(buf, buf+sizeof(buf), fmt[0], (u32int)ival);
	else
		seprint(buf, buf+sizeof(buf), fmt[1], ival);
	s = nil;
	kstrdup(&s, buf);
	return s;
}

static u64int
buftoint(uchar *p, int len)
{
	u64int ival;
	int i;

	ival = 0;
	if(acpisz < len)
		len = acpisz;
	for(i = 0; i < len; i++)
		ival |= p[i] << i*8;
	return ival;
}

static u64int
strtoint(char *s, int type)
{
	int n, c, i;
	u64int ival;

	ival = 0;
	n = acpisz * 2;
	for(i = 0; i < n; i++){
		if(s[i] >= '0' && s[i] <= '9')
			c = s[i]- '0';
		else if(s[i] >= 'a' && s[i] <= 'f')
			c = s[i] - 'a' + 10;
		else if(s[i] >= 'A' && s[i] <= 'F')
			c = s[i] - 'A' + 10;
		else
			break;
		if(type == Ostr)
			ival = (ival << 4) | c;
		else
			ival = ival * 10 + c;
	}
	return ival;
}

static Aobj*
toint(Aobj *o)
{
	Aobj *no, *bo;

	switch(o->type){
	case Ohandle:
		no = amlnewi(o->ival);
		break;
	case Ostr:
	case Odecstr:
		no = amlnewi(strtoint(o->sval, o->type));
		break;
	case Obuf:
		no = amlnewi(buftoint(o->buf.p, o->buf.len));
		break;
	default:
		bo = tobuf(o);
		no = toint(bo);
		amlclose(bo);
		break;
	}
	return no;
}

static Aobj*
tobuf(Aobj *o)
{
	Aobj *no, *so;
	void *p;
	int len;

	len = 0;
	p = nil;
	no = amlnew(Obuf);
	switch(o->type){
	case Oslice:
		so = o->slice.src;
		no->buf.p = amlmalloc(so->buf.len);
		bitcpy(no->buf.p, 0, so->buf.p, o->slice.off, o->slice.len);
		no->buf.len = o->slice.len >> 3;
		if(o->slice.len & 7)
			no->buf.len++;
		return no;
	case Ofield:
		len = o->field.len >> 3;
		if(o->field.len & 7)
			len++;
		no->buf.p = amlmalloc(len);
		no->buf.len = len;
		fieldrd(o, no);
		return no;
	case Oint:
	case Ohandle:
		p = &o->ival;
		len = acpisz;
		break;
	case Ostr:
	case Odecstr:
		p = o->sval;
		len = strlen(o->sval) + 1;
		if(len == 1)
			len = 0;
		break;
	default:
		amlclose(no);
		amlerror("cast %s -> buf not supported", amltname(o->type));
	}
	no->buf.len = len;
	no->buf.p = amlmalloc(len);
	if(len > 0)
		memmove(no->buf.p, p, len);
	return no;
}

static Aobj*
tostr(Aobj *o, int type)
{
	Aobj *no, *bo;
	char *s, *se;
	int i;
	char *fmt;

	fmt = "%#02ux";
	if(type == Odecstr)
		fmt = "%ud";
	no = amlnew(type);
	switch(o->type){
	case Ohandle:
	case Oint:
		no->sval = inttostr(o->ival, type);
		break;
	default:
		if(wasamlerror()){
			amlclose(no);
			amlerror("cast %s -> str not supported",  amltname(o->type));
		}	
		bo = tobuf(o);
		popamlerror();
		if(bo->buf.len == 0)
			kstrdup(&no->sval, "");
		else if((o->type == Oslice || o->type == Ofield) && bo->buf.len <= acpisz){
			/* must handle as integer */
			no->sval = inttostr(buftoint(bo->buf.p, bo->buf.len), Ostr);
		}else{
			s = no->sval = amlmalloc(bo->buf.len * 4 + 1);
			se = s + bo->buf.len * 4 + 1;
			for(i = 0; i < bo->buf.len; i++){
				s = seprint(s, se, fmt, bo->buf.p[i]);
				if(i < bo->buf.len - 1)
					s = seprint(s, se, ",");
			}
		}
		amlclose(bo);
	}
	return no;
}

static Aobj*
duplist(Aobj *l)
{
	Aobj *nl, *r, **p;

	if(l == nil)
		return nil;
	r = nl = amlnew(l->type);
	dupobj(nl, l);
	p = &nl->next;
	for(; l != nil; l = l->next){
		nl = amlnew(Onone);
		dupobj(nl, l);
		*p = nl;
		p = &nl->next;
	}
	return r;
}

/*
 * only memory copying and field I/O.
 * building block for amlcpy.
 */
static Aobj*
dupobj(Aobj* o, Aobj *so)
{
	Aobjhdr hdr;

	/*
	 * args and locals must always keep their types.
	 */
	assert(o->type != Oarg && o->type != Olocal);

	/*
	 * we copy field and slices always from buffers
	 */
	if(o->type == Oslice){
		amltcheck(so, Obuf, 0);
		slicecpy(o, so->buf.p, 0, so->buf.len * 8);
		return o;
	}
	if(o->type == Ofield){
		fieldwr(o, so);
		return o;
	}

	hdr = o->Aobjhdr;
	amlclean(o);
	memmove(o, so, sizeof(Aobj));
	o->Aobjhdr = hdr;

	switch(so->type){
	case Onone:
	case Oint:
	case Oscope:
	case Omethod:
	case Odev:
	case Opwr:
	case Oproc:
	case Othermal:
	case Ohandle:
	case Oreg:
		break;
	case Opkg:
		o->list.hd = duplist(o->list.hd);
		break;
	case Oref:
		if(o->oref.o != nil)
			incref(o->oref.o);   
		break;
	case Ostr:
	case Odecstr:
	case Oname:
		o->sval = nil;
		kstrdup(&o->sval, so->sval);
		break;
	case Obuf:
		o->buf.p = amlmalloc(o->buf.len);
		memmove(o->buf.p, so->buf.p, o->buf.len);
		break;
	case Omutex:
		memset(&o->mutex, 0, sizeof(Nmutex));
		break;
	case Oevent:
		memset(&o->event, 0, sizeof(Nevent));
		break;
	default:
		panic("dupobj: unknown type %s\n", amltname(o->type));
	}
	if(o->type == Onone)
		o->type = so->type;
	return o;
}

/*
 * return the object we should read to copy from it or to
 * build references to it (does not add a new ref).
 */
Aobj*
amlread(Aobj *o)
{
	Aname *nm;

	switch(o->type){
	case Oarg:
		o = o->oref.o;
		if(o == nil)
			amlerror("read from null argument");
		if(o->type == Oref)
			o = o->oref.o;
		break;
	case Olocal:
		o = o->oref.o;
		if(o == nil)
			amlerror("read from null local");
		break;
	case Oname:
		nm = amlwalk(o->sval, Walkns, nil);
		if(nm == nil)
			amlerror("read from unbound name %s", o->sval);
		o = nm->o;
		break;
	}
	if(o == nil)
		amlerror("read from null object");
	return o;
}

static Aobj*
tcast(Aobj *o, int type)
{
	if(o->type == type || type == Onone || type == Oany){
		incref(o);
		return o;
	}
	dxprint("cast %s <- %s %O\n", amltname(type), amltname(o->type), o);

	switch(type){
	case Ostr:
	case Odecstr:
		o = tostr(o, type);
		break;
	case Obuf:
		o = tobuf(o);
		break;
	case Oint:
		o = toint(o);
		break;
	default:
		o = nil;
		amlerror("tcast: cast to %s not supported", amltname(type));
	}
	dxprint("cast %s -> %O\n", amltname(type), o);
	return o;
}

/*
 * Copy object so into o, perhaps doing a type conversion
 * if castok. It's ok if o is nil (may be a null target).
 * Copy must follow requirements regarding how to copy
 * to local objects and to argument objects. See 18.2.5.8.
 */
void
amlcpy(Aobj *o, Aobj *so, int castok)
{
	int dtype;
	Aobj *no, *to;
	Aname *nm;

	if(o == nil)
		return;


	dxprint("amlcpy [%O] <- [%O] %s\n", o, so, castok ? "castok" : "");
	to = no = nil;
	if(so->type == Ofield || so->type == Oslice)
		to = so = tcast(so, Obuf);
	if(wasamlerror()){
		amlclose(to);
		amlclose(no);
		amlerror(nil);
	}
	switch(o->type){
	case Oarg:
		no = o->oref.o;
		if(no != nil && no->type == Oref && no->oref.o != nil)
			o = no->oref.o;
		else{
			incref(so);
			popamlerror();
			amlclose(o->oref.o);
			o->oref.o = so;
			amlclose(to);
			return;
		}
		break;
	case Olocal:
		popamlerror();
		amlclose(o->oref.o);
		o->oref.o = amlnew(so->type);
		dupobj(o->oref.o, so);
		amlclose(to);
		dxprint("amlcpy [%O]\n", o);
		return;
	case Oname:
		nm = amlwalk(o->sval, Walkns, nil);
		if(nm == nil)
			amlerror("copy to unbound name %s", o->sval);
		o = nm->o;
		break;
	}

	if(castok && o->type != Ofield){
		dtype = o->type;
		if(dtype == Oslice)
			dtype = Obuf;
	}else
		dtype = so->type;

	so = amlread(so);
	no = nil;
	if(dtype != so->type)
		so = no = tcast(so, dtype);
	dupobj(o, so);
	popamlerror();
	amlclose(no);
	amlclose(to);
}

/*
 * Evaluate an object trying to get something of the type wanted.
 */
Aobj*
amlcast(Aobj *o, int twanted)
{
	Aname *nm;
	Aobj *no, *ro;

	if(o == nil)
		return nil;
	incref(o);
	ro = o;
	if(wasamlerror()){
		amlclose(ro);
		amlerror(nil);
	}
	while(ro != nil && ro->type != twanted){
		switch(ro->type){
		case Olocal:
		case Oarg:
		case Oref:
			no = ro->oref.o;
			if(no != nil)
				incref(no);
			break;
		case Oname:
			nm = amlwalk(ro->sval, Walkns, nil);
			if(nm == nil){
				amlwarn("cast to %s: '%s' not found in '%s'\n",
					amltname(twanted), amlenv->dot->s, ro->sval);
				amlclose(ro);
				ro = nil;
				goto Found;
			}
			no = nm->o;
			if(no != nil)
				incref(no);
			break;
		case Omethod:
			no = opcall(ro, nil);
			break;
		default:
			if(twanted == Oany)
				goto Found;
			no = tcast(ro, twanted);
		}
		amlclose(ro);
		ro = no;
	}
	if(ro == nil || (twanted != Oany && ro->type != twanted))
		if((amlenv->flags&Enoeval) == 0)
			amlerror("no such object");
		else
			amlwarn("objeval: null result\n");
Found:
	popamlerror();
	return ro;
}

