/*
 * Main driver for AML operations.
 *
 * On most of them, the comment describes which
 * args are taken and which type of result is produced, if any.
 * Syntax is the same used for argument strings. 
 */
#include "kernel.h"
#include "acpi.h"
#include "aml.h"

/* Op flags */
enum {
	Dry = 1,	/* call op on dry runs */
	Bdy = 2,	/* print { } around body */
};

static Aobj*	pamlobj(void);
static Aobj**	checkargs(char*);
static Op*	decode(uchar* ps, uchar *pe, int *icp);

static void
opintprefix(int sz)
{
	int i;

	amlenv->res = amlnew(Oint);
	for(i = 0; i < sz; i += 8)
		amlenv->res->ival |= amlnext() << i;
	amlnext();
}

/* -> I */
static void
opbyteprefix(void)
{
	opintprefix(8);
}

/*
 * DANGER: prev is to be used only when we want to call an operation
 * just to decode its arguments and we are faking the instruction.
 * We step back so that the op call will skip the fake op and
 * start with the first byte of the arguments.
 */
static void
prev(void)
{
	amlenv->ps--;
}

/* process arg 'b' */
static Aobj*
pbyte(void)
{
	prev();
	opintprefix(8);
	return amlenv->res;
}

/* -> I */
static void
opwordprefix(void)
{
	opintprefix(16);
}

/* process arg 'w' */
static Aobj*
pword(void)
{
	prev();
	opintprefix(16);
	return amlenv->res;
}

/* -> I */
static void
opdwordprefix(void)
{
	opintprefix(32);
}

/* process arg 'd' */
static Aobj*
pdword(void)
{
	prev();
	opintprefix(32);
	return amlenv->res;
}

/* -> I */
static void
opqwordprefix(void)
{
	opintprefix(64);
}

/* -> S */
static void
opstrprefix(void)
{
	Aobj *o;
	int l;

	for(l = 0; amlenv->ps[l] != 0; l++)
		if(amlenv->ps == amlenv->pe)
			amlerror("short string");
	o = amlenv->res = amlnew(Ostr);
	o->sval = amlmalloc(l+1);
	if(l > 0)
		memmove(o->sval, amlenv->ps, l);
	o->sval[l] = 0;
	amlenv->ps += l+1;

	diprint("%!'%s'\n", o->sval);	/* see the value */
}

/* -> O (it's Olocal) */
static void
oplocal(void)
{
	int n;
	Mframe *mp;

	n = amlenv->ps[0] - OpLocal0;
	amlnext();
	mp = amlenv->mp;
	if(mp == nil)
		panic("oplocal outside a call");
	if(n < 0 || n >= Nmlcls)
		amlerror("bad local number");
	if(mp->lcls[n] == nil){
		if(mp->nlcls < n+1)
			mp->nlcls = n+1;
		mp->lcls[n] = amlnew(Olocal);
		mp->lcls[n]->oref.no = n;
	}
	incref(mp->lcls[n]);
	amlenv->res = mp->lcls[n];
}

/* -> O (it's Oarg) */
static void
oparg(void)
{
	int n;
	Mframe *mp;

	n = amlenv->ps[0] - OpArg0;
	amlnext();
	mp = amlenv->mp;
	if(mp == nil)
		panic("oparg outside a call");
	if(n < 0 || n >= mp->nargs)
		amlerror("bad arg number");
	
	if(mp->args[n] == nil){
		mp->args[n] = amlnew(Oarg);
		mp->args[n]->oref.no = n;
	}
	incref(mp->args[n]);
	amlenv->res = mp->args[n];
}

/*
 * -> N
 *
 * Args and locals can be used as names in the grammar.
 * It's called a supername. sic.
 * This operation encapsulates this, and returns a
 * local or an argument when it's a supername and not
 * a name what we want.
 */
static Aobj*
pnamechar(void)
{
	Aobj *o;
	int nels;
	int i;
	char prefix[50];
	char *s;

	if(amlenv->ps[0] >= OpArg0 && amlenv->ps[0] <= OpArg6){
		oparg();
		return amlenv->res;
	}
	if(amlenv->ps[0] >= OpLocal0 && amlenv->ps[0] <= OpLocal7){
		oplocal();
		return amlenv->res;
	}

	o = amlenv->res = amlnew(Oname);
	s = prefix;
	*s = 0;
	if(amlenv->ps[0] == '\\'){
		*s++ = amlenv->ps[0];
		amlnext();
	}
	while(amlenv->ps[0] == '^'){
		*s++ = '^';
		amlnext();
	}
	*s = 0;
	switch(amlenv->ps[0]){
	case OpDnameprefix:
		nels = 2;
		amlnext();
		break;
	case OpMnameprefix:
		nels = amlnext();
		amlnext();
		break;
	case 0:
		kstrdup(&o->sval, "");
		amlnext();
		return o;
	default:
		if((amlenv->ps[0] >= 'A' && amlenv->ps[0] <= 'Z') ||
		   amlenv->ps[0] == '_')
			nels = 1;
		else{
			amlclose(amlenv->res);
			return amlenv->res = nil;
		}
		break;
	}
	o->sval = amlmalloc(sizeof(prefix)+nels*5+1);
	strcpy(o->sval, prefix);
	s = o->sval + (s-prefix);
	for(i = 0; i < nels; i++){
		memmove(&s[i*5], amlenv->ps, 4);
		if(i+1 < nels)
			s[i*5+4] = '.';
		else
			s[i*5+4] = 0;
		amlenv->ps += 4;
	}
	return o;
}

/*
 * -> n
 */
static void
opnamechar(void)
{
	Aobj *o;
	Aname *n;

	o = amlenv->res = pnamechar();
	if(o == nil || o->sval == nil)
		amlerror("bad name");


	if(amlenv->flags&(Enocall)){
		diprint("%!\"%s\"\n", o->sval);
		return;
	}

	n = amlwalk(o->sval, Walkns, nil);
	if(n == nil || n->o == nil){
		diprint("%!\"%s\"\n", o->sval);
		if((amlenv->flags&Enoeval) == 0)
			amlerror("name %s not found or bound to nil\n", o->sval);
		else
			amlwarn("name %s not found or bound to nil\n", o->sval);
		return;
	}
	amlenv->res = nil;
	if(n->o->type != Omethod){
		diprint("%!\"%s\"\n", o->sval);
		incref(n->o);
		amlclose(o);
		amlenv->res = n->o;
	}else{
		amlclose(o);
		amlenv->res = opcall(n->o, nil);
	}
}


/*
 * -> I
 */
static void
opnum(void)
{
	Aobj *o;

	o = amlenv->res = amlnew(Oint);
	switch(amlenv->ic){
	case OpZero:
	case OpOne:
		o->ival = amlenv->ic;
		break;
	case OpOnes:
		o->ival = ~0;
		break;
	case OpTimer:
	case OpDebug:
		o->ival = perfticks();
		break;
	case OpRevision:
		o->ival = 9;
		break;
	default:
		panic("opnum bug");
	}
}

/*
 * name:n obj:O
 */
static void
opname(void)
{
	Aobj **args;

	args = checkargs("nO");
	amlbind(args[0]->sval, args[1]);
}

/*
 * new name for object.
 * srcobj:O  name:n
 */
static void
opalias(void)
{
	Aobj **args;

	args = checkargs("On");
	amlbind(args[1]->sval, args[0]);
}

/*
 * name:n
 */
static void
opevent(void)
{
	Aobj **args;

	args = checkargs("n");
	amlenv->res = amlnew(Oevent);
	amlbind(args[0]->sval, amlenv->res);
}

/*
 * name:n flags:b
 */
static void
opmutex(void)
{
	Aobj **args, *o;

	args = checkargs("nb");
	o = amlenv->res = amlnew(Omutex);
	o->mutex.flags = args[1]->ival;
	amlbind(args[0]->sval, amlenv->res);
}

/*
 * named region in memory, I/O, etc.
 * name:n space:b base:I len:I
 */
static void
opopregion(void)
{
	Aobj **args, *o;

	args = checkargs("nbII");
	o = amlenv->res = amlnew(Oreg);
	o->reg.spc = args[1]->ival;
	o->reg.base = args[2]->ival;
	o->reg.len = args[3]->ival;
	amlbind(args[0]->sval, o);
}

/*
 * len sz:I data -> B
 */
static void
opbuf(void)
{
	Aobj **args, *o;
	int n, sz;

	args = checkargs("I");
	if(args[0]->ival > 64 * 1024)
		amlerror("bad buffer sz %lld", args[0]->ival);
	o = amlenv->res = amlnew(Obuf);
	sz = o->buf.len = args[0]->ival;
	o->buf.p = amlmalloc(sz+1);
	n = amlenv->argpe - amlenv->ps;
	if(n > sz)
		n = sz;
	if(n > 0)
		memmove(o->buf.p, amlenv->ps, n);
	if(sz > n)
		memset(o->buf.p + n, 0, sz - n);

	amlenv->ps = amlenv->argpe;
}

/*
 * It's an array, we use a list.
 * package:  len nitems:b ... -> P
 * varpackage: pkglen nitems:I ... -> P
 */
static void
oppackage(void)
{
	Aobj **args, *o, **ol;
	Env env2;

	args = checkargs("I");
	if(args[0]->ival > 256)
		amlerror("package with more than 256 objects");
	o = amlenv->res = amlnew(Opkg);
	o->list.n = args[0]->ival;
	amlpush(&env2, amlenv->op->name, nil, amlenv->argpe);

	env2.flags |= Enocall;

	/* Ignore errors and try to get as many elements as
	 * we can. Users may always check out if they got what
	 * they need.
	 * We may return more elements than declared upon errors,
	 * but that should not harm and it's easier.
	 */
	if(!wasamlerror()){
		pamlblock(Keep);
		popamlerror();
	}
	o->list.hd = amlenv->olist.hd;
	if(o->list.n > amlenv->olist.n){
		for(ol = &o->list.hd; *ol != nil; ol = &(*ol)->next)
			;
		while(o->list.n > amlenv->olist.n){
			*ol = amlnewi(0);
			ol = &(*ol)->next;
			amlenv->olist.n++;
		}
	}

	amlenv->olist.hd = nil;
	amlenv->olist.n = 0;
	amlpop();
}

static void
setfieldflags(Aobj *o, uint ival)
{
	o->field.accsz = ival & 0xF;
	switch(o->field.accsz){
	default:
		o->field.accsz = 8;
		break;
	case AFword:
		o->field.accsz = 16;
		break;
	case AFdword:
		o->field.accsz = 32;
		break;
	case AFqword:
		o->field.accsz = 64;
		break;
	}
	o->field.locking = ival>>4 & 1;
	o->field.update = ival>>5 & 3;
}

/*
 * Create a bit field for a region.
 * field: len reg:R flags:b fields:f
 */
static void
opfield(void)
{
	Aobj *ol, **args;

	args = checkargs("Rbf");
	for(ol = args[2]->list.hd; ol != nil; ol = ol->next){
		amltcheck(ol, Ofield, 1);
		incref(args[0]);
		assert(ol->field.reg == nil);
		ol->field.reg = args[0];
		setfieldflags(ol, args[1]->ival);
		if(ol->field.fname != nil)
			amlbind(ol->field.fname, ol);
		else
			amlwarn("field with no name %O\n", ol);
	}
}

/*
 * Create bit field for a banked register.
 * len reg:R bankname:n bankval:I flags:b fields:f
 */
static void
opbankfield(void)
{
	Aobj *ol, **args;
	char *bname;
	Aname *n;

	args = checkargs("RnIbf");
	for(ol = args[4]->list.hd; ol != nil; ol = ol->next){
		amltcheck(ol, Ofield, 1);
		incref(args[0]);
		ol->field.reg = args[0];
		bname = args[1]->sval;
		n = amlwalk(bname, Walkns, nil);
		if(n != nil && n->o != nil && n->o->type == Ofield){
			incref(n->o);
			ol->field.bank = n->o;
		}
		ol->field.bankval = amlnewi(args[2]->ival);
		setfieldflags(ol, args[3]->ival);
		if(ol->field.fname == nil)
			amlwarn("bankfield with no name %O\n", ol);
		else if(ol->field.bank == nil)
			amlwarn("bankfield with no bank '%s' %O\n", bname, ol);
		else
			amlbind(ol->field.fname, ol);
	}
}

/*
 * Create a bit field within a data object indexed by a value.
 * len idx:F data:F flags:b fields:f
 */
static void
opindexfield(void)
{
	Aobj *ol, **args;

	args = checkargs("FFbf");
	for(ol = args[3]->list.hd; ol != nil; ol = ol->next){
		amltcheck(ol, Ofield, 1);
		/* adjust offset back to bytes (value for index) */
		incref(args[0]);
		ol->field.idx = args[0];
		ol->field.idxval = amlnewi(ol->field.off >> 3);
		ol->field.off &= 7;
		incref(args[1]);
		ol->field.data = args[1];
		setfieldflags(ol, args[2]->ival);
		if(ol->field.fname != nil)
			amlbind(ol->field.fname, ol);
		else
			amlwarn("indexfield with no name %O\n", ol);
	}
}

/*
 * Create a slice of a buffer.
 * We always use bit offsets/lengths, to keep all of them homogeneous.
 * mkfield: buf:B off:I len:I name:n
 * mk{bit,byte,word,dword,qword}field: buf:B off:I name:n
 */
static void
opmkxfield(void)
{
	Aobj **args, *o, *no;

	if(amlenv->ic == OpMkfield){
		args = checkargs("BIIn");
		no = args[3];
	}else{
		args = checkargs("BIn");
		no = args[2];
	}
	o = amlenv->res = amlnew(Oslice);
	incref(args[0]);
	o->slice.src = args[0];
	switch(amlenv->ic){
	case OpMkfield:
		o->slice.off = args[1]->ival;
		o->slice.len = args[2]->ival;
		break;
	case OpMkbitfield:
		o->slice.off = args[1]->ival;
		o->slice.len = 1;
		break;
	case OpMkbytefield:
		o->slice.off = args[1]->ival<<3;
		o->slice.len = 8;
		break;
	case OpMkwordfield:
		o->slice.off = args[1]->ival<<3;
		o->slice.len = 16;
		break;
	case OpMkdwordfield:
		o->slice.off = args[1]->ival<<3;
		o->slice.len = 32;
		break;
	case OpMkqwordfield:
		o->slice.off = args[1]->ival<<3;
		o->slice.len = 64;
		break;
	default:
		panic("opmkxfield bug");
	}
	amlbind(no->sval, o);
}

Aobj*
nth(Aobj *o, int n)
{
	Aobj *l;

	for(l = o->list.hd; l != nil; l = l->next)
		if(n-- == 0)
			return l;
	return nil;
}

/*
 * reference to ith object in another object.
 * bufstrpkg:O idx:I res:t -> Oref
 */
static void
opindex(void)
{
	Aobj **args, *o, *no;

	args = checkargs("OIt");
	o = amlenv->res = amlnew(Oref);
	/*
	 * return ref to idx-th object within O
	 */
	switch(args[0]->type){
	case Opkg:
		no = o->oref.o = nth(args[0], args[1]->ival);
		if(no != nil)
			incref(no);
		break;
	case Oint:
	case Ostr:
	case Odecstr:
	case Obuf:
		no = o->oref.o = amlnew(Oslice);
		no->slice.src = amlcast(args[0], Obuf);
		no->slice.off = args[1]->ival << 3;
		no->slice.len = 8;
		break;
	default:
		amlerror("index on object of type %s", amltname(args[0]->type));
	}
	amlcpy(args[2], o, Cast);
}

/*
 * Complex object with nested scope.
 * scope: len name:n ...
 * device: len name:n ...
 * method: len name:n flags:b ...
 * pwrrsrc: len name:n syslvl:b rorder:w ...
 * processor: len name:n id:b baddr:d blen:b ...
 * thermalzone: len name:n ...
 */
static void
opscope(void)
{
	Aobj *o, **args;
	Env env2;
	char *name, *nval;
	Aname *dot;
	int ic;

	name = amlenv->op->name;
	args = checkargs("n");
	nval = args[0]->sval;
	ic = amlenv->ic;
	switch(ic){
	case OpMethod:
		o = amlenv->res = amlnew(Omethod);
		if((amlenv->flags&Enoeval) == 0){
			amltcheck(args[1], Oint, 0);
			kstrdup(&o->method.name, nval);
			o->method.isexcl = args[1]->ival & 8;
			o->method.synclvl = (args[1]->ival>>4) & 0xF;
			o->method.nargs = args[1]->ival & 7;
			o->method.ps = amlenv->ps;
			o->method.pe = amlenv->argpe;
		}
		/* methods are not evaluated now */
		amlenv->ps = amlenv->argpe;
		if((amlenv->flags&Enoeval) == 0)
			amlbind(o->method.name, o);
		return;
	case OpScope:
		o = amlenv->res = amlnew(Oscope);
		if(amlenv->flags&Enoeval)
			break;
		o->scope.pe = amlenv->argpe;
		break;
	case OpDevice:
		o = amlenv->res = amlnew(Odev);
		o->dev.pe = amlenv->argpe;
		break;
	case OpPwrrsrc:
		o = amlenv->res = amlnew(Opwr);
		if(amlenv->flags&Enoeval)
			break;
		amltcheck(args[1], Oint, 0);
		amltcheck(args[2], Oint, 0);
		o->pwr.syslvl = args[1]->ival;
		o->pwr.rorder = args[2]->ival;
		break;
	case OpProcessor:
		o = amlenv->res = amlnew(Oproc);
		if(amlenv->flags&Enoeval)
			break;
		amltcheck(args[1], Oint, 0);
		amltcheck(args[2], Oint, 0);
		amltcheck(args[3], Oint, 0);
		o->proc.id = args[1]->ival;
		o->proc.baddr = args[2]->ival;
		o->proc.blen = args[3]->ival;
		break;
	case OpThermalzone:
		o = amlenv->res = amlnew(Othermal);
		break;
	default:
		SET(o);
		panic("opscope bug");
	}
	if((amlenv->flags&Enoeval) == 0){
		dot = amlwalk(nval, Walkcreating, nil);
		if(dot == nil)
			amlerror("%s: can't walk to %s", name, nval);
	}else
		dot = amlenv->dot;
	amlpush(&env2, amlenv->op->name, nil, amlenv->argpe);
	amlenv->dot = dot;
	if(ic == OpMethod)
		env2.flags |= Enoeval;	/* dump only */
	if(wasamlerror()){
		amlpop();
		amlerror(nil);
	}

	pamlblock(Dontkeep);


	if((amlenv->flags&Enoeval) == 0 && ic != OpScope){
		amlenv->dot = amlenv->prev->dot;
		amlbind(nval, o);
	}
	popamlerror();
	amlpop();
}

/*
 * tobuf: arg:B res:t ->b
 * todecstr: arg:D res:t -> S
 * tohexstr: arg:S res:t ->S
 * toint: arg:I res:t ->I
 * copyobject: arg:O res:t ->O
 * store: arg:O res:t ->O
 */
static void
opcast(void)
{
	Aobj **args;
	int docast;

	args = checkargs(amlenv->op->args);
	incref(args[0]);
	amlenv->res = args[0];
	docast = amlenv->ic != OpCopyobject;
	amlcpy(args[1], args[0], docast);
}

/*
 * arg:B len:I res:t -> S
 */
static void
optostr(void)
{
	Aobj *o, **args;
	int n;

	args = checkargs("BIt");
	n = args[1]->ival;
	if(n > args[0]->buf.len)
		n = args[0]->buf.len;
	o = amlenv->res = amlnew(Ostr);
	o->sval = amlmalloc(n + 1);
	if(n > 0)
		memmove(o->sval, args[0]->buf.p, n);
	o->sval[n] = 0; /* in case the buffer does to have a '\0' */
	amlcpy(args[2], o, Cast);
}

/*
 * inc: arg:I -> I
 * dec: arg:I -> I
 */
static void
opincdec(void)
{
	Aobj **args;

	args = checkargs("I");
	dxprint("%s %O\n", amlenv->op->name, args[0]);
	if(amlenv->ic == OpInc)
		args[0]->ival++;
	else
		args[0]->ival--;
	amlenv->res = amlnewi(args[0]->ival);
}

static int
objtype(Aobj *o)
{
	while(o != nil && (o->type == Oref || o->type == Oarg || o->type == Olocal))
		o = o->oref.o;

	if(o == nil)
		return Tnone;
	switch(o->type){
	case Oslice:	return Tslice;
	case Ofield:	return Tfield;
	case Omethod:	return Tmethod;
	case Odev:	return Tdev;
	case Oproc:	return Tproc;
	case Opwr:	return Tpwr;
	case Othermal:	return Tthermal;
	case Ohandle:	return Thandle;
	case Obuf:	return Tbuf;
	case Oevent:	return Tevent;
	case Oint:	return Tint;
	case Omutex:	return Tmutex;
	case Opkg:	return Tpkg;
	case Oreg:	return Treg;
	case Odecstr:
	case Ostr:
		return Tstr;
	case Oname:
		/* don't want to resolve the name */
		/* leave it so by now */
	default:
		return Tnone;
	}
}

static long
objsize(Aobj *o)
{
	switch(o->type){
	case Ostr:
	case Odecstr:
		return strlen(o->sval);
	case Obuf:
		return o->buf.len;
	case Opkg:
		return o->list.n;
	default:
		amlerror("not a buffer, string, or package");
	}
	return 0;
}

/*
 * objecttype: arg:O -> I
 * sizeof: arg:O -> I
 */
static void
opunaryobj(void)
{
	Aobj *o, **args;

	args = checkargs("O");
	o = amlenv->res = amlnew(Oint);
	switch(amlenv->ic){
	case OpObjecttype:
		o->ival = objtype(args[0]);
		break;
	case OpSizeof:
		o->ival = objsize(args[0]);
		break;
	default:
		amlerror("aml opunary bug");
	}
}

/*
 * term:I term:I -> I
 */
static void
oplandor(void)
{
	Aobj **args;

	args = checkargs("II");
	if(amlenv->ic == OpLand)
		amlenv->res = amlnewi(args[0]->ival && args[1]->ival);
	else
		amlenv->res = amlnewi(args[0]->ival || args[1]->ival);
}

/*
 * lnot: term:I -> I
 */
static void
oplnot(void)
{
	Aobj **args;

	args = checkargs("I");
	amlenv->res = amlnewi(args[0]->ival == 0);
}

/*
 * Inefficient, as everything else; and not compliant. but should suffice.
 * lequal, lgreater, lless: term:B term:B -> I
 */
static void
oplcmp(void)
{
	Aobj **args, *o, *arg0, *arg1;
	int len, r;
	int isint;

	args = checkargs("OO");
	arg1 = arg0 = nil;
	isint = args[0]->type == Oint && args[1]->type == Oint;
	if(wasamlerror()){
		amlclose(arg0);
		amlclose(arg1);
		amlerror(nil);
	}
	r = 0;
	if(!isint){
		arg0 = amlcast(args[0], Obuf);
		arg1 = amlcast(args[1], Obuf);
		len = arg0->buf.len;
		if(len > arg1->buf.len)
			len = arg1->buf.len;
		r = memcmp(arg0->buf.p, arg1->buf.p, len);
	}
	o = amlenv->res = amlnew(Oint);
	switch(amlenv->ic){
	case OpLequal:
		if(isint)
			o->ival = args[0]->ival == args[1]->ival;
		else
			o->ival = (r == 0 && arg0->buf.len == arg1->buf.len);
		break;
	case OpLgreater:
		if(isint)
			o->ival = args[0]->ival > args[1]->ival;
		else{
			o->ival = r > 0;
			if(r == 0 && arg0->buf.len > arg1->buf.len)
				o->ival = 1;
		}
		break;
	case OpLless:
		if(isint)
			o->ival = args[0]->ival < args[1]->ival;
		else{
			o->ival = r < 0;
			if(r == 0 && arg0->buf.len < arg1->buf.len)
				o->ival = 1;
		}
		break;
	default:
		panic("oplcmp bug");
	}
	popamlerror();
	amlclose(arg0);
	amlclose(arg1);
}

static int
msb(u64int n)
{
	int r;

	if(n == 0)
		return 0;
	for(r = 1; n != 1; r++)
		n >>= 1;
	return r;
}

static int
lsb(u64int n)
{
	int r;

	if(n == 0)
		return 0;
	for(r = 1; (n & 1) == 0; r++)
		n >>= 1;
	return r;
}

static uvlong
baseconv(uvlong v, int ibase, int obase)
{
	uvlong r, p;

	r = 0;
	p = 1;
	while(v != 0){
		r += (v % ibase) * p;
		v /= ibase;
		p *= obase;
	}
	return r;
}

static uvlong
frombcd(uvlong v)
{
	return baseconv(v, 16, 10);
}

static uvlong
tobcd(uvlong v)
{
	return baseconv(v, 10, 16);
}

/*
 * arg:I res:t -> I
 */
static void
opunaryt(void)
{
	Aobj *o, **args;

	args = checkargs("It");
	o = amlenv->res = amlnew(Oint);
	switch(amlenv->ic){
	case OpLeftsetbit:
		o->ival = msb(args[0]->ival);
		break;
	case OpRightsetbit:
		o->ival = lsb(args[0]->ival);
		break;
	case OpFrombcd:
		o->ival = frombcd(args[0]->ival);
		break;
	case OpTobcd:
		o->ival = tobcd(args[0]->ival);
		break;
	case OpNot:
		o->ival = ~args[0]->ival;
		break;
	default:
		panic("opunaryt bug");
	}
	amlcpy(args[1], o, Cast);
}

/*
 * term:I term:I res:t -> I
 */
static void
opbinaryt(void)
{
	Aobj *o, **args;

	args = checkargs("IIt");
	o = amlenv->res = amlnew(Oint);
	switch(amlenv->ic){
	case OpAdd:
		o->ival = args[0]->ival + args[1]->ival;
		break;
	case OpSub:
		o->ival = args[0]->ival - args[1]->ival;
		break;
	case OpMultiply:
		o->ival = args[0]->ival * args[1]->ival;
		break;
	case OpMod:
		if(args[1]->ival == 0)
			amlerror("divide by zero");
		o->ival = args[0]->ival % args[1]->ival;
		break;
	case OpAnd:
		o->ival = args[0]->ival & args[1]->ival;
		break;
	case OpOr:
		o->ival = args[0]->ival | args[1]->ival;
		break;
	case OpNor:
		o->ival = ~(args[0]->ival | args[1]->ival);
		break;
	case OpNand:
		o->ival = ~(args[0]->ival & args[1]->ival);
		break;
	case OpXor:
		o->ival = args[0]->ival ^ args[1]->ival;
		break;
	case OpShl:
		o->ival = args[0]->ival << args[1]->ival;
		break;
	case OpShr:
		o->ival = args[0]->ival >> args[1]->ival;
		break;
	default:
		panic("opbinaryt bug");
	}
	amlcpy(args[2], o, Cast);
}

/*
 * num:I den:I rem:t quot:t -> I
 */
static void
opdivide(void)
{
	Aobj *o, **args;

	args = checkargs("IItt");
	o = amlenv->res = amlnew(Oint);
	if(args[1]->ival == 0)
		amlerror("divide by zero");
	o->ival = args[0]->ival % args[1]->ival;
	amlcpy(args[2], o, Cast);
	o->ival = args[0]->ival / args[1]->ival;
	amlcpy(args[3], o, Cast);
}

/*
 * concat strings if arg1 is a string, or buffers otherwise.
 * arg1:O arg2:O res:t -> B|S
 */
static void
opconcat(void)
{
	Aobj **args, *o, *arg0, *arg1;

	args = checkargs("OOt");
	arg1 = nil;
	if(args[0]->type == Ostr || args[0]->type == Odecstr){
		incref(args[0]);
		arg0 = args[0];
	}else
		arg0 = amlcast(args[0], Obuf);
	if(wasamlerror()){
		amlclose(arg0);
		amlclose(arg1);
		amlerror(nil);
	}
	arg1 = amlcast(args[1], arg0->type);

	if(arg0->type == Obuf){
		o = amlenv->res = amlnew(Obuf);
		o->buf.len = arg0->buf.len + arg1->buf.len;
		o->buf.p = amlmalloc(o->buf.len);
		memmove(o->buf.p, arg0->buf.p, arg0->buf.len);
		memmove(o->buf.p + arg0->buf.len, arg1->buf.p, arg1->buf.len);
	}else{
		o = amlenv->res = amlnew(Ostr);
		o->sval = amlmalloc(strlen(arg0->sval) + strlen(arg1->sval) + 1);
		strcpy(o->sval, arg0->sval);
		strcat(o->sval, arg1->sval);
	}
	popamlerror();
	amlclose(arg0);
	amlclose(arg1);
	amlcpy(args[2], o, Cast);
}

/*
 * return ACPI resource length not counting the end tag, if any.
 */
static int
reslen(uchar *p, int n)
{
	int tot, len;

	for(tot = 0; tot < n; tot += len){
		if(p[tot] & 0x80)
			/* small resource */
			len = 1 + (p[tot] & 3);
		else if(p[tot] != ARendtag && tot + 3 <= n)
			/* large resource and not the end tag */
			len = 3 + l16get(&p[tot+1]);
		else
			break;
	}
	return tot;
}

/*
 * arg1:B arg2:B res:t -> B
 */
static void
opconcatres(void)
{
	Aobj **args, *o;
	int b0len, b1len;

	args = checkargs("BBt");
	b0len = reslen(args[0]->buf.p, args[0]->buf.len);
	b1len = reslen(args[0]->buf.p, args[0]->buf.len);

	o = amlenv->res = amlnew(Obuf);
	o->buf.len = b0len + b1len + 2;
	o->buf.p = amlmalloc(o->buf.len);
	memmove(o->buf.p, args[0]->buf.p, b0len);
	memmove(o->buf.p + b0len, args[1]->buf.p, b1len);
	o->buf.p[b0len + b1len] = ARendtag;

	o->buf.p[b0len + b1len + 1] = ARdontcksum;
	amlcpy(args[2], o, Cast);
}

/*
 * obj:n ref:t -> I
 */
static void
opcondrefof(void)
{
	Aobj **args, *no, *ro;

	args = checkargs("nt");
	if(wasamlerror()){
		if(isamlbranch())
			amlerror(nil);
		/* ignore errors and return false */
		amlenv->res = amlnewi(0);
		return;
	}

	no = amlread(args[0]);	/* check we can read it */
	popamlerror();

	ro = amlenv->res = amlnew(Oref);
	incref(no);
	ro->oref.o = no;

	amlcpy(args[1], ro, Nocast);
	amlclose(ro);
	amlenv->res = amlnewi(1);
}

/*
 * obj:o  -> O
 */
static void
oprefof(void)
{
	Aobj **args, *o;

	args = checkargs("o");
	o = amlenv->res = amlnew(Oref);
	o->oref.o = amlread(args[0]);
	incref(o->oref.o);
}

/*
 * arg:o -> O
 */
static void
opderefof(void)
{
	Aobj **args, *o;
	Aname *n;

	args = checkargs("o");
	o = amlread(args[0]);
	switch(o->type){
	case Ostr:
	case Odecstr:
		n = amlwalk(o->sval, Walkns, nil);
		if(n == nil)
			amlerror("name not found");
		o = n->o;
		break;
	case Oref:
		o = o->oref.o;
		break;
	}
	incref(o);
	amlenv->res = o;
}

/*
 * src:O idx:I len:I res:t -> B|S
 */
static void
opmid(void)
{
	Aobj **args, *o, *arg0;
	void *sp, *dp;
	int len, off, n;

	args = checkargs("OIIt");
	switch(args[0]->type){
	case Ostr:
	case Odecstr:
		arg0 = nil;
		sp = args[0]->sval;
		n = strlen(args[0]->sval);
		o = amlenv->res = amlnew(Ostr);
		break;
	default:
		arg0 = amlcast(args[0], Obuf);
		o = amlenv->res = amlnew(Obuf);
		sp = arg0->buf.p;
		n = arg0->buf.len;
		break;
	}
	off = args[1]->ival;
	len = args[2]->ival;
	if(off + len > n)
		len = n - off;
	if(len > 0){
		if(o->type == Obuf){
			dp = o->buf.p = amlmalloc(len);
			o->buf.len = len;
		}else{
			dp = o->sval = amlmalloc(len + 1);
			o->sval[len] = 0;
		}
		memmove(dp, sp, len);
	}
	amlclose(arg0);
	amlcpy(args[3], o, Cast);
}

static int
matches(int op, uvlong a1, uvlong a2)
{
	switch(op){
	case Mtrue:
		return 1;
	case Meq:
		return a1 == a2;
	case Mle:
		return a1 <= a2;
	case Mlt:
		return a1 < a2;
	case Mge:
		return a1 >= a2;
	case Mgt:
		return a1 > a2;
	default:
		amlerror("matched: unknown op %d", op);
	}
	return 0;
}

/*
 * starting at idx in src search for item i s.t.
 * src[i] op1 arg1 and src[i] op2 arg2.
 * Return i or -1.
 * [0]src:P [1]op1:b [2]arg1:I [3]op2:b [4]arg2:I [5]idx:I -> I
 */
static void
opmatch(void)
{
	Aobj **args, *o, *po, *io;
	int i, r;

	args = checkargs("PbIbII");
	o = amlenv->res = amlnewi(-1);

	po = args[0]->list.hd;
	for(i = 0; po != nil && i < args[5]->ival; i++)
		po = po->next;
	for(r = 0; r == 0 && po != nil; po = po->next){
		io = amlcast(po, Oint);
		r = matches(args[1]->ival, io->ival, args[2]->ival) &&
		   matches(args[3]->ival, io->ival, args[4]->ival);
		amlclose(io);
		if(r != 0)
			o->ival = i;
		else
			i++;
	}
}

/*
 * search XSDT for sig, oemid and oemtblid
 * and load it using root scope.
 * Then copy paramv into parm and return dbhandle or 0
 * [0]sig:S [1]oemid:S [2]oemtblid:S [3]root:S [4]param:S [5]paramv:B -> I
 */
static void
oploadtable(void)
{
	amlwarn("load table not implemented");
	/*
	 * found = acpixsdt(args[0]->sval, args[1]->sval, args[2]->sval);
	 * t = findtable(args[0]->sval, args[1]->sval, args[2]->sval);
	 * create env for acpins and use it to eval t->data.
	 * copyobj(args[4], args[5], Cast);
	 * return handle obj with t->id;
	 */
	amlenv->res = amlnewi(0);
}

/*
 * load SSDT found in region and store its handle.
 * reg:O handle:t
 */
static void
opload(void)
{
	Aobj **args, *o;
	Nreg *r;
	long off, len;

	off = len = 0;
	r = nil;
	args = checkargs("Ot");
	switch(args[0]->type){
	case Oreg:
		r = &args[0]->reg;
		len = r->len;
		break;
	case Ofield:
		if(args[0]->field.reg == nil)
			amlerror("null region");
		r = &args[0]->field.reg->reg;
		off = args[0]->field.off >> 3;
		len = args[0]->field.len >> 3;
		break;
	default:
		amlerror("load requires region or field");
	}
	if(r->spc != Rsysmem)
		amlerror("can load only from system memory");
	if(r->p == nil)
		r->p = vmap(r->base, r->len);
	if(off + len > r->len)
		len = r->len - off;
	if(len < Sdthdrsz)
		amlerror("region is too small for load");
	if(acpidsdt(r->p + off, len) < 0)
		amlerror("table load failed");
	
	o = amlenv->res = amlnew(Ohandle);
	o->ival = -1; // (uintptr)t;
	amlcpy(args[1], o, Cast);
	amlerror("load not implemented");
}

/*
 * unload SSDT named by handle.
 * handle:I
 */
static void
opunload(void)
{
	Aobj **args;

	args = checkargs("I");
	dxprint("unload %O\n", args[0]);
//	closetable((Atable*)args[0]->ival);
}

/*
 * named memory region for an ACPI table.
 * name:n sig:S oem:S tbl:S
 */
static void
opdataregion(void)
{
	Aobj **args, *o;

	args = checkargs("nSSS");
	o = amlenv->res = amlnew(Oreg);
	/*
	 * t = findtable(args[1]->sval, args[2]->sval, args[3]->sval);
	 * if(t == nil)
	 *	amlerror("no such table");
	 * o->reg.spc = Rsysmem;
	 * o->reg.p = (uintptr)t->tbl;
	 * o->reg.len = t->tbl->len;
	 */
	amlwarn("dataregion called but not really implemented");
	amlbind(args[0]->sval, o);
}

static void
opnop(void)
{
}

/*
 * n:I
 */
static void
opsleep(void)
{
	Aobj **args;

	args = checkargs("I");

	dxprint("delay %lld\n", args[0]->ival);
	switch(amlenv->ic){
	case OpSleep:
		/* should use tsleep if up */
		delay(args[0]->ival);		/* ms */
		break;
	case OpStall:
		/* can't use tsleep */
		microdelay(args[0]->ival);	/* µs */
		break;
	default:
		panic("opsleep bug");
	}
}

/*
 * obj:O ev:I
 */
static void
opnotify(void)
{
	Aobj **args;
	int ev;

	args = checkargs("OI");
	dxprint("notify %O %O\n", args[0], args[1]);
	ev = args[1]->ival;
	switch(args[0]->type){
	case Odev:
		args[0]->dev.handler(args[0], ev);
		break;
	case Oproc:
		args[0]->proc.handler(args[0], ev);
		break;
	case Othermal:
		args[0]->thermal.handler(args[0], ev);
		break;
	}
}

/*
 * The same invocation may re-acquire a mutex
 * it holds.The mutex can be the global lock.
 *
 * As long as we use a single proc to run the
 * interpreter, we can assume we can always acquire locks.
 * The first call for the global lock acquires it.
 * The last one releasing releases it.
 */

/*
 * mutex:M tmout:w -> I
 */
static void
opacquire(void)
{
	Aobj **args;

	args = checkargs("Mw");
	dxprint("acquire %O\n", args[0]);
	args[0]->mutex.count++;
	if(args[0]->mutex.count == 1 && args[0] == acpilock)
		acpiglock();
	amlenv->res = amlnewi(0); /* 0 == ok */
}

/*
 * mutex:M
 */
static void
oprelease(void)
{
	Aobj **args;

	args = checkargs("M");
	dxprint("release %O\n", args[0]);
	args[0]->mutex.count--;
	if(args[0]->mutex.count == 0 && args[0] == acpilock)
		acpigunlock();
}

/*
 * event:E
 */
static void
opreset(void)
{
	Aobj **args;

	args = checkargs("E");
	dxprint("reset %O\n", args[0]);
	args[0]->event.fire = 0;
}

/*
 * event:E
 */
static void
opsignal(void)
{
	Aobj **args;

	args = checkargs("E");
	dxprint("signal %O\n", args[0]);
	args[0]->event.fire++;
}

/*
 * event:E tmout:I -> I
 */
static void
opwait(void)
{
	Aobj **args;

	args = checkargs("E");
	dxprint("wait %O\n", args[0]);
	amlenv->res = amlnewi(0);
	do{
		if(args[0]->event.fire)
			return;
		/* should wait for the event if up.
		 * There may be many invocations calling wait
		 * on the same event.
		 */
	}while(0);
	amlenv->res->ival = 1;
}

/*
 * type:b code:d arg:I
 */
static void
opfatal(void)
{
	Aobj **args;

	args = checkargs("bdI");
	xprint("acpi: fatal(%lld, %lld, %lld)\n",
		args[0]->ival, args[1]->ival, args[2]->ival);
	amlerror("fatal");
}

/*
 * We assume that a method won't bind a child and then its parent.
 */
static void
unbindnames(Aname *n)
{
	Aname *nn;

	for(; n != nil; n = nn){
		nn = n->mnext;
		amlunbind(n);
	}
}

/*
 * call a method. used by opnamechar and amlcast.
 * If uargs is nil, args are parsed from aml code at current ps.
 */
Aobj*
opcall(Aobj *o, Aobj **uargs)
{
	Aobj *r;
	Mframe mfr;
	Env env2;
	int i, nargs;
	uchar *oldps;

	if(o->type != Omethod)
		panic("aml: call on non method");
	nargs = o->method.nargs;
	memset(&mfr, 0, sizeof(mfr));

	/*
	 * 1. parse and evaluate arguments in caller's environment.
	 * Keep them in targs so that stack schecks may find
	 * those objects in the stack and errors raised know they
	 * must be cleaned up. They are unused otherwise.
	 */
	amlpush(&env2, "call", nil, nil);
	env2.dot = o->name;
	if(uargs == nil && nargs > 0){
		/* record temp. args, for stack checks and errors */
		env2.mp->targs = mfr.args;
		if(wasamlerror()){
			if(env2.mp->targs != nil)
				for(i = 0; i < nargs; i++){
					amlclose(mfr.args[i]);
					mfr.args[i] = nil;
				}
			amlpop();
			amlerror(nil);
		}
		diprint("%!call \"%s\"(\n", o->name->path);
		for(i = 0; i < nargs; i++){
			mfr.args[i] = amlcast(pamlobj(), Oany);
			diprint("%!marg%d: %O\n", i, mfr.args[i]);
			amlclose(amlenv->res);
			amlenv->res = nil;
		}
		popamlerror();
		diprint("%!)\n");
		env2.mp->targs = nil;	/* they are in mp->args now. */
	}else{
		for(i = 0; i < nargs && uargs[i] != nil; i++)
			mfr.args[i] = uargs[i];
		nargs = i;
		for(; i < nelem(mfr.args); i++)
			mfr.args[i] = nil;
	}
	mfr.nargs = nargs;
	env2.mp = &mfr;	/* It's official now */

	if(amlenv->flags&Enoeval){
		amlpop();
		incref(o);
		return o;
	}

	/*
	 * 2. Call the method.
	 */
	if(o->method.trace > 0)
		amlenv->flags |= Exdump|Eidump;
	dxprint("call %O ps %#p\n", o, o->method.ps);
	if(o->method.trace < 0)
		amlenv->flags &= ~(Exdump|Eidump);
	oldps = amlenv->ps;
	r = nil;
	if(wasamlerror()){
		assert(amlenv->mp == &mfr);
		if(strncmp(amlerrstk.msg, "return at", 9) == 0){
			r = amlenv->res;
			amlenv->res = nil;
			goto Done;
		}
		unbindnames(mfr.binds);
		amlpop();
		amlenv->ps = oldps;	/* amlpop fails to set ps */
		amlerror(nil);
	}
	env2.ps = o->method.ps;
	env2.pe = o->method.pe;
	if(o->method.f != nil)
		o->method.f(o);
	else
		pamlblock(Dontkeep);
	popamlerror();
Done:
	unbindnames(mfr.binds);
	amlpop();
	amlenv->ps = oldps;	/* amlpop fails to set ps */
	amlclose(amlenv->res);
	amlenv->res = nil;
	if(r != nil)
		amlenv->res = r;
	dxprint("ret %s: %O\n", o->method.name, r);
	return amlenv->res;
}

/*
 * arg:O
 */
static void
opreturn(void)
{
	Aobj **args;
	Env *e;

	if((amlenv->flags&Enoeval) == 0){
		args = checkargs("O");
		/* leave the result in the environment that made the call */
		for(e = amlenv; e->prev != nil && e->prev->mp == e->mp; e = e->prev)
			;
		if(e->mp == nil)
			amlerror("return not in method call");
		amlclose(e->res);
		incref(args[0]);
		e->res = args[0];
		amlerror("return");	/* catched by opcall() */
	}
}

/*
 * Both if and while.
 * len cond:I ...
 */
static void
opif(void)
{
	Env env2, *env;
	int loop;
	uchar *p0, *pe;
	char *name;

	checkargs("I");
	env = amlenv;

	loop = env->ic == OpWhile;
	name = "if";
	if(loop)
		name = "while";
	p0 = env->opp0;
	pe = env->argpe;
	if((env->flags&Enoeval) == 0){
		if(env->args[0]->ival != 0)
			env->flags |= Ewastrue;
		else
			env->flags &= ~Ewastrue;
		if((env->flags&Ewastrue) == 0){
			env->ps = pe;	/* do not evaluate anything now */
			return;
		}
	}
	amlpush(&env2, name, nil, pe);
	if(wasamlerror()){
		amlpop();
		if(isamlbranch()){
			if(loop == 0 || strncmp(amlerrstk.msg, "return", 6) == 0)
				amlerror(nil);
			if(strncmp(amlerrstk.msg, "continue", 8) == 0)
				goto Next;
			/* break the loop */
			xprint("%s: error: %s\n", name, amlerrstk.msg);
		}
		return;
	}
	pamlblock(Dontkeep);
	amlpop();
	popamlerror();
Next:
	if((env->flags&Enoeval) == 0 && loop != 0){
		dxprint("loop\n");
		env->ps = p0;	/* pretend we are unparsed: loop */
	}
}

/*
 * len ...
 */
static void
opelse(void)
{
	Env env2, *env;

	env = amlenv;
	if((env->flags&Enoeval) == 0 && (env->flags&Ewastrue) != 0){
		env->ps = env->argpe;	/* do not evaluate now */
		return;
	}
	amlpush(&env2, "elsescope", nil, env->argpe);
	if(wasamlerror()){
		amlpop();
		amlerror(nil);
	}
	env->res = pamlblock(Dontkeep);
	amlpop();
	popamlerror();
}

/*
 * continue and break raise their own names, both cached by opif()
 * continue continues the innermost while (opif).
 * break breaks the innermost while (opif);
 */
static void
opjump(void)
{
	if((amlenv->flags&Enoeval) == 0){
		dxprint("%s\n", amlenv->op->name);
		amlerror(amlenv->op->name);
	}
}


/*
 * Argument processing tools.
 *
 * Parsing functions below are delicate regarding
 * ref counting and env handling. That's the price for
 * keeping ops above more simple in that respect.
 */

static int
ppkglen(void)
{
	int len;

	len = amlenv->ps[0];

	switch(len & 0xC0){
	case 0x00:
		break;
	case 0x40:
		len &= 0xF;
		len |= amlnext() << 4;
		break;
	case 0x80:
		len &= 0xF;
		len |= amlnext() << 4;
		len |= amlnext() << 12;
		break;
	case 0xC0:
		len &= 0xF;
		len |= amlnext() << 4;
		len |= amlnext() << 12;
		len |= amlnext() << 20;
		break;
	default:
		amlerror("bad pkglen %#ux\n", len);
	}
	amlnext();
	return len;
		
}

/*
 * Parse a field list argument and return Opkg with Ofields
 */
static Aobj*
pfieldlist(void)
{
	Aobj *lo;
	Aobj **ol;
	Aobj *fo;
	Aobj *n;
	Nfield f;
	int off;

	lo = amlnew(Opkg);
	ol = &lo->list.hd;
	if(wasamlerror()){
		amlclose(lo);
		amlerror(nil);
	}
	off = 0;
	while(amlenv->ps < amlenv->pe){
		memset(&f, 0, sizeof(f));
		switch(amlenv->ps[0]){
		case 0:				/* reserved field */
			amlnext();
			off += ppkglen();
			diprint("%#!off = %#x\n", off/8);
			break;
		case 1:				/* access field */
			amlnext();
			amlnext();
			amlnext();
			break;
		default:
			amlclose(amlenv->res);
			amlenv->res = nil;
			n = pnamechar();
			if(n == nil)
				amlerror("bad name");
			amlenv->res = nil;
			f.off = off;
			kstrdup(&f.fname, n->sval);
			amlclose(n);
			f.len = ppkglen();
			fo = amlnew(Ofield);
			fo->field = f;
			*ol = fo;
			ol = &fo->next;
			off += f.len;
			diprint("%#!field '%s' %d\n", f.fname, f.len);
		}
	}
	popamlerror();
	return amlenv->res = lo;
}

static int
isintop(Op *op)
{
	return op->x == opbyteprefix || op->x == opwordprefix ||
		op->x == opdwordprefix || op->x == opqwordprefix;
}

static int
isstrop(Op *op)
{
	return op->x == opstrprefix || op->x == opnamechar;
}

/*
 * Decode arguments according to fmt.
 * See aml.h:/Pargs types in aml.h.
 */
static uchar*
pargs(void)
{
	uchar *newpe, *oldpe;
	Aobj *o, **args;
	char *fmt;
	Env env2;
	int len, flags, hasbody, hasargs;

	fmt = amlenv->op->args;
	if(fmt[0] == 0){		/* no args; optimize */
		if(!isstrop(amlenv->op))
			diprint("\n");
		return amlenv->pe;
	}
	args = amlenv->args;
	hasbody = amlenv->op->flags&Bdy;
	oldpe = nil;

	/*
	 * 'l' limits the length of the program for the following
	 * items. We must enforce the limit and restore env later.
	 */
	newpe = amlenv->pe;
	if(fmt[0] == 'l'){
		newpe = amlenv->ps;
		len = ppkglen();
		newpe += len;
		if(newpe > amlenv->pe){
			amlwarn("pargs: overlap (truncating)\n");
			newpe = amlenv->pe;
		}else
			oldpe = amlenv->pe;
		fmt++;
		diprint("[%#x]", len);
	}
	hasargs = fmt[0] != 0;
	if(hasargs)
		diprint("(\n");
	else
		diprint("\n");
	amlenv->pe = newpe;
	flags = amlenv->flags;
	if(wasamlerror()){
		amlenv->flags = flags;
		if(oldpe != nil)
			amlenv->pe = oldpe;
		amlerror(nil);
	}
	for(; fmt[0] != 0 && amlenv->ps < amlenv->pe; fmt++){
		amlclose(amlenv->res);
		amlenv->res = nil;
		switch(fmt[0]){
		case 'b':
			o = pbyte();
			break;
		case 'w':
			o = pword();
			break;
		case 'd':
			o = pdword();
			break;
		case 'f':
			o = pfieldlist();
			break;
		case 'n':
			amlenv->flags |= Enocall;
			o = pnamechar();
			amlenv->flags = flags;
			if(fmt[0] == 's' && (o == nil || o->type != Oname))
				amlerror("bad argument for '%c'", fmt[0]);
			break;
		case 't':
			amlenv->flags |= Enocall;
			if(fmt[1] != 0 && fmt[1] != 't')
				amlerror("pargs bug: 't' is not last");
			if(amlenv->ps[0] == OpZero){
				o = amlenv->res = nil;	/* null target */
				amlnext();		/* not an int */
				break;
			}
			/* else fall */
		case 'B':
		case 'D':
		case 'E':
		case 'F':
		case 'I':
		case 'M':
		case 'O':
		case 'o':
		case 'P':
		case 'R':
		case 'S':
		case 's':
			/*
			 * pamlobj may evaluate ops with args
			 * we must save the current state
			 * and leave the resulting object at amlenv->res
			 * where it belongs.
			 */
			amlpush(&env2, "pargs", nil, newpe);
			if(wasamlerror()){
				amlpop();
				amlerror(nil);
			}
			o = pamlobj();
			popamlerror();
			amlenv->res = nil;	/* env2's res */
			amlpop();
			amlenv->res = o;	/* env's res */
			amlenv->flags = flags;
			break;
		case 'l':
			amlerror("pargs bug: 'l' is not first");
			return nil;
		default:
			amlerror("pargs bug: bad arg fmt %c", fmt[0]);
			return nil;
		}
		diprint("%#!arg%d/%c %O\n", amlenv->nargs, fmt[0], o);

		if((amlenv->flags&Enoeval) == 0 && o == nil && fmt[0] != 't')
			amlerror("null argument for '%c'", fmt[0]);
		if(amlenv->nargs == nelem(amlenv->args))
			amlerror("pargs: increase Nargs");

		if((amlenv->flags&Enoeval) == 0 && fmt[0] >= 'A' && fmt[0] <= 'Z')
			args[amlenv->nargs++] = amlcast(o, fmt[0]);
		else{
			args[amlenv->nargs++] = o;
			if(o != nil)
				incref(o);
		}
		amlclose(o);
		amlenv->res = nil;

	}
	if(fmt[0] != 0)
		amlerror("argument expected");
	if(oldpe != nil)
		amlenv->pe = oldpe;
	popamlerror();

	if(hasargs || hasbody){
		diprint("%!");
		if(hasargs)
			diprint(")");
		if(hasbody)
			diprint("{");
		diprint("\n");
	}
	return newpe;
}

/*
 * Pure paranoia. Double check that we got args ok.
 * They should be, but bugs in arg. processing may show up
 * and we might make a mistake in the op tables below.
 */
static Aobj**
checkargs(char *s)
{
	Aobj **args;

	if(amlenv->flags&Enoeval)
		return amlenv->args;

	args = amlenv->args;
	for(; s != nil && s[0] != 0; s++){
		if(*args == nil)
			if(s[0] == 't')
				break;
			else
				amlerror("checkargs: argument expected");
		switch(s[0]){
		case 'B':
			amltcheck(*args, Obuf, 1);
			break;
		case 'b':
		case 'w':
		case 'd':
		case 'q':
		case 'I':
			amltcheck(*args, Oint, 1);
			break;
		case 'D':
			amltcheck(*args, Odecstr, 0);
			break;
		case 'E':
			amltcheck(*args, Oevent, 1);
			break;
		case 'F':
			amltcheck(*args, Ofield, 1);
			break;
		case 'M':
			amltcheck(*args, Omutex, 1);
			break;
		case 'n':
		case 'N':
			amltcheck(*args, Oname, 0);
			break;
		case 'O':
		case 'o':
			break;
		case 'l':
		case 'L':
		case 'f':
		case 'P':
			amltcheck(*args, Opkg, 1);
			break;
		case 'R':
			amltcheck(*args, Oreg, 1);
			break;
		case 's':
		case 'S':
			amltcheck(*args, Ostr, 0);
			break;
		case 't':
			if((*args)->type != Oname && (*args)->type != Olocal &&
			   (*args)->type != Oarg)
				amlerror("checkargs: %O is not a target", *args);
			break;
		default:
			panic("aml: checkargs: unknown argument letter %c", s[0]);
		}
		args++;
	}
	return amlenv->args;
}

static void
fetch(void)
{
	Env *env;
	int ic;
	uchar *ps;

	env = amlenv;
	ps = env->ps;
	env->op = decode(env->ps, env->pe, &ic);
	env->ic = ic;
	if(env->op == nil)
		amlerror("unknown instr %#04x at %#p", env->ic, ps);
	if(env->ps[0] == OpExtprefix && env->ps < env->pe)
		amlnext();
	if(!isintop(env->op) && !isstrop(env->op))
		diprint("%!%s", env->op->name);
	env->op->ncalls++;
}

/*
 * Process aml for a single object and return it.
 */
static Aobj*
pamlobj(void)
{
	Aobj *o;
	Op *op;
	int i;
	int ic;

	amlenv->opp0 = amlenv->ps;
	fetch();
	amlclose(amlenv->res);
	amlenv->res = nil;
	for(i = 0; i < amlenv->nargs; i++){
		amlclose(amlenv->args[i]);
		amlenv->args[i] = nil;
	}
	amlenv->nargs = 0;
	op = amlenv->op;
	ic = amlenv->ic;
	if(op->args != nil){
		amlnext();
		amlenv->argpe = pargs();
	}else{
		amlenv->argpe = amlenv->pe;
		if(!isintop(op) && !isstrop(op))
			diprint("\n");
	}
	amlenv->ic = ic;	/* pargs may have changed it */

	if((amlenv->flags&Enoeval) == 0 || (op->flags&Dry) != 0)
		op->x();
	o = amlenv->res;

	if((op->flags&Bdy) != 0)
		diprint("%!}\n");
	if(isintop(op) && o != nil)
		diprint("%!%s %#ullx\n", amlenv->op->name, o->ival);

	if(amlenv->flags&Eedump)
		amlenvdump(amlenv);

	if(amlenv->flags&Escheck)
		amlstackcheck();

	return o;
}

/*
 * Process all objects in [ps:pe] and return a list with them.
 * or throw them away after evaluating if keep == 0.
 */
Aobj*
pamlblock(int keep)
{
	Aobj **ol, *o;
	int i;

	assert(amlenv->olist.hd == nil);
	ol = &amlenv->olist.hd;
	amlenv->olist.n = 0;
	for(i = 0; amlenv->ps < amlenv->pe && i < 1000; i++){
		pamlobj();
		o = amlenv->res;
		if(o != nil)
			if(keep){
				*ol = o;
				if(o->next != nil)
					panic("pamlblock: non-null next: %O", *ol);
				ol = &o->next;
				amlenv->olist.n++;
			}else
				amlclose(o);
		amlenv->res = nil;
	}
	if(i == 1000)
		panic("pamlblock: loop or huge pamlblock?");
	return amlenv->olist.hd;
}


/*
 * Plain (1-byte) operations.
 * The argument string indicates which arguments are expected
 * for each op.
 * A nil as argument str. means that no arg processing is done
 * (env->ps points to the operation, which must call amlnext()
 * to advance the pc).
 * An empty ("") argument str. means that there is no argument,
 * but argument processing will happen and advance the PC
 * to point past the last argument (i.e., past the op code).
 * A non-empty argument str. means that pargs will attempt to
 * build objects for arguments as explained in aml.h:/Pargs.
 * See also the comment at the start of amlrt.c.
 */
static Op amlops[256] =
{
	[OpZero]	{opnum,		"zero",		Dry, ""},
	[OpOne]		{opnum,		"one",		Dry, ""},
	[OpAlias]	{opalias,	"alias",		0, "On"},
	[OpName]	{opname,	"name",		0, "nO"},
	[OpByteprefix]	{opbyteprefix,	"byteprefix",	Dry, nil},
	[OpWordprefix]	{opwordprefix,	"wordprefix",	Dry, nil},
	[OpDwordprefix]	{opdwordprefix,	"dwordprefix",	Dry, nil},
	[OpStrprefix]	{opstrprefix,	"strprefix",	Dry, ""},
	[OpQwordprefix]	{opqwordprefix,	"qwordprefix",	Dry, nil},
	[OpScope]	{opscope,	"scope",	Bdy|Dry, "ln"},
	[OpBuffer]	{opbuf,		"buf",		Dry, "lI"},
	[OpPackage]	{oppackage,	"package",	Bdy|Dry, "lb"},
	[OpVarpackage]	{oppackage,	"varpackage",	Bdy|Dry, "lI"},
	[OpMethod]	{opscope,	"method",	Dry, "lnb"},
	[OpDnameprefix]	{opnamechar,	"dnameprefix",	Dry, nil},
	[OpMnameprefix]	{opnamechar,	"mnameprefix",	Dry, nil},
	[OpExtprefix]	{nil,		"extprefix",	Dry, nil},
	[OpRootchar]	{opnamechar,	"rootchar",	Dry, nil},
	[OpParentprefix]	{opnamechar,	"parentprefix",	Dry, nil},
	[OpNamechar]	{opnamechar,	"namechar",	Dry, nil},
	[OpLocal0]	{oplocal,	"local0",	Dry, nil},
	[OpLocal1]	{oplocal,	"local1",	Dry, nil},
	[OpLocal2]	{oplocal,	"local2",	Dry, nil},
	[OpLocal3]	{oplocal,	"local3",	Dry, nil},
	[OpLocal4]	{oplocal,	"local4",	Dry, nil},
	[OpLocal5]	{oplocal,	"local5",	Dry, nil},
	[OpLocal6]	{oplocal,	"local6",	Dry, nil},
	[OpLocal7]	{oplocal,	"local7",	Dry, nil},
	[OpArg0]	{oparg,		"arg0",		Dry, nil},
	[OpArg1]	{oparg,		"arg1",		Dry, nil},
	[OpArg2]	{oparg,		"arg2",		Dry, nil},
	[OpArg3]	{oparg,		"arg3",		Dry, nil},
	[OpArg4]	{oparg,		"arg4",		Dry, nil},
	[OpArg5]	{oparg,		"arg5",		Dry, nil},
	[OpArg6]	{oparg,		"arg6",		Dry, nil},
	[OpStore]	{opcast,	"store",	0, "Ot"},
	[OpRefof]	{oprefof,	"refof",	0, "o"},
	[OpAdd]		{opbinaryt,	"add",		0, "IIt"},
	[OpConcat]	{opconcat,	"concat",	0, "OOt"},
	[OpSub]		{opbinaryt,	"sub",		0, "IIt"},
	[OpInc]		{opincdec,	"inc",		0, "I"},
	[OpDec]		{opincdec,	"dec",		0, "I"},
	[OpMultiply]	{opbinaryt,	"multiply",	0, "IIt"},
	[OpDivide]	{opdivide,	"divide",	0, "IItt"},
	[OpShl]		{opbinaryt,	"shl",		0, "IIt"},
	[OpShr]		{opbinaryt,	"shr",		0, "IIt"},
	[OpAnd]		{opbinaryt,	"and",		0, "IIt"},
	[OpNand]	{opbinaryt,	"nand",		0, "IIt"},
	[OpOr]		{opbinaryt,	"or",		0, "IIt"},
	[OpNor]		{opbinaryt,	"nor",		0, "IIt"},
	[OpXor]		{opbinaryt,	"xor",		0, "IIt"},
	[OpNot]		{opunaryt,	"not",		0, "It"},
	[OpLeftsetbit]	{opunaryt,	"leftsetbit",	0, "It"},
	[OpRightsetbit]	{opunaryt,	"rightsetbit",	0, "It"},
	[OpDerefof]	{opderefof,	"derefof",	0, "o"},
	[OpConcatres]	{opconcatres,	"concatres",	0, "BBt"},
	[OpMod]		{opbinaryt,	"mod",		0, "IIt"},
	[OpNotify]	{opnotify,	"notify",	0, "OI"},
	[OpSizeof]	{opunaryobj,	"sizeof",	0, "O"},
	[OpIndex]	{opindex,	"index",		0, "OIt"},
	[OpMatch]	{opmatch,	"match",	0, "PbIbII"},
	[OpMkdwordfield]	{opmkxfield,	"mkdwordfield",	0, "BIn"},
	[OpMkwordfield]	{opmkxfield,	"mkwordfield", 	0, "BIn"},
	[OpMkbytefield]	{opmkxfield,	"mkbytefield", 	0, "BIn"},
	[OpMkbitfield]	{opmkxfield,	"mkbitfield", 	0, "BIn"},
	[OpObjecttype]	{opunaryobj,	"objecttype",	0, "O"},
	[OpMkqwordfield]	{opmkxfield,	"mkqwordfield",	0, "BIn"},
	[OpLand]	{oplandor,	"land",		0, "II"},
	[OpLor]		{oplandor,	"lor",		0, "II"},
	[OpLnot]	{oplnot,	"lnot",		0, "I"},
	[OpLequal]	{oplcmp,	"lequal",	0, "OO"},
	[OpLgreater]	{oplcmp,	"lgreater",	0, "OO"},
	[OpLless]	{oplcmp,	"lless",	0, "OO"},
	[OpTobuf]	{opcast,	"tobuf",	0, "Bt"},
	[OpTodecstr]	{opcast,	"todecstr",	0, "Dt"},
	[OpTohexstr]	{opcast,	"tohexstr",	0, "St"},
	[OpToint]	{opcast,	"toint",		0, "It"},
	[OpTostr]	{optostr,	"tostr",	0, "BIt"},
	[OpCopyobject]	{opcast,	"copyobject",	0, "On"},
	[OpMid]		{opmid,		"mid",		0, "OIIt"},
	[OpContinue]	{opjump,	"continue",	0, ""},
	[OpIf]		{opif,		"if",		Bdy|Dry, "lI"},
	[OpElse]	{opelse,	"else",		Bdy|Dry, "l"},
	[OpWhile]	{opif,		"while",		Bdy|Dry,"lI"},
	[OpNop]		{opnop,		"nop",		0, ""},
	[OpReturn]	{opreturn,	"return",	0, "O"},
	[OpBreak]	{opjump,	"break",	0, ""},
	[OpBreakpoint]	{opnop,		"breakpoint",	0, ""},
	[OpOnes]	{opnum,		"ones",		Dry, ""},
};

/*
 * extended operation, found after OpExtprefix.
 * codes may be found also in the plain (non-extended) operation
 * table. But env->op will always point to the correct operation.
 * When env->ic is looked at to switch among several codes,
 * context makes it clear which one is the operation at hand.
 */
static Xop amlxops[] = {
	{OpDevice,	opscope,	"device",	Dry, "ln"},
	{OpDataregion,	opdataregion,	"dataregion",	0, "nSSS"},
	{OpField,	opfield,		"field",		0, "lRbf"},
	{OpBankfield,	opbankfield,	"bankfield",	0, "lRnIbf"},
	{OpIndexfield,	opindexfield,	"indexfield",	0, "lFFbf"},
	{OpCondrefof,	opcondrefof,	"condrefof",	0, "nt"},
	{OpMutex,	opmutex,	"mutex",	0, "nb"},
	{OpEvent,	opevent,	"event",	0, "n"},
	{OpMkfield,	opmkxfield,	"mkfield",	0, "BIIn"},
	{OpOpregion,	opopregion,	"opregion",	0, "nbII"},
	{OpLoadtable,	oploadtable,	"loadtable",	0, "SSSSSB"},
	{OpLoad,	opload,		"load",		0, "Ot"},
	{OpStall,	opsleep,	"stall",	0, "I"},
	{OpSleep,	opsleep,	"sleep",	0, "I"},	
	{OpAcquire,	opacquire,	"acquire",	0, "Mw"},
	{OpSignal,	opsignal,	"signal",	0, "E"},
	{OpWait,		opwait,		"wait",		0, "EI"},
	{OpReset,	opreset,	"reset",	0, "E"},
	{OpRelease,	oprelease,	"release",	0, "M"},
	{OpFrombcd,	opunaryt,	"frombcd",	0, "It"},
	{OpTobcd,	opunaryt,	"tobcd",	0, "It"},
	{OpUnload,	opunload,	"unload",	0, "I"},
	{OpRevision,	opnum,		"revision",	0, ""},
	{OpDebug,	opnum,		"debug",	0, ""},
	{OpFatal,	opfatal,	"fatal",	0, "bdI"},
	{OpTimer,	opnum,		"timer",		0, ""},
	{OpProcessor,	opscope,	"processor",	Dry, "lnbdb"},
	{OpPwrrsrc,	opscope,	"pwrrsrc",	Dry, "lnbw"},
	{OpThermalzone,	opscope,	"thermalzone",	Dry, "ln"},
	{0, nil, nil, nil},
};

void
initamlops(void)
{
	int i;
	for(i = 'A'; i <= 'Z'; i++){
		assert(amlops[i].name == nil);
		amlops[i].name = "namechar";
		amlops[i].x = opnamechar;
		amlops[i].flags = Dry;
	}
}

static Op*
decode(uchar* ps, uchar *pe, int *icp)
{
	Op *op;
	Xop *xop;
	int ic;

	assert(ps < pe);
	*icp = ic = ps[0];
	if(ic == OpExtprefix){
		if(ps >= pe-1)
			return nil;
		*icp = ic = ps[1];
		for(xop = amlxops; xop->ic != 0 && xop->ic != ic; xop++)
			;
		if(xop->ic == 0)
			return nil;
		op = &xop->op;
	}else
		op = &amlops[ic];
	if(op == nil || op->x == nil)
		return nil;
	else
		return op;
}

void
dumpopstats(void)
{
	Xop *xop;
	int i;

	dprint("\ncalls:\n");
	for(i = 0; i < 256; i++)
		if(amlops[i].x != nil && amlops[i].ncalls != 0)
			dprint("%3d %s\n",
				amlops[i].ncalls, amlops[i].name);
	for(xop = amlxops; xop->ic != 0; xop++)
		if(xop->op.ncalls != 0)
			dprint("%3d %s\n",xop->op.ncalls,xop->op.name);
}
