/*
 * Advanced Configuration and Power Interface Specification 4.0.
 * Abstract Machine Language (AML) interpreter.
 *
 * This file provides run time support and configuration routines,
 * and contains the main entry point for the interpreter.
 *
 * See the aml.ms document before reading the implementation.
 */

#include "kernel.h"
#include "acpi.h"
#include "aml.h"

Aname *acpins;		/* namespace */
Env *amlenv;		/* last (most nested) AML env */
Error amlerrstk;	/* our error stack */
Aobj* acpilock;
int amlflags;		/* global debug flags when not evaluating */

static int naobj;	/* number of allocated objects */
static int maxnaobj;	/* max. number of allocated objects */
static int nanames;	/* number of names allocated */
static int maxdepth;
static uchar* minsp;
static Aobj *freeobj;

char*
amltname(int t)
{
	static char* amltnames[] = {
		"none", "BUG", "slice", "scope", "method", "dev",
		"proc", "pwr", "thermal", "ref", "local", "arg", "handle", };

	if(t >= 0 && t < nelem(amltnames))
		return amltnames[t];
	switch(t){
	case Obuf:
		return "buf";
	case Oevent:
		return "event";
	case Oint:
		return "int";
	case Oname:
		return "name";
	case Omutex:
		return "mutex";
	case Oany:
		return "any";
	case Opkg:
		return "pkg";
	case Oreg:
		return "reg";
	case Ostr:
		return "str";
	case Odecstr:
		return "decstr";
	case Ofield:
		return "field";
	default:
		return "unknown";
	}
}

void*
amlmalloc(int l)
{
	void *p;

	p = malloc(l);
	if(p == nil)
		panic("aml: no memory");
	memset(p, 0, l);
	return p;
}

/*
 * Push a nested environment continuing evaluation
 * from where the parent is right now.
 * dot, ps, pe, and table may be nil when not a top-level environment,
 * and those of the current environment are used in that case.
 */
void
amlpush(Env *new, char* name, uchar *ps, uchar *pe)
{
	Env *env;

	memset(new, 0, sizeof(Env));
	env = new->prev = amlenv;
	new->name = name;
	assert(env != nil || ps != nil);
	assert(env != nil || pe != nil);

	if(ps != nil && env != nil && (ps < env->ps || ps > env->pe))
		amlwarn("amlpush: wrong ps at %#ulx", getcallerpc(&new));
	if(pe != nil && env != nil && (pe < env->ps || pe > env->pe))
		amlwarn("amlpush: wrong pe at %#ulx", getcallerpc(&new));
	if(ps != nil)
		new->ps = ps;
	else
		new->ps = env->ps;
	if(pe != nil)
		new->pe = pe;
	else
		new->pe = env->pe;
	if(env != nil){
		new->ic = env->ic;
		new->lvl = env->lvl+1;
		new->op = env->op;
		new->flags = env->flags;
		new->table = env->table;
		new->dot = env->dot;
		new->mp = env->mp;
	}else
		new->dot = acpins;

	if(new->lvl > maxdepth)
		maxdepth = new->lvl;
	if(minsp == 0 || (uchar*)&env < minsp)
		minsp = (uchar*)&env;
	amlenv = new;
}

/*
 * Pop out the current environment but update the parent's
 * idea of what's evaluated so far and the pc. This works fine
 * most of the time, but for calling methods. Upon return, the
 * calling sequence is responsible for restoring ps to point
 * past the call, and not past the method aml.
 */
Env*
amlpop(void)
{
	Env *old;
	Aobj *ol;
	int i;
	Mframe *mp;

	if(amlenv == nil)
		panic("pop on an empty env stack");

	old = amlenv->prev;
	if(old != nil)
		old->ps = amlenv->ps;
	while(amlenv->olist.hd != nil){
		ol = amlenv->olist.hd;
		amlenv->olist.hd = ol->next;
		amlclose(ol);
	}
	amlenv->olist.n = 0;
	for(i = 0; i < amlenv->nargs; i++){
		amlclose(amlenv->args[i]);
		amlenv->args[i] = nil;
	}
	amlenv->nargs = 0;
	mp = amlenv->mp;
	if(mp != nil && (amlenv->prev == nil || amlenv->prev->mp != mp)){
		/* method call; collect objects */
		for(i = 0; i < mp->nargs; i++){
			amlclose(mp->args[i]);
			mp->args[i] = nil;
		}
		for(i = 0; i < mp->nlcls; i++){
			amlclose(mp->lcls[i]);
			mp->lcls[i] = nil;
		}
	}
	amlclose(amlenv->res);
	amlenv->res = nil;

	return amlenv = old;
}

int
amlnext(void)
{
	if(amlenv->ps >= amlenv->pe)
		amlerror("end of program");
	amlenv->ps++;
	if(amlenv->ps < amlenv->pe)
		return amlenv->ps[0];
	else
		return 0;
}

void
popamlerror(void)
{
	if (amlerrstk.nerr-- == 0)
		panic("popamlerror w/o wasamlerror");
}

void
amlerror(char* msg, ...)
{
	va_list arg;
	char *s;
	char *se;

	if(msg != nil){
		va_start(arg, msg);
		s = amlerrstk.msg;
		se = amlerrstk.msg + sizeof(amlerrstk.msg);
		s = vseprint(s, se, msg, arg);
		if(amlenv != nil)
			seprint(s, se, " at %08p", amlenv->ps);
		va_end(arg);

		if(!isamlbranch())
			dprint("amlerror: %s\n", amlerrstk.msg);
		if(!isamlbranch() && amlenv != nil && (amlenv->flags&Eabort) != 0){
			amlstackdump();
			panic("acpi abort");
		}
	}
	if(amlerrstk.nerr-- == 0)
		panic("aml error stack is empty");
	gotolabel(&amlerrstk.label[amlerrstk.nerr]);
}

void
amlwarn(char* msg, ...)
{
	va_list arg;
	char buf[128];
	char *s;

	s = buf;
	if(amlenv != nil)
		s = seprint(s, buf+sizeof(buf), "pc=%p: ", amlenv->ps);
	s = seprint(s, buf+sizeof(buf), "warning: ");
	va_start(arg, msg);
	vseprint(s, buf+sizeof(buf), msg, arg);
	va_end(arg);
	xprint("%s", buf);
}

/*
 * Is the error raised just a branch, and not an error?
 * we use "break", "continue", and "return" to implement those.
 */
int
isamlbranch(void)
{
	char *err;

	err = amlerrstk.msg;
	return strncmp(err, "break at", 8) == 0 ||
		strncmp(err, "return at", 9) == 0 ||
		strncmp(err, "continue at", 11) == 0;
}

Aobj*
amlnew(int t)
{
	Aobj *o;

	o = freeobj;
	if(o != nil){
		freeobj = freeobj->next;
		if(o->type != Ofree)
			panic("acpi: free object has been used");
		memset(o, 0, sizeof(Aobj));
	}else
		o = amlmalloc(sizeof(Aobj));
	o->type = t;
	o->ref = 1;
	naobj++;
	if(maxnaobj < naobj)
		maxnaobj = naobj;
	return o;
}

Aobj*
amlnewi(int v)
{
	Aobj *o;

	o = amlnew(Oint);
	o->ival = v;
	return o;
}

Aobj*
amlnews(char* v)
{
	Aobj *o;

	o = amlnew(Ostr);
	if(v == nil)
		v = "";
	kstrdup(&o->sval, v);
	return o;
}

void
amltcheck(Aobj *o, int t, int nullrefok)
{
	uintptr pc;

	pc = getcallerpc(&o);
	if(o == nil)
		amlerror("amltcheck: %#p null object", pc);
	if(o->type != t)
		amlerror("amltcheck: %#p %O is not of type %#02x\n", pc, o, t);
	if(nullrefok != 0)
		return;
	switch(o->type){
	case Oint:
	case Oscope:
	case Omethod:
	case Odev:
	case Oproc:
	case Opwr:
	case Othermal:
	case Ohandle:
	case Oevent:
	case Omutex:
	case Opkg:
	case Oreg:
		return;
	case Ofield:
		if(o->field.idx != nil)
			if(o->field.idx->type != Ofield || o->field.idxval == nil)
				amlerror("bad index in field");
		if(o->field.bank != nil){
			if(o->field.bank->type != Ofield || o->field.bankval == nil)
				amlerror("bad bank in field");
			if(o->field.data == nil || o->field.data->type != Ofield)
				amlerror("bad data in field");
		}
		if(o->field.idx != nil || o->field.reg != nil)
			return;
		break;
	case Obuf:
		if(o->buf.p != nil)
			return;
		break;
	case Oname:
	case Ostr:
	case Odecstr:
		if(o->sval != nil)
			return;
		break;
	case Oslice:
		if(o->slice.src != nil){
			if(o->slice.src->type != Obuf)
				amlerror("slice src is not a buffer");
			return;
		}
		break;
	case Oref:
		if(o->oref.o != nil)
			return;
		break;
	default:
		panic("amltcheck: %#p objtype is %s", pc, amltname(o->type));
	}
	amlerror("amltcheck: %#p %O has a null value", pc, o);
}

static void
freeres(Res *res)
{
	Res *r;

	while(res != nil){
		r = res;
		res = r->next;
		if(r->type == Rpir)
			amlclose(r->pir.src);
		free(r);
	}
}

void
amlclean(Aobj *o)
{
	Aobj *l;

	switch(o->type){
	case Onone:
	case Oint:
	case Oscope:
	case Omethod:
	case Opwr:
	case Oproc:
	case Othermal:
	case Ohandle:
	case Omutex:
	case Oevent:
		break;
	case Odev:
		free(o->dev.name);
		freeres(o->dev.res);
		freeres(o->dev.other);
		o->dev.name = nil;
		o->dev.res = o->dev.other = nil;
		break;
	case Opkg:
		while(o->list.hd != nil){
			l = o->list.hd;
			o->list.hd = l->next;
			amlclose(l);
		}
		break;
	case Oreg:
		if(o->reg.p != nil)
			vunmap(o->reg.p, o->reg.len);
		break;
	case Oref:
	case Oarg:
	case Olocal:
		amlclose(o->oref.o);
		break;
	case Ostr:
	case Odecstr:
	case Oname:
		free(o->sval);
		break;
	case Ofield:
		amlclose(o->field.reg);
		amlclose(o->field.bank);
		amlclose(o->field.bankval);
		free(o->field.fname);
		amlclose(o->field.idx);
		amlclose(o->field.idxval);
		amlclose(o->field.data);
		break;
	case Oslice:
		amlclose(o->slice.src);
		break;
	case Obuf:
		free(o->buf.p);
		break;
	default:
		panic("amlclose: unknown type %#x\n", o->type);
	}
	o->type = Onone;
}

void
amlclose(Aobj *o)
{
	if(o == nil || decref(o) > 0)
		return;
	if(o->type == Ofree)
		panic("acpi: double free");
	if(o->name != nil)
		panic("amlclose: bound %O", o);
	naobj--;
	amlclean(o);
	o->type = Ofree;
	o->next = freeobj;
	freeobj = o;
}

static Aname*
newname(Aname *p, char *n)
{
	Aname *nm;

	nm = amlmalloc(sizeof(Aname));
	strncpy(nm->s, n, 4);
	nm->s[4] = 0;
	nm->parent = p;
	nm->path = nil;
	nm->hasdevs = 0;
	if(p != nil){
		nm->path = amlmalloc(strlen(p->path) + 1 + 4 + 1);
		strcpy(nm->path, p->path);
		if(p->parent != nil)
			strcat(nm->path, "/");
		strcat(nm->path, nm->s);
		if(p->childtl != nil)
			p->childtl->sibling = nm;
		p->childtl = nm;
		if(p->childhd == nil)
			p->childhd = nm;
	}else
		kstrdup(&nm->path, n);
	if(amlenv != nil && amlenv->mp != nil){
		/* record table and method call, for unbind */
		nm->table = amlenv->table;
		nm->mnext = amlenv->mp->binds;
		amlenv->mp->binds = nm;
	}
	nanames++;
	return nm;
}

/* Didn't they have enough? */
static int
amlmatch(char *n, char *s)
{
	while(n[0] != 0 && n[0] == s[0]){
		n++;
		s++;
	}
	while(n[0] == '_' && s[0] == 0)
		n++;
	return n[0] == 0 && s[0] == 0;
}

static Aname*
walkup(Aname *n, char *s)
{
	Aname *cn;

	for(; n != nil; n = n->parent)
		for(cn = n->childhd; cn != nil; cn = cn->sibling){
			if(amlmatch(cn->s, s))
				return cn;
		}
	return nil;
}

Aname*
amlwalk(char *s, int how, Aname *dot)
{
	Aname *n, *cn;
	char *d;
	char *els[20];
	int nels, i;

	n = dot;
	if(n == nil)
		if(amlenv == nil)
			n = acpins;
		else
			n = amlenv->dot;
	if(0 && how == Walkns)
		dbprint("walk '%s' dot=%s\n", s, n->path);

	if(strlen(s) == 4 && how == Walkns)	/* NS search rule */
		return walkup(n, s);

	while(s[0] == '\\' || s[0] == '/'){
		while(n->parent != nil)
			n = n->parent;
		s++;
	}
	while(s[0] == '^'){
		if(n->parent != nil)
			n = n->parent;
		s++;
	}
	if(s[0] == 0)
		return n;
	d = nil;
	kstrdup(&d, s);
	nels = getfields(d, els, nelem(els), 0, "./");
	for(i = 0; n != nil && i < nels; i++){
		for(cn = n->childhd; cn != nil; cn = cn->sibling)
			if(amlmatch(cn->s, els[i]))
				break;
		if(cn == nil)
			if(how == Walkcreating)
				cn = newname(n, els[i]);
			else{
				free(d);
				return nil;
			}
		n = cn;
	}
	if(n == nil && how == Walkcreating)
		panic("bug: amlwalk: nil name but we are creating it");
	free(d);
	return n;
}

/*
 * objects may be bound to more than one name.
 * we only record the first one in the object.
 */
void
amlbind(char *s, Aobj *o)
{
	Aname *n;

	n = amlwalk(s, Walkcreating, nil);
	if(n == nil)
		amlerror("can't bind '%s'", s);
	incref(o);
	if(n->o != nil){
		n->o->name = nil;
		amlclose(n->o);
	}
	n->o = o;
	if(o->name == nil)
		o->name = n;
	dbprint("bind %s -> %O\n", n->path, o);
	if(o->type == Odev)
		for(n = n->parent; n != nil; n = n->parent)
			n->hasdevs++;
}

void
amlunbind(Aname *n)
{
	Aname *pn, *l, *prev;

	dbprint("unbind %s\n", n->path);
	pn = n->parent;
	if(pn == nil)
		panic("acpi: unbind of root name");
	prev = nil;
	for(l = pn->childhd; l != nil && l != n; l = l->sibling)
		prev = l;
	if(l == nil)
		panic("acpi: name not in parent");
	if(prev == nil)
		pn->childhd = n->sibling;
	else
		prev->sibling = n->sibling;
	if(n == pn->childtl)
		pn->childtl = prev;

	n->o->name = nil;
	amlclose(n->o);
	while(n->childhd != nil)
		amlunbind(n->childhd);
	free(n->path);
	free(n);
}

static int
tfmt(Fmt *f)
{
	char buf[80];
	int n;

	n = va_arg(f->args, int);
	assert(n >= 0);
	switch(n){
	case 0:
		return fmtstrcpy(f, "");
	case 1:
		return fmtstrcpy(f, "\t");
	default:
		if(n >= sizeof(buf)/3-3)
			n = sizeof(buf)/3-3;
		buf[0] = '\t';
		n--;
		memset(buf+1, ' ', n*3);
		buf[1+n*3] = 0;
		return fmtstrcpy(f, buf);
	}
}

void
dumpstats(void)
{
	dprint("\n%d objects allocated\n", naobj);
	dprint("\n%d objects allocated at most \n", maxnaobj);
	dprint("%d names allocated\n", nanames);
	dprint("max stack depth %d\n", maxdepth);
	if(minsp != nil)
		dprint("max stack size %ld\n", (uchar*)acpistackbottom - minsp);
}

void
amlenvdump(Env *e)
{
	int i, lvl;
	Aobj *o;
	Mframe *mp;

	lvl = 0;
	dprint("%>env %s %#p prev %#p ps %#p pe %#p\n",
		lvl, e->name, e, e->prev, e->ps, e->pe);
	dprint("%>flg %#x lvl %d ic %#x opp0 %#p",
		lvl, e->flags, e->lvl, e->ic, e->opp0);
	if(e->flags&Enoeval)
		dprint(" noeval");
	if(e->op != nil)
		dprint(" op=%s", e->op->name);
	if(e->dot != nil)
		dprint(" dot=%s\n", e->dot->path);
	else
		dprint("\n");
	lvl++;
	if(e->res != nil)
		dprint("%>res  %O\n", lvl, e->res);
	for(i = 0; i < nelem(e->args); i++)
		if(e->args[i] != nil)
			if(i < e->nargs)
				dprint("%>arg%d %O\n", lvl, i, e->args[i]);
			else
				panic("amlenvdump: argument used beyond nargs");
	mp = e->mp;
	if(mp != nil && (e->prev == nil || e->prev->mp != mp)){
		dprint("%>method frame:\n", lvl);
		for(i = 0; i < mp->nargs; i++)
			if(mp->args[i] != nil)
				dprint("%>marg%d %O\n", lvl+1, i, mp->args[i]);
		for(i = 0; i < mp->nlcls; i++)
			if(mp->lcls[i] != nil)
				dprint("%>lcl%d %O\n", lvl, i, mp->lcls[i]);
	}
	for(o = e->olist.hd; o != nil; o = o->next)
		dprint("%>obj  %O\n", lvl, o);
	dprint("\n");
}

void
amlstackdump(void)
{
	Env *e;

	xprint("aml stack:\n");
	for(e = amlenv; e != nil; e = e->prev)
		amlenvdump(e);
}

void
dumpns(Aname *ns, int d, int objs)
{
	Aname *cn;

	xprint("%>%s", d, ns->path);
	if(objs && ns->o != nil && ns->o->type != Oscope)
		xprint("= %O", ns->o);
	if(ns->childhd != nil){
		xprint(" {\n");
		for(cn = ns->childhd; cn != nil; cn = cn->sibling)
			dumpns(cn, d+1, objs);
		xprint("%>}\n", d);
	}else
		xprint("\n");
}

static int
Pfmt(Fmt *f)
{
	int lvl;

	if(amlenv == nil)
		return 0;
	lvl = amlenv->lvl;
	if(f->flags & FmtSharp)
		lvl++;
	return fmtprint(f, "%08p: %>", amlenv->ps, lvl);
}


static int
Ofmt(Fmt *f)
{
	static int verb = 0;
	Aobj *o, *ol;
	char str[40], *s;
	int i;

	o = va_arg(f->args, Aobj*);
	if(o == nil)
		return fmtprint(f, "<null>");
	str[0] = 0;
	s = str;
	if(verb || (f->flags&FmtSharp) != 0 ||
	   (o->type != Ostr && o->type != Oint && o->type != Ostr))
		s = seprint(str, str+sizeof(str), "%s", amltname(o->type));
	if(verb || (f->flags&FmtSharp) != 0){
		s = seprint(s, str+sizeof(str),  " r=%uld %#p", o->ref, o);
		if(o->name != nil)
			s = seprint(s, str+sizeof(str), " bound");
	}
	switch(o->type){
	case Oint:
	case Ohandle:
		return fmtprint(f, "%s %#ullx", str, o->ival);
	case Oname:
	case Ostr:
	case Odecstr:
		return fmtprint(f, "%s \"%s\"", str, o->sval);
	case Ofield:
		fmtprint(f, "%s %s", str, o->field.fname);
		if(o->field.idx != nil)
			fmtprint(f, " idx %s[%#llx]%s %#ulx:%uld[%#x]",
				o->field.idx->field.fname, o->field.idxval->ival,
				o->field.data->field.fname,
				o->field.off/8, o->field.off&7,
				o->field.len);
		else {
			if(o->field.reg != nil)
				fmtprint(f, " %s", o->field.reg->name->s);
			if(o->field.bank != nil)
				fmtprint(f, " bnk %s[%#llx]",
					o->field.bank->field.fname,
					o->field.bankval->ival);
			fmtprint(f, " %#ulx:%uld[%#x]",
				o->field.off/8, o->field.off&7,
				o->field.len);
		}
		return fmtprint(f, " a%xs%xu%x", o->field.accsz, o->field.locking,
				o->field.update);
	case Opkg:
		fmtprint(f, "%s[%d] {", str, o->list.n);
		i = 0;
		for(ol = o->list.hd; ol != nil; ol = ol->next){
			if(o->list.n > 5 || ol->type == Opkg)
				fmtprint(f, "\n\t\t\t[%2d] ", i++);
			fmtprint(f, "%O%s", ol, ol->next ? ", " : "");
		}
		return fmtprint(f, "}");
	case Obuf:
		fmtprint(f, "%s %#p[%#lx]{", str, o, o->buf.len);
		if(o->buf.len > 6)
			fmtprint(f, "\n\t\t\t");
		for(i = 0; i < o->buf.len; i++){
			fmtprint(f, "%02x ", o->buf.p[i]);
			if((i % 16) == 15 && o->buf.len > 6)
				fmtprint(f, "\n\t\t\t");
		}
		return fmtprint(f, "}");
	case Oslice:
		fmtprint(f, "%s %#ulx:%ld[%#ulx]",
				str, o->slice.off/8, o->slice.off&7, o->slice.len);
		if(o->slice.src != nil && o->slice.src->type == Obuf)
			return fmtprint(f, " src %#p", o->slice.src);
		else
			return fmtprint(f, " src %O", o->slice.src);
	case Oscope:
		return fmtprint(f, "%s %s", str, o->scope.name);
	case Omethod:
		if(verb || (f->flags&FmtSharp) != 0)
			return fmtprint(f, "%s %s/%d %s sl%d ps %#p pe %#p",
				str, o->method.name, o->method.nargs,
				o->method.isexcl ? "excl " : "",
				o->method.synclvl,
				o->method.ps, o->method.pe);
		else
			return fmtprint(f, "%s %s/%d %#p",
				str, o->method.name, o->method.nargs, o->method.ps);
	case Odev:
		fmtprint(f, "%s %s", str, o->dev.name);
		if(o->dev.pnpid != nil)
			fmtprint(f, " pnp %s", o->dev.pnpid);
		return fmtprint(f, " uid %#ullx addr %#llux",
			o->dev.uid, o->dev.addr);
	case Oproc:
		return fmtprint(f, "%s %s id %#x baddr %#ulx blen %d",
			str, o->proc.name, o->proc.id, o->proc.baddr, o->proc.blen);
	case Opwr:
		return fmtprint(f, "%s %s syslvl %#x rorder %#x",
			str, o->pwr.name, o->pwr.syslvl, o->pwr.rorder);
	case Othermal:
		return fmtprint(f, "%s %s", str, o->thermal.name);
	case Oreg:
		return fmtprint(f, "%s %s base %#ullx p %#p len %#ullx",
			str, acpiregstr(o->reg.spc), o->reg.base,
			o->reg.p, o->reg.len);
	case Omutex:
	case Oevent:
		return fmtprint(f, "%s %#p", str, o);
	case Oarg:
	case Olocal:
		seprint(s, str+sizeof(str), "#%d ", o->oref.no);
		/* fall */
	case Oref:
		if(o->oref.o == nil)
			return fmtprint(f, "%s", str);
		if(verb || (f->flags&FmtSharp) != 0)
			return fmtprint(f, "%s-> %#O", str, o->oref.o);
		else
			return fmtprint(f, "%s-> %O", str, o->oref.o);
	default:
		return fmtprint(f, "<type %#x>", o->type);
	}
}

Aobj*
amlcall(Aobj *mo, Aobj **args, Atable *table, int flags)
{
	Env toplvl;
	Mframe mfr;
	Aobj *r;

	memset(&mfr, 0, sizeof(mfr));
	amlpush(&toplvl, "toplvl", mo->method.ps, mo->method.pe);
	amlenv->table = table;
	amlenv->flags = flags;
	amlenv->mp = &mfr;
	r = nil;
	if(!wasamlerror()){
		r = opcall(mo, args);
		amlenv->res = nil;	/* don't close r */
		popamlerror();
		if(r == nil)
			r = amlnewi(0);	/* return something */
	}
	amlpop();
	return r;

}

/*
 * Evaluate code is [ps:pe]. We pretend we are a method call
 * so that dry runs may still find arguments and locals.
 */
char*
amleval(uchar *ps, uchar *pe, Aobj **po, Aname *n, Atable* table, int flags)
{
	Env toplvl;
	Aobj *o;
	Mframe mfr;

	memset(&mfr, 0, sizeof(mfr));
	mfr.nargs = Nmargs;
	amlpush(&toplvl, "toplvl", ps, pe);
	if(n != nil)
		amlenv->dot = n;
	amlenv->table = table;
	amlenv->flags = flags;
	amlenv->mp = &mfr;
	if(wasamlerror()){
		amlpop();
		return amlerrstk.msg;
	}

	o = pamlblock(Dontkeep);
	if(po != nil)
		*po = o;
	else
		amlclose(o);
	popamlerror();
	amlpop();
	return nil;
}

static void
osi(Aobj*)
{
	int i;
	Aobj *arg;

	static char *osis[] = {
		"Windows 2000",
		"Windows 2001",
		"Windows 2001.1",
		"Windows 2001 SP0",
		"Windows 2001 SP1",
		"Windows 2001 SP2",
		"Windows 2001 SP3",
		"Windows 2001 SP4",
		"Windows 2006",
	};

	arg = amlenv->mp->args[0];
	if(arg == nil || arg->type != Ostr)
		amlerror("argument is not a string");
	amlenv->res = amlnewi(0);
	for(i = 0; i < nelem(osis); i++)
		if(strcmp(arg->sval, osis[i]) == 0){
			amlenv->res->ival = 1;
			break;
		}
	amlerror("return");
}

/*
 * Most of the _INI AML code I've seen defines the OS as "Linux"
 * if it identifies itself as BSD. However, BSD code identifies itself
 * as windows.
 * We could either impersonate Windows or Linux. Otherwise, all the
 * bugs will be for us (besides our own ones).
 */
void
amlinit(void)
{
	Aobj *o;

	fmtinstall('O', Ofmt);
	fmtinstall('>', tfmt);
	fmtinstall('!', Pfmt);
	initamlops();
	acpins = newname(nil, "/");
	o = amlnews("Microsoft Windows NT"); 	/* sic */
	amlbind("_OS_", o);
	amlclose(o);
	o = amlnewi(9);
	amlbind("_REV", o);
	amlclose(o);
	acpilock = amlnew(Omutex);
	amlbind("_GL_", acpilock);
	amlclose(acpilock);
	o = amlnew(Omethod);
	o->method.name = "_OSI";
	o->method.nargs = 1;
	o->method.f = osi;
	o->method.ps = o->method.pe = (uchar*)osi;
	amlbind("_OSI", o);
	amlclose(o);
}

/* debug prints and output. only for a single thread.
 * Trying to get things ok for both bio and the kernel print.
 * b -> bind; x -> exec; i -> instruction; d -> any debug; 0 -> print
 */
void
amlprint(char d, char *fmt, ...)
{
	va_list v;
	int flags;

	flags = acpiflags;
	if(amlenv != nil)
		flags |= amlenv->flags;
	switch(d){
	case 'x':
		if((flags & Exdump) == 0)
			return;
		break;
	case 'i':
		if((flags & Eidump) == 0)
			return;
		break;
	case 'b':
		if((flags & Ebdump) == 0)
			return;
		break;
	case 'o':
		if((flags & Eodump) == 0)
			return;
		break;
	case 'd':
		if((flags & Edebugflags) == 0)
			return;
		break;
	}
	va_start(v, fmt);
	vfprint(2, fmt, v);
	va_end(v);
}

/*
 * This part is pure paranoia and could probably go.
 * It's a full check on all objects on the aml stack,
 * to ensure that objects seem to be ok and that references are
 * ok also.
 * It's easy to make a mistake regarding RCs while implementing
 * some ops and this is a safeguard.
 * This check is activated when any debug flag is active.
 */
static int dumprefs;

static int
amlforall(Aname *n, int (*f)(Aobj*,void*), void *a)
{
	int r, c;
	Aname *cn;

	r = 0;
	if(n->o != nil){
		c = f(n->o, a);
		if(c != 0 && dumprefs)
			dprint("\tns %s %#p\n", n->path, n->o);
		r += c;
	}
	for(cn = n->childhd; cn != nil; cn = cn->sibling)
		r += amlforall(cn, f, a);
	return r;
}

static int
forallargs(Env *e, Aobj **args, char *s, int nargs, int(*f)(Aobj*,void*), void* a)
{
	int i, r, n;

	if(args == nil)
		return 0;
	r = 0;
	for(i = 0; i < nargs; i++)
		if(args[i] != nil){
			n = f(args[i], a);
			if(n != 0 && dumprefs)
				dprint("\tenv %#p %s%d %#p\n", e, s, i, args[i]);
			r += n;
		}
	return r;
}

static int
forallobjs(int (*f)(Aobj*,void*), void *a)
{
	Env *e;
	int r, i, n;
	Aobj *o;
	Mframe *mp;

	r = 0;
	for(e = amlenv; e!= nil; e = e->prev){
		if(e->res != nil){
			n = f(e->res, a);
			if(n != 0 && dumprefs)
				dprint("\tenv %#p res %#p\n", e, e->res);
			r += n;
		}
		for(o = e->olist.hd; o != nil; o = o->next){
			n = f(o, a);
			if(n != 0 && dumprefs)
				dprint("\tenv %#p olist %#p\n", e, o);
			r += n;
		}
		for(i = 0; i < e->nargs; i++){
			if(e->args[i] == nil)
				continue;
			n = f(e->args[i], a);
			if(n != 0 && dumprefs)
				dprint("\tenv %#p arg%d %#p\n", e, i, e->args[i]);
			r += n;
		}
		mp = e->mp;
		if(mp != nil && (e->prev == nil || e->prev->mp != mp)){
			r += forallargs(e, mp->args, "marg", mp->nargs, f, a);
			r += forallargs(e, mp->targs, "tmarg", Nmargs, f, a);
			r += forallargs(e, mp->lcls, "lcl", mp->nlcls, f, a);
		}
	}

	r += amlforall(acpins, f, a);
	return r;
}

enum {Nrefs = 1023};

static int
addref(Aobj **refs, Aobj *o)
{
	int i;

	for(i = 0; i < Nrefs; i++)
		if(refs[i] == nil){
			refs[i] = o;
			return 1;
		}else if(refs[i] == o)
			return 0;
	panic("countrefs: increase Nrefs %d\n", Nrefs);
	return 1;
}

static int
countrefs(Aobj *o, void *a)
{
	Aobj **refs = a;
	Aobj *arg;
	Aobj *l;
	int nr;

	arg = refs[0];
	refs++;

	if(o == nil)
		panic("aml: contrefs on nil obj");
	if(o == arg)
		return 1;
	switch(o->type){
	case Oref:
	case Olocal:
	case Oarg:
		if(o->oref.o == arg)
			return addref(refs, o);
		else if(o->oref.o != nil)
			return countrefs(o->oref.o, refs-1);
		break;
	case Ofield:
		nr = 0;
		if(o->field.reg == arg)
			nr += addref(refs, o);
		if(o->field.bank == arg)
			nr += addref(refs, o);
		if(o->field.bankval == arg)
			nr += addref(refs, o);
		if(o->field.idx == arg)
			nr += addref(refs, o);
		if(o->field.idxval == arg)
			nr += addref(refs, o);
		if(o->field.data == arg)
			nr += addref(refs, o);
		return nr;
	case Oslice:
		if(o->slice.src == arg)
			return addref(refs, o);
		break;
	case Opkg:
		for(l = o->list.hd; l != nil; l = l->next)
			if(l == arg)
				return addref(refs, o);
		break;
	}
	return 0;
}

static int
checkobj(Aobj *o, void *)
{
	int nrefs;
	Aobj **refs;

	refs = amlmalloc(sizeof(Aobj*) * (Nrefs+1));
	if((char*)o < end || (char*)o > (char*)&o){
		amlwarn("bad object address %#p", o);
		amlstackdump();
		panic("acpi abort");
	}
	refs[0] = o;
	nrefs = forallobjs(countrefs, refs);
	if(nrefs != o->ref){
		amlwarn("object %#O\n\t%d refs found on stack.\n", o, nrefs);
		dumprefs=1;
		forallobjs(countrefs, refs);
		amlstackdump();
		panic("acpi stack check failed");
	}
	free(refs);
	return 0;
}

void
amlstackcheck(void)
{
	forallobjs(checkobj, nil);
}

