#include "kernel.h"
#include "acpi.h"
#include "aml.h"

/*
 * Configure apics and interrupts.
 * This is only necessary for:
 * - hot plugging of processors.
 * - parsing entries obtained from other systems.
 *
 * MADT entries only guarantee that the first lapic is
 * for the boot processor. Processing can't be linear
 * because entries for an apic might be found before
 * the entry declaring the apic.
 *
 * System interrupt numbers are virtual, with each
 * ioapic knowing the base number for its first input.
 *
 * Interrupt flags are always converted into Apic flags.
 */

static Madt	*apics;

static void
acpilapic(Apicst *a)
{
	Apicst *l;
	uint lint[2];

	lint[0] = lint[1] = -1;

	for(l = apics->st; l != nil; l = l->next)
		if(l->type == ASlnmi && l->lnmi.pid == a->lapic.pid){
			if(l->lnmi.lint != 0 && l->lnmi.lint != 1)
				dprint("acpilapic: bad lint %d\n", l->lnmi.lint);
			else
				lint[l->lnmi.lint] = l->lnmi.flags;
		}
	acpictl("lapic %d %#llux %I %I",
		a->lapic.id, apics->laddr, lint[0], lint[1]);
}

void
acpiapics(void)
{
	Apicst *l;
	Aobj *o, *args[Nmargs];
	Aname *n;
	int nlapics;

	if(apics == nil)
		return;

	n = amlwalk("/PIC", Walkdev, acpins);
	if(n != nil && n->o != nil && n->o->type == Omethod){
		args[0] = amlnewi(Papic);
		args[1] = nil;
		o = amlcall(n->o, args, nil, acpiflags);
		if(o == nil)
			xprint("call to _PIC failed\n");
		amlclose(o);
	}

	/*
	 * Configure apics. The first one is always the boot processor.
	 * Ensure IO apics do not get same ids than local apics.
	 */
	nlapics = 0;
	for(l = apics->st; l != nil; l = l->next)
		if(l->type == ASlapic){
			nlapics++;
			acpilapic(l);
		}
	for(l = apics->st; l != nil; l = l->next)
		if(l->type == ASioapic){
			acpictl("ioapic %d %#llux %d",
				l->ioapic.id + nlapics, l->ioapic.addr,
				l->ioapic.ibase);
			break;
		}

	/*
	 * Add interrupt info now that we have configured all apics.
	 */
	for(l = apics->st; l != nil; l = l->next)
		switch(l->type){
		case ASintovr:
			acpictl("int isa %d %d %I",
				l->intovr.irq, l->intovr.intr,
				l->intovr.flags);
			break;
		case ASnmi:
			acpictl("int isa 2 %d %I",
				l->nmi.intr, l->nmi.flags|IntNMI);
			break;
		}
	
}

static void
dumpmadt(Madt *apics)
{
	Apicst *st;

	if((acpiflags&Edebugflags) == 0)
		return;
	dprint("acpi: madt lapic addr %llux pcat %d:\n", apics->laddr, apics->pcat);
	for(st = apics->st; st != nil; st = st->next)
		switch(st->type){
		case ASlapic:
			dprint("\tlapic pid %d id %d\n", st->lapic.pid, st->lapic.id);
			break;
		case ASioapic:
		case ASiosapic:
			dprint("\tioapic id %d addr %#llux ibase %d\n",
				st->ioapic.id, st->ioapic.addr, st->ioapic.ibase);
			break;
		case ASintovr:
			dprint("\tintovr irq %d intr %d %I\n",
				st->intovr.irq, st->intovr.intr,
				st->intovr.flags);
			break;
		case ASnmi:
			dprint("\tnmi intr %d %I\n", st->nmi.intr,
				st->nmi.flags);
			break;
		case ASlnmi:
			dprint("\tlnmi pid %d lint %d %I\n",
				st->lnmi.pid, st->lnmi.lint,
				st->lnmi.flags);
			break;
		case ASlsapic:
			dprint("\tlsapic pid %d id %d eid %d puid %d puids %s\n",
				st->lsapic.pid, st->lsapic.id,
				st->lsapic.eid, st->lsapic.puid,
				st->lsapic.puids);
			break;
		case ASintsrc:
			dprint("\tintr type %d pid %d peid %d"
				" iosv %d intr %d %I\n",
				st->type, st->intsrc.pid,
				st->intsrc.peid, st->intsrc.iosv,
				st->intsrc.intr, st->intsrc.flags);
			break;
		case ASlx2apic:
			dprint("\tlx2apic puid %d id %d\n", st->lx2apic.puid, st->lx2apic.id);
			break;
		case ASlx2nmi:
			dprint("\tlx2nmi puid %d intr %d %I\n",
				st->lx2nmi.puid, st->lx2nmi.intr,
				st->lx2nmi.flags);
			break;
		default:
			dprint("\t<unknown madt entry>\n");
		}
	dprint("\n");
}

int
acpimadt(uchar *p, int len)
{
	uchar *pe;
	Apicst *st, *l, **stl;
	int stlen, id;

	apics = amlmalloc(sizeof(Madt));
	apics->laddr = l32get(p+36);
	apics->pcat = l32get(p+40);
	apics->st = nil;
	stl = &apics->st;
	pe = p + len;
	for(p += 44; p < pe; p += stlen){
		st = amlmalloc(sizeof(Apicst));
			st->type = p[0];
		st->next = nil;
		stlen = p[1];
		switch(st->type){
		case ASlapic:
			st->lapic.pid = p[2];
			st->lapic.id = p[3];
			if(l32get(p+4) == 0){
				free(st);
				st = nil;
			}
			break;
		case ASioapic:
			st->ioapic.id = p[2];
			st->ioapic.addr = l32get(p+4);
			st->ioapic.ibase = l32get(p+8);
			break;
		case ASintovr:
			st->intovr.irq = p[3];
			st->intovr.intr = l32get(p+4);
			st->intovr.flags = l16get(p+8);
			break;
		case ASnmi:
			st->nmi.flags = IntNMI|l16get(p+2);
			st->nmi.intr = l32get(p+4);
			break;
		case ASlnmi:
			st->lnmi.pid = p[2];
			st->lnmi.flags = IntNMI|l16get(p+3);
			st->lnmi.lint = p[5];
			break;
		case ASladdr:
			apics->laddr = l64get(p+8);
			break;
		case ASiosapic:
			id = st->iosapic.id = p[2];
			st->iosapic.ibase = l32get(p+4);
			st->iosapic.addr = l64get(p+8);
			for(l = apics->st; l != nil; l = l->next)
				if(l->type == ASioapic && l->ioapic.id == id){
					l->ioapic = st->iosapic;
					free(st);
					st = nil;
					break;
				}
			break;
		case ASlsapic:
			st->lsapic.pid = p[2];
			st->lsapic.id = p[3];
			st->lsapic.eid = p[4];
			st->lsapic.puid = l32get(p+12);
			if(l32get(p+8) == 0){
				free(st);
				st = nil;
			}else
				kstrdup(&st->lsapic.puids, (char*)p+16);
			break;
		case ASintsrc:
			st->intsrc.flags = l16get(p+2);
			st->type = p[4];
			st->intsrc.pid = p[5];
			st->intsrc.peid = p[6];
			st->intsrc.iosv = p[7];
			st->intsrc.intr = l32get(p+8);
			st->intsrc.any = l32get(p+12);
			break;
		case ASlx2apic:
			st->lx2apic.id = l32get(p+4);
			st->lx2apic.puid = l32get(p+12);
			if(l32get(p+8) == 0){
				free(st);
				st = nil;
			}
			break;
		case ASlx2nmi:
			st->lx2nmi.flags = IntNMI|l16get(p+2);
			st->lx2nmi.puid = l32get(p+4);
			st->lx2nmi.intr = p[8];
			break;
		default:
			xprint("unknown APIC structure\n");
			free(st);
			st = nil;
		}
		if(st != nil){
			*stl = st;
			stl = &st->next;
		}
	}

	dumpmadt(apics);
	return 0;
}

